@@ -6,8 +6,6 @@
 #include <math.h> /* for : double rint (double) */
 
 
-char *WMBrowserDidScrollNotification = "WMBrowserDidScrollNotification";
-
 
 typedef struct W_Browser {
     W_Class widgetClass;
@@ -19,33 +17,33 @@ typedef struct W_Browser {
     short columnCount;
     short usedColumnCount;	       /* columns actually being used */
     short minColumnWidth;
- 
+
     short maxVisibleColumns;
     short firstVisibleColumn;
 
     short titleHeight;
-    
+
     short selectedColumn;
 
     WMSize columnSize;
-    
+
 
     void *clientData;
     WMAction *action;
     void *doubleClientData;
     WMAction *doubleAction;
 
-    WMBrowserFillColumnProc *fillColumn;
-    
+    WMBrowserDelegate *delegate;
+
     WMScroller *scroller;
 
     char *pathSeparator;
-    
+
     struct {
 	unsigned int isTitled:1;
 	unsigned int allowMultipleSelection:1;
 	unsigned int hasScroller:1;
-	
+
 	/* */
 	unsigned int loaded:1;
 	unsigned int loadingColumn:1;
@@ -79,7 +77,7 @@ static void setupScroller(WMBrowser *bPtr);
 
 static void scrollToColumn(WMBrowser *bPtr, int column, Bool updateScroller);
 
-static void paintItem(WMList *lPtr, int index, Drawable d, char *text, 
+static void paintItem(WMList *lPtr, int index, Drawable d, char *text,
 		      int state, WMRect *rect);
 
 static void loadColumn(WMBrowser *bPtr, int column);
@@ -89,12 +87,15 @@ static void removeColumn(WMBrowser *bPtr, int column);
 static char*
 createTruncatedString(WMFont *font, char *text, int *textLen, int width);
 
-static void resizeBrowser(WMWidget*, unsigned int, unsigned int);
+static void willResizeBrowser(W_ViewDelegate*, WMView*, 
+			      unsigned int*, unsigned int*);
 
-W_ViewProcedureTable _BrowserViewProcedures = {
+W_ViewDelegate _BrowserViewDelegate = {
     NULL,
-	resizeBrowser,
-	NULL
+	NULL,
+	NULL,
+	NULL,
+	willResizeBrowser
 };
 
 
@@ -111,17 +112,19 @@ WMCreateBrowser(WMWidget *parent)
     memset(bPtr, 0, sizeof(WMBrowser));
 
     bPtr->widgetClass = WC_Browser;
-    
+
     bPtr->view = W_CreateView(W_VIEW(parent));
     if (!bPtr->view) {
-	free(bPtr);
+	wfree(bPtr);
 	return NULL;
     }
     bPtr->view->self = bPtr;
 
+    bPtr->view->delegate = &_BrowserViewDelegate;
+
     WMCreateEventHandler(bPtr->view, ExposureMask|StructureNotifyMask
 			 |ClientMessageMask, handleEvents, bPtr);
-	
+
     /* default configuration */
     bPtr->flags.hasScroller = DEFAULT_HAS_SCROLLER;
 
@@ -129,8 +132,8 @@ WMCreateBrowser(WMWidget *parent)
     bPtr->flags.isTitled = DEFAULT_IS_TITLED;
     bPtr->maxVisibleColumns = DEFAULT_MAX_VISIBLE_COLUMNS;
 
-    resizeBrowser(bPtr, DEFAULT_WIDTH, DEFAULT_HEIGHT);
-    
+    WMResizeWidget(bPtr, DEFAULT_WIDTH, DEFAULT_HEIGHT);
+
     bPtr->pathSeparator = wstrdup(DEFAULT_SEPARATOR);
 
     if (bPtr->flags.hasScroller)
@@ -140,13 +143,20 @@ WMCreateBrowser(WMWidget *parent)
 	WMAddBrowserColumn(bPtr);
     }
     bPtr->usedColumnCount = 0;
-    
+
     bPtr->selectedColumn = -1;
 
     return bPtr;
 }
 
 
+int
+WMGetBrowserMaxVisibleColumns(WMBrowser *bPtr)
+{
+    return bPtr->maxVisibleColumns;
+}
+
+
 void
 WMSetBrowserMaxVisibleColumns(WMBrowser *bPtr, int columns)
 {
@@ -154,7 +164,7 @@ WMSetBrowserMaxVisibleColumns(WMBrowser *bPtr, int columns)
     int newFirstVisibleColumn = 0;
 
     assert ((int) bPtr);
-    
+
     columns = (columns < MIN_VISIBLE_COLUMNS) ? MIN_VISIBLE_COLUMNS : columns;
     columns = (columns > MAX_VISIBLE_COLUMNS) ? MAX_VISIBLE_COLUMNS : columns;
     if (columns == bPtr->maxVisibleColumns) {
@@ -196,7 +206,7 @@ WMSetBrowserMaxVisibleColumns(WMBrowser *bPtr, int columns)
 	    removeColumn(bPtr, newFirstVisibleColumn + columns);
 	}
     }
-    resizeBrowser(bPtr, bPtr->view->size.width, bPtr->view->size.height);
+    WMResizeWidget(bPtr, bPtr->view->size.width, bPtr->view->size.height);
     if (bPtr->flags.loaded) {
 	XClearArea(bPtr->view->screen->display, bPtr->view->window, 0, 0,
 		   bPtr->view->size.width, bPtr->titleHeight, False);
@@ -205,7 +215,7 @@ WMSetBrowserMaxVisibleColumns(WMBrowser *bPtr, int columns)
 }
 
 
-int 
+int
 WMGetBrowserNumberOfColumns(WMBrowser *bPtr)
 {
     return bPtr->usedColumnCount;
@@ -215,7 +225,7 @@ void
 WMSetBrowserPathSeparator(WMBrowser *bPtr, char *separator)
 {
     if (bPtr->pathSeparator)
-	free(bPtr->pathSeparator);
+	wfree(bPtr->pathSeparator);
     bPtr->pathSeparator = wstrdup(separator);
 }
 
@@ -226,7 +236,7 @@ drawTitleOfColumn(WMBrowser *bPtr, int column)
 {
     WMScreen *scr = bPtr->view->screen;
     int x;
-    
+
     x=(column-bPtr->firstVisibleColumn)*(bPtr->columnSize.width+COLUMN_SPACING);
 
     XFillRectangle(scr->display, bPtr->view->window, WMColorGC(scr->darkGray), x, 0,
@@ -239,17 +249,17 @@ drawTitleOfColumn(WMBrowser *bPtr, int column)
 	int widthC = bPtr->columnSize.width-8;
 
 	if (WMWidthOfString(scr->boldFont, bPtr->titles[column], titleLen)
-	    > widthC) {     
+	    > widthC) {
 	    char *titleBuf = createTruncatedString(scr->boldFont,
 						   bPtr->titles[column],
 						   &titleLen, widthC);
-	    W_PaintText(bPtr->view, bPtr->view->window, scr->boldFont, x, 
+	    W_PaintText(bPtr->view, bPtr->view->window, scr->boldFont, x,
 			(bPtr->titleHeight-WMFontHeight(scr->boldFont))/2,
 			bPtr->columnSize.width, WACenter, WMColorGC(scr->white),
 			False, titleBuf, titleLen);
 	    free (titleBuf);
 	} else {
-	    W_PaintText(bPtr->view, bPtr->view->window, scr->boldFont, x, 
+	    W_PaintText(bPtr->view, bPtr->view->window, scr->boldFont, x,
 			(bPtr->titleHeight-WMFontHeight(scr->boldFont))/2,
 			bPtr->columnSize.width, WACenter, WMColorGC(scr->white),
 			False, bPtr->titles[column], titleLen);
@@ -258,41 +268,24 @@ drawTitleOfColumn(WMBrowser *bPtr, int column)
 }
 
 
-void
-WMSetBrowserColumnTitle(WMBrowser *bPtr, int column, char *title)
-{
-    assert(column >= 0);
-    assert(column < bPtr->usedColumnCount);
-
-    if (bPtr->titles[column])
-	free(bPtr->titles[column]);
-
-    bPtr->titles[column] = wstrdup(title);
-   
-    if (COLUMN_IS_VISIBLE(bPtr, column) && bPtr->flags.isTitled) {
-	drawTitleOfColumn(bPtr, column);
-    }
-}
-
-
 WMList*
 WMGetBrowserListInColumn(WMBrowser *bPtr, int column)
 {
     if (column < 0 || column >= bPtr->usedColumnCount)
 	return NULL;
-    
+
     return bPtr->columns[column];
 }
 
 
-void 
-WMSetBrowserFillColumnProc(WMBrowser *bPtr, WMBrowserFillColumnProc *proc)
+void
+WMSetBrowserDelegate(WMBrowser *bPtr, WMBrowserDelegate *delegate)
 {
-    bPtr->fillColumn = proc;
+    bPtr->delegate = delegate;
 }
 
 
-int 
+int
 WMGetBrowserFirstVisibleColumn(WMBrowser *bPtr)
 {
     return bPtr->firstVisibleColumn;
@@ -305,9 +298,9 @@ removeColumn(WMBrowser *bPtr, int column)
     int i, clearEnd, destroyEnd;
     WMList **clist;
     char **tlist;
-    
+
     assert ((int) bPtr);
-    
+
     column = (column < 0) ? 0 : column;
     if (column >= bPtr->columnCount) {
 	return;
@@ -326,14 +319,14 @@ removeColumn(WMBrowser *bPtr, int column)
     }
     for (i=column; i < clearEnd; i++) {
 	if (bPtr->titles[i]) {
-	    free(bPtr->titles[i]);
+	    wfree(bPtr->titles[i]);
 	    bPtr->titles[i] = NULL;
 	}
 	WMClearList(bPtr->columns[i]);
     }
     for (;i < destroyEnd; i++) {
 	if (bPtr->titles[i]) {
-	    free(bPtr->titles[i]);
+	    wfree(bPtr->titles[i]);
 	    bPtr->titles[i] = NULL;
 	}
         WMRemoveNotificationObserverWithName(bPtr,
@@ -346,8 +339,8 @@ removeColumn(WMBrowser *bPtr, int column)
     tlist = wmalloc(sizeof(char*) * (bPtr->columnCount));
     memcpy(clist, bPtr->columns, sizeof(WMList*) * (bPtr->columnCount));
     memcpy(tlist, bPtr->titles, sizeof(char*) * (bPtr->columnCount));
-    free(bPtr->titles);
-    free(bPtr->columns);
+    wfree(bPtr->titles);
+    wfree(bPtr->columns);
     bPtr->titles = tlist;
     bPtr->columns = clist;
 }
@@ -379,7 +372,24 @@ WMGetBrowserSelectedRowInColumn(WMBrowser *bPtr, int column)
     } else {
 	return -1;
     }
-} 
+}
+
+
+void
+WMSetBrowserColumnTitle(WMBrowser *bPtr, int column, char *title)
+{
+    assert(column >= 0);
+    assert(column < bPtr->usedColumnCount);
+
+    if (bPtr->titles[column])
+	wfree(bPtr->titles[column]);
+
+    bPtr->titles[column] = wstrdup(title);
+
+    if (COLUMN_IS_VISIBLE(bPtr, column) && bPtr->flags.isTitled) {
+	drawTitleOfColumn(bPtr, column);
+    }
+}
 
 
 void
@@ -392,7 +402,7 @@ WMSetBrowserTitled(WMBrowser *bPtr, Bool flag)
 	return;
 
     columnX = 0;
-    
+
     if (!bPtr->flags.isTitled) {
 	columnY = TITLE_SPACING + bPtr->titleHeight;
 
@@ -418,23 +428,23 @@ WMSetBrowserTitled(WMBrowser *bPtr, Bool flag)
 	    WMMoveWidget(bPtr->columns[i], columnX, 0);
 	}
     }
-    
+
     bPtr->flags.isTitled = flag;
 }
 
 
-WMListItem*
-WMAddSortedBrowserItem(WMBrowser *bPtr, int column, char *text, Bool isBranch)
+void
+WMSortBrowserColumn(WMBrowser *bPtr, int column)
 {
-    WMListItem *item;
+    WMSortListItems(bPtr->columns[column]);
+}
 
-    if (column < 0 || column >= bPtr->columnCount)
-	return NULL;
-	
-    item = WMAddSortedListItem(bPtr->columns[column], text);
-    item->isBranch = isBranch;
 
-    return item;
+void
+WMSortBrowserColumnWithComparer(WMBrowser *bPtr, int column,
+				int (f)(const void*, const void*))
+{
+    WMSortListItemsWithComparer(bPtr->columns[column], f);
 }
 
 
@@ -447,7 +457,7 @@ WMInsertBrowserItem(WMBrowser *bPtr, int column, int row, char *text,
 
     if (column < 0 || column >= bPtr->columnCount)
 	return NULL;
-
+    
     item = WMInsertListItem(bPtr->columns[column], row, text);
     item->isBranch = isBranch;
 
@@ -457,33 +467,34 @@ WMInsertBrowserItem(WMBrowser *bPtr, int column, int row, char *text,
 
 
 
-static void 
-resizeBrowser(WMWidget *w, unsigned int width, unsigned int height)
+static void
+willResizeBrowser(W_ViewDelegate *self, WMView *view, 
+		  unsigned int *width, unsigned int *height)
 {
-    WMBrowser *bPtr = (WMBrowser*)w;
+    WMBrowser *bPtr = (WMBrowser*)view->self;
     int cols = bPtr->maxVisibleColumns;
     int colX, colY;
     int i;
 
-    assert(width > 0);
-    assert(height > 0);
-    
-    bPtr->columnSize.width = (width-(cols-1)*COLUMN_SPACING) / cols;
-    bPtr->columnSize.height = height;
-    
+    assert(*width > 0);
+    assert(*height > 0);
+
+    bPtr->columnSize.width = (*width-(cols-1)*COLUMN_SPACING) / cols;
+    bPtr->columnSize.height = *height;
+
     if (bPtr->flags.isTitled) {
 	colY = TITLE_SPACING + bPtr->titleHeight;
 	bPtr->columnSize.height -= colY;
     } else {
 	colY = 0;
     }
-    
+
     if (bPtr->flags.hasScroller) {
 	bPtr->columnSize.height -= SCROLLER_WIDTH + 4;
 
 	if (bPtr->scroller) {
-	    WMResizeWidget(bPtr->scroller, width-2, 1);
-	    WMMoveWidget(bPtr->scroller, 1, height-SCROLLER_WIDTH-1);
+	    WMResizeWidget(bPtr->scroller, *width-2, 1);
+	    WMMoveWidget(bPtr->scroller, 1, *height-SCROLLER_WIDTH-1);
 	}
     }
 
@@ -491,20 +502,18 @@ resizeBrowser(WMWidget *w, unsigned int width, unsigned int height)
     for (i = 0; i < bPtr->columnCount; i++) {
 	WMResizeWidget(bPtr->columns[i], bPtr->columnSize.width,
 		       bPtr->columnSize.height);
-	
+
 	WMMoveWidget(bPtr->columns[i], colX, colY);
-	
+
 	if (COLUMN_IS_VISIBLE(bPtr, i)) {
 	    colX += bPtr->columnSize.width+COLUMN_SPACING;
 	}
     }
-
-    W_ResizeView(bPtr->view, width, height);
 }
 
 
 static void
-paintItem(WMList *lPtr, int index, Drawable d, char *text, int state, 
+paintItem(WMList *lPtr, int index, Drawable d, char *text, int state,
 	  WMRect *rect)
 {
     WMView *view = W_VIEW(lPtr);
@@ -519,21 +528,21 @@ paintItem(WMList *lPtr, int index, Drawable d, char *text, int state,
     if (state & WLDSSelected)
 	XFillRectangle(scr->display, d, WMColorGC(scr->white), x, y,
 		       width, height);
-    else 
+    else
 	XClearArea(scr->display, d, x, y, width, height, False);
 
     if (text) {
-	/* Avoid overlaping... */
+        /* Avoid overlaping... */
+        WMFont *font = (state & WLDSIsBranch) ? scr->boldFont : scr->normalFont;
 	int textLen = strlen(text);
 	int widthC = (state & WLDSIsBranch) ? width-20 : width-8;
-	if (WMWidthOfString(scr->normalFont, text, textLen) > widthC) {
-	    char *textBuf = createTruncatedString(scr->normalFont,
-		                                  text, &textLen, widthC);
-            W_PaintText(view, d, scr->normalFont,  x+4, y, widthC,
+	if (WMWidthOfString(font, text, textLen) > widthC) {
+	    char *textBuf = createTruncatedString(font, text, &textLen, widthC);
+            W_PaintText(view, d, font,  x+4, y, widthC,
 		    	WALeft, WMColorGC(scr->black), False, textBuf, textLen);
-	    free(textBuf);
+	    wfree(textBuf);
 	} else {
-      	    W_PaintText(view, d, scr->normalFont,  x+4, y, widthC,
+      	    W_PaintText(view, d, font,  x+4, y, widthC,
 		    	WALeft, WMColorGC(scr->black), False, text, textLen);
 	}
     }
@@ -567,7 +576,7 @@ scrollCallback(WMWidget *scroller, void *self)
 	    scrollToColumn(bPtr, bPtr->firstVisibleColumn-1, True);
 	}
 	break;
-	
+
      case WSDecrementPage:
 	if (bPtr->firstVisibleColumn > 0) {
 	    newFirst = bPtr->firstVisibleColumn - bPtr->maxVisibleColumns;
@@ -576,13 +585,13 @@ scrollCallback(WMWidget *scroller, void *self)
 	}
 	break;
 
-	
+
      case WSIncrementLine:
 	if (LAST_VISIBLE_COLUMN < bPtr->usedColumnCount) {
 	    scrollToColumn(bPtr, bPtr->firstVisibleColumn+1, True);
 	}
 	break;
-	
+
      case WSIncrementPage:
 	if (LAST_VISIBLE_COLUMN < bPtr->usedColumnCount) {
 	    newFirst = bPtr->firstVisibleColumn + bPtr->maxVisibleColumns;
@@ -593,7 +602,7 @@ scrollCallback(WMWidget *scroller, void *self)
 	    scrollToColumn(bPtr, newFirst, True);
 	}
 	break;
-	
+
      case WSKnob:
 	{
 	    double floatValue;
@@ -631,14 +640,14 @@ setupScroller(WMBrowser *bPtr)
     int y;
 
     y = bPtr->view->size.height - SCROLLER_WIDTH - 1;
-    
+
     sPtr = WMCreateScroller(bPtr);
     WMSetScrollerAction(sPtr, scrollCallback, bPtr);
     WMMoveWidget(sPtr, 1, y);
     WMResizeWidget(sPtr, bPtr->view->size.width-2, SCROLLER_WIDTH);
-    
+
     bPtr->scroller = sPtr;
-    
+
     WMMapWidget(sPtr);
 }
 
@@ -683,6 +692,9 @@ WMSetBrowserPath(WMBrowser *bPtr, char *path)
 
     removeColumn(bPtr, 1);
 
+    WMSelectListItem(bPtr->columns[0], -1);
+    WMSetListPosition(bPtr->columns[0], 0);
+
     i = 0;
     tmp = strtok(str, bPtr->pathSeparator);
     while (tmp) {
@@ -698,25 +710,25 @@ WMSetBrowserPath(WMBrowser *bPtr, char *path)
 	listItem = WMGetListItem(bPtr->columns[i], item);
 	if (!listItem || !listItem->isBranch) {
 	    break;
-	} 
+	}
 
 	/* load next column */
 	WMAddBrowserColumn(bPtr);
 
 	loadColumn(bPtr, i+1);
-	
+
 	tmp = strtok(NULL, bPtr->pathSeparator);
 
 	i++;
     }
-    free(str);
+    wfree(str);
 
     for (i = bPtr->usedColumnCount - 1;
     	 (i > -1) && !WMGetListSelectedItem(bPtr->columns[i]);
 	 i--);
 
     bPtr->selectedColumn = i;
-    
+
     if (bPtr->columnCount < bPtr->maxVisibleColumns) {
 	int i = bPtr->maxVisibleColumns - bPtr->columnCount;
 	int curUsedColumnCount = bPtr->usedColumnCount;
@@ -740,13 +752,14 @@ WMGetBrowserPath(WMBrowser *bPtr)
 }
 
 
+
 char*
 WMGetBrowserPathToColumn(WMBrowser *bPtr, int column)
 {
     int i, size;
     char *path;
     WMListItem *item;
-    
+
     if (column >= bPtr->usedColumnCount)
 	column = bPtr->usedColumnCount-1;
 
@@ -762,7 +775,7 @@ WMGetBrowserPathToColumn(WMBrowser *bPtr, int column)
 	    break;
 	size += strlen(item->text);
     }
-    
+
     /* get the path */
     path = wmalloc(size+(column+1)*strlen(bPtr->pathSeparator)+1);
     /* ignore first / */
@@ -782,11 +795,28 @@ WMGetBrowserPathToColumn(WMBrowser *bPtr, int column)
 static void
 loadColumn(WMBrowser *bPtr, int column)
 {
-    assert(bPtr->fillColumn);
-    
+    assert(bPtr->delegate);
+    assert(bPtr->delegate->createRowsForColumn);
+
     bPtr->flags.loadingColumn = 1;
-    (*bPtr->fillColumn)(bPtr, column);
+    (*bPtr->delegate->createRowsForColumn)(bPtr->delegate, bPtr, column,
+					   bPtr->columns[column]);
     bPtr->flags.loadingColumn = 0;
+
+    if (bPtr->delegate->titleOfColumn) {
+	char *title;
+
+	title = (*bPtr->delegate->titleOfColumn)(bPtr->delegate, bPtr, column);
+
+	if (bPtr->titles[column])
+	    wfree(bPtr->titles[column]);
+
+	bPtr->titles[column] = wstrdup(title);
+
+        if (COLUMN_IS_VISIBLE(bPtr, column) && bPtr->flags.isTitled) {
+            drawTitleOfColumn(bPtr, column);
+        }
+    }
 }
 
 
@@ -798,10 +828,10 @@ paintBrowser(WMBrowser *bPtr)
     if (!bPtr->view->flags.mapped)
 	return;
 
-    W_DrawRelief(bPtr->view->screen, bPtr->view->window, 0, 
+    W_DrawRelief(bPtr->view->screen, bPtr->view->window, 0,
 		 bPtr->view->size.height-SCROLLER_WIDTH-2,
 		 bPtr->view->size.width, 22, WRSunken);
-    
+
     if (bPtr->flags.isTitled) {
 	for (i=0; i<bPtr->maxVisibleColumns; i++) {
 	    drawTitleOfColumn(bPtr, i+bPtr->firstVisibleColumn);
@@ -822,11 +852,11 @@ handleEvents(XEvent *event, void *data)
      case Expose:
 	paintBrowser(bPtr);
 	break;
-	
+
      case DestroyNotify:
 	destroyBrowser(bPtr);
 	break;
-	
+
     }
 }
 
@@ -839,13 +869,18 @@ scrollToColumn(WMBrowser *bPtr, int column, Bool updateScroller)
     int x;
     int notify = 0;
 
-    
-    if (column != bPtr->firstVisibleColumn)
+
+    if (column != bPtr->firstVisibleColumn) {
 	notify = 1;
+    }
 
     if (column < 0)
 	column = 0;
 
+    if (notify && bPtr->delegate && bPtr->delegate->willScroll) {
+	(*bPtr->delegate->willScroll)(bPtr->delegate, bPtr);
+    }
+
     x = 0;
     bPtr->firstVisibleColumn = column;
     for (i = 0; i < bPtr->columnCount; i++) {
@@ -855,7 +890,7 @@ scrollToColumn(WMBrowser *bPtr, int column, Bool updateScroller)
 	    if (!WMWidgetView(bPtr->columns[i])->flags.realized)
 		WMRealizeWidget(bPtr->columns[i]);
 	    WMMapWidget(bPtr->columns[i]);
-	    x += bPtr->columnSize.width+COLUMN_SPACING;
+	    x += bPtr->columnSize.width + COLUMN_SPACING;
 	} else {
 	    WMUnmapWidget(bPtr->columns[i]);
 	}
@@ -874,12 +909,13 @@ scrollToColumn(WMBrowser *bPtr, int column, Bool updateScroller)
 	    WMSetScrollerParameters(bPtr->scroller, 0, 1);
 	}
     }
-    
+
     if (bPtr->view->flags.mapped)
 	paintBrowser(bPtr);
 
-    if (notify)
-	WMPostNotificationName(WMBrowserDidScrollNotification, bPtr, NULL);
+    if (notify && bPtr->delegate && bPtr->delegate->didScroll) {
+	(*bPtr->delegate->didScroll)(bPtr->delegate, bPtr);
+    }
 }
 
 
@@ -905,20 +941,24 @@ listCallback(void *self, void *clientData)
         }
         assert(i<bPtr->columnCount);
 
+	bPtr->selectedColumn = i;
+
         /* columns at right must be cleared */
         removeColumn(bPtr, i+1);
         /* open directory */
         if (item->isBranch) {
             WMAddBrowserColumn(bPtr);
-            loadColumn(bPtr, bPtr->usedColumnCount-1);
         }
         if (bPtr->usedColumnCount < bPtr->maxVisibleColumns)
             i = 0;
         else
             i = bPtr->usedColumnCount-bPtr->maxVisibleColumns;
         scrollToColumn(bPtr, i, True);
+	if (item->isBranch) {
+            loadColumn(bPtr, bPtr->usedColumnCount-1);
+	}
     }
-  
+
 
     /* call callback for click */
     if (bPtr->action)
@@ -953,12 +993,12 @@ WMLoadBrowserColumnZero(WMBrowser *bPtr)
 	WMAddBrowserColumn(bPtr);
 
 	loadColumn(bPtr, 0);
-	    
+
 	/* make column 0 visible */
 	scrollToColumn(bPtr, 0, True);
 
 	bPtr->flags.loaded = 1;
-    } 
+    }
 }
 
 
@@ -993,7 +1033,7 @@ listSelectionObserver(void *observerData, WMNotification *notification)
     int column, item = (int)WMGetNotificationClientData(notification);
     WMList *lPtr = (WMList*)WMGetNotificationObject(notification);
 
-    for (column=0; column<bPtr->usedColumnCount; column++)
+    for (column = 0; column < bPtr->usedColumnCount; column++)
         if (bPtr->columns[column] == lPtr)
             break;
 
@@ -1019,7 +1059,7 @@ WMAddBrowserColumn(WMBrowser *bPtr)
     int colY;
     int index;
 
-    
+
     if (bPtr->usedColumnCount < bPtr->columnCount) {
 	return bPtr->usedColumnCount++;
     }
@@ -1039,9 +1079,9 @@ WMAddBrowserColumn(WMBrowser *bPtr)
     memcpy(clist, bPtr->columns, sizeof(WMList*)*(bPtr->columnCount-1));
     memcpy(tlist, bPtr->titles, sizeof(char*)*(bPtr->columnCount-1));
     if (bPtr->columns)
-	free(bPtr->columns);
+	wfree(bPtr->columns);
     if (bPtr->titles)
-	free(bPtr->titles);
+	wfree(bPtr->titles);
     bPtr->columns = clist;
     bPtr->titles = tlist;
 
@@ -1058,7 +1098,7 @@ WMAddBrowserColumn(WMBrowser *bPtr)
 
     WMResizeWidget(list, bPtr->columnSize.width, bPtr->columnSize.height);
     WMMoveWidget(list, (bPtr->columnSize.width+COLUMN_SPACING)*index, colY);
-    if (COLUMN_IS_VISIBLE(bPtr, index)) 
+    if (COLUMN_IS_VISIBLE(bPtr, index))
 	WMMapWidget(list);
 
     /* update the scroller */
@@ -1081,17 +1121,17 @@ destroyBrowser(WMBrowser *bPtr)
 {
     int i;
 
-    for (i=0; i<bPtr->columnCount; i++) {
+    for (i = 0; i < bPtr->columnCount; i++) {
 	if (bPtr->titles[i])
-	    free(bPtr->titles[i]);
+	    wfree(bPtr->titles[i]);
     }
-    free(bPtr->titles);
-    
-    free(bPtr->pathSeparator);
-    
+    wfree(bPtr->titles);
+
+    wfree(bPtr->pathSeparator);
+
     WMRemoveNotificationObserver(bPtr);
-    
-    free(bPtr);
+
+    wfree(bPtr);
 }
 
 
@@ -1104,7 +1144,7 @@ createTruncatedString(WMFont *font, char *text, int *textLen, int width)
     if (width >= 3*dLen) {
 	int dddLen = 3*dLen;
 	int tmpTextLen = *textLen;
-	
+
 	strcpy(textBuf, text);
 	while (tmpTextLen
 	       && (WMWidthOfString(font, textBuf, tmpTextLen)+dddLen > width))
