@@ -6,11 +6,16 @@
 #include <unistd.h>
 #include <dirent.h>
 #include <limits.h>
+#include <errno.h>
 
 #ifndef PATH_MAX
 #define PATH_MAX 1024
 #endif
 
+#ifndef FLOPPY_PATH
+#define FLOPPY_PATH "/floppy"
+#endif
+
 typedef struct W_FilePanel {
     WMWindow *win;
     
@@ -26,6 +31,10 @@ typedef struct W_FilePanel {
     WMButton *cancelButton;
     
     WMButton *homeButton;
+    WMButton *trashcanButton;
+    WMButton *createDirButton;
+    WMButton *disketteButton;
+    WMButton *unmountButton;
 
     WMView *accessoryView;
 
@@ -45,9 +54,6 @@ typedef struct W_FilePanel {
 	unsigned int canFreeFileTypes:1;
 	unsigned int fileMustExist:1;
 	unsigned int panelType:1;
-
-	/**/
-	unsigned int ignoreTextChangeNotification:1;
     } flags;
 } W_FilePanel;
 
@@ -56,17 +62,26 @@ typedef struct W_FilePanel {
 #define WP_OPEN         0
 #define WP_SAVE         1
 
-#define PWIDTH		320
+#define PWIDTH		330
 #define PHEIGHT 	360
 
 static void listDirectoryOnColumn(WMFilePanel *panel, int column, char *path);
 static void browserClick();
 static void browserDClick();
 
-static void fillColumn(WMBrowser *bPtr, int column);
+static void fillColumn(WMBrowserDelegate *self, WMBrowser *bPtr, int column,
+		       WMList *list);
+
+static void deleteFile();
+
+static void createDir();
 
 static void goHome();
 
+static void goFloppy();
+
+static void goUnmount();
+
 static void buttonClick();
 
 static char *getCurrentFileName(WMFilePanel *panel);
@@ -74,25 +89,35 @@ static char *getCurrentFileName(WMFilePanel *panel);
 static void handleEvents(XEvent *event, void *data);
 
 
+
+static WMBrowserDelegate browserDelegate = {
+    NULL, /* data */
+	fillColumn, /* createRowsForColumn */
+	NULL, /* titleOfColumn */
+	NULL, /* didScroll */
+	NULL  /* willScroll */
+};
+
+
 static int
 closestListItem(WMList *list, char *text, Bool exact)
 {
-    WMListItem *item = WMGetListItem(list, 0);
-    int i = 0;
+    WMListItem *item;
+    WMBag *items = WMGetListItems(list);
+    WMBagIterator i;
     int len = strlen(text);
 
     if (len==0)
 	return -1;
 
-    while (item) {
+    WM_ITERATE_BAG(items, item, i) {
         if (strlen(item->text) >= len &&
             ((exact && strcmp(item->text, text)==0) ||
              (!exact && strncmp(item->text, text, len)==0))) {
-	    return i;
+	    return WMBagIndexForIterator(items, i);
 	}
-	item = item->nextPtr;
-	i++;
     }
+
     return -1;
 }
 
@@ -106,9 +131,6 @@ textChangedObserver(void *observerData, WMNotification *notification)
     int col = WMGetBrowserNumberOfColumns(panel->browser) - 1;
     int i, textEvent;
 
-    if (panel->flags.ignoreTextChangeNotification)
-	return;
-
     if (!(list = WMGetBrowserListInColumn(panel->browser, col)))
 	return;
 
@@ -131,19 +153,16 @@ textChangedObserver(void *observerData, WMNotification *notification)
         if (textEvent!=WMDeleteTextEvent) {
             WMRange range;
 
-            panel->flags.ignoreTextChangeNotification = 1;
 	    WMInsertTextFieldText(panel->fileField, &item->text[textLen],
                                   textLen);
-            panel->flags.ignoreTextChangeNotification = 0;
-
-	    WMSetTextFieldCursorPosition(panel->fileField, itemTextLen);
 	    range.position = textLen;
 	    range.count = itemTextLen - textLen;
 	    WMSelectTextFieldRange(panel->fileField, range);
+            /*WMSetTextFieldCursorPosition(panel->fileField, itemTextLen);*/
 	}
     }
 
-    free(text);
+    wfree(text);
 }
 
 
@@ -198,10 +217,11 @@ makeFilePanel(WMScreen *scrPtr, char *name, char *title)
     WMSetFrameRelief(fPtr->line, WRGroove);
 
     fPtr->browser = WMCreateBrowser(fPtr->win);
-    WMSetBrowserFillColumnProc(fPtr->browser, fillColumn);
+    WMSetBrowserDelegate(fPtr->browser, &browserDelegate);
     WMSetBrowserAction(fPtr->browser, browserClick, fPtr);
     WMSetBrowserDoubleAction(fPtr->browser, browserDClick, fPtr);
     WMMoveWidget(fPtr->browser, 7, 72);
+    WMResizeWidget(fPtr->browser, PWIDTH-14,200);
     WMHangData(fPtr->browser, fPtr);
 
     fPtr->nameLabel = WMCreateLabel(fPtr->win);
@@ -220,8 +240,8 @@ makeFilePanel(WMScreen *scrPtr, char *name, char *title)
 			      fPtr->fileField);
     
     fPtr->okButton = WMCreateCommandButton(fPtr->win);
-    WMMoveWidget(fPtr->okButton, 230, 325);
-    WMResizeWidget(fPtr->okButton, 80, 28);
+    WMMoveWidget(fPtr->okButton, 245, 325);
+    WMResizeWidget(fPtr->okButton, 75, 28);
     WMSetButtonText(fPtr->okButton, "OK");
     WMSetButtonImage(fPtr->okButton, scrPtr->buttonArrow);
     WMSetButtonAltImage(fPtr->okButton, scrPtr->pushedButtonArrow);
@@ -229,19 +249,53 @@ makeFilePanel(WMScreen *scrPtr, char *name, char *title)
     WMSetButtonAction(fPtr->okButton, buttonClick, fPtr);
     
     fPtr->cancelButton = WMCreateCommandButton(fPtr->win);
-    WMMoveWidget(fPtr->cancelButton, 140, 325);
-    WMResizeWidget(fPtr->cancelButton, 80, 28);
+    WMMoveWidget(fPtr->cancelButton, 165, 325);
+    WMResizeWidget(fPtr->cancelButton, 75, 28);
     WMSetButtonText(fPtr->cancelButton, "Cancel");
     WMSetButtonAction(fPtr->cancelButton, buttonClick, fPtr);
 
+    fPtr->trashcanButton = WMCreateCommandButton(fPtr->win);
+    WMMoveWidget(fPtr->trashcanButton, 7, 325);
+    WMResizeWidget(fPtr->trashcanButton, 28, 28);
+    WMSetButtonImagePosition(fPtr->trashcanButton, WIPImageOnly);
+    WMSetButtonImage(fPtr->trashcanButton, scrPtr->trashcanIcon);
+    WMSetButtonAltImage(fPtr->trashcanButton, scrPtr->altTrashcanIcon);
+    WMSetButtonAction(fPtr->trashcanButton, deleteFile, fPtr);
+
+    fPtr->createDirButton = WMCreateCommandButton(fPtr->win);
+    WMMoveWidget(fPtr->createDirButton, 37, 325);
+    WMResizeWidget(fPtr->createDirButton, 28, 28);
+    WMSetButtonImagePosition(fPtr->createDirButton, WIPImageOnly);
+    WMSetButtonImage(fPtr->createDirButton, scrPtr->createDirIcon);
+    WMSetButtonAltImage(fPtr->createDirButton, scrPtr->altCreateDirIcon);
+    WMSetButtonAction(fPtr->createDirButton, createDir, fPtr);
+
     fPtr->homeButton = WMCreateCommandButton(fPtr->win);
-    WMMoveWidget(fPtr->homeButton, 55, 325);
+    WMMoveWidget(fPtr->homeButton, 67, 325);
     WMResizeWidget(fPtr->homeButton, 28, 28);
     WMSetButtonImagePosition(fPtr->homeButton, WIPImageOnly);
     WMSetButtonImage(fPtr->homeButton, scrPtr->homeIcon);
     WMSetButtonAltImage(fPtr->homeButton, scrPtr->altHomeIcon);
     WMSetButtonAction(fPtr->homeButton, goHome, fPtr);
 
+    fPtr->disketteButton = WMCreateCommandButton(fPtr->win);
+    WMMoveWidget(fPtr->disketteButton, 97, 325);
+    WMResizeWidget(fPtr->disketteButton, 28, 28);
+    WMSetButtonImagePosition(fPtr->disketteButton, WIPImageOnly);
+    WMSetButtonImage(fPtr->disketteButton, scrPtr->disketteIcon);
+    WMSetButtonAltImage(fPtr->disketteButton, scrPtr->altDisketteIcon);
+    WMSetButtonAction(fPtr->disketteButton, goFloppy, fPtr);
+
+    fPtr->unmountButton = WMCreateCommandButton(fPtr->win);
+    WMMoveWidget(fPtr->unmountButton, 127, 325);
+    WMResizeWidget(fPtr->unmountButton, 28, 28);
+    WMSetButtonImagePosition(fPtr->unmountButton, WIPImageOnly);
+    WMSetButtonImage(fPtr->unmountButton, scrPtr->unmountIcon);
+    WMSetButtonAltImage(fPtr->unmountButton, scrPtr->altUnmountIcon);
+    WMSetButtonAction(fPtr->unmountButton, goUnmount, fPtr);
+    WMSetButtonEnabled(fPtr->unmountButton, False);
+
+
     WMRealizeWidget(fPtr->win);
     WMMapSubwidgets(fPtr->win);
 
@@ -250,6 +304,10 @@ makeFilePanel(WMScreen *scrPtr, char *name, char *title)
 
     WMLoadBrowserColumnZero(fPtr->browser);
     
+    WMSetWindowInitialPosition(fPtr->win,
+	       (scrPtr->rootView->size.width - WMWidgetWidth(fPtr->win))/2,
+               (scrPtr->rootView->size.height - WMWidgetHeight(fPtr->win))/2);
+
     fPtr->flags.canChooseFiles = 1;
     fPtr->flags.canChooseDirectories = 1;
     fPtr->flags.autoCompletion = 1;
@@ -306,7 +364,7 @@ WMFreeFilePanel(WMFilePanel *panel)
     WMRemoveNotificationObserver(panel);
     WMUnmapWidget(panel->win);
     WMDestroyWidget(panel->win);
-    free(panel);
+    wfree(panel);
 }
 
 
@@ -349,10 +407,6 @@ WMRunModalFilePanelForDirectory(WMFilePanel *panel, WMWindow *owner,
     scr->modalView = W_VIEW(panel->win);
     WMMapWidget(panel->win);
 
-    WMMoveWidget(panel->win, 
-		 (scr->rootView->size.width - WMWidgetWidth(panel->win))/2,
-		 (scr->rootView->size.height - WMWidgetHeight(panel->win))/2);
-
     scr->modal = 1;
     while (!panel->flags.done) {	
 	WMNextEvent(scr->display, &event);
@@ -476,6 +530,19 @@ filterFileName(WMFilePanel *panel, char *file, Bool isDirectory)
 }
 
 
+#define CAST(item) (*((WMListItem**)item))
+static int
+comparer(const void *a, const void *b)
+{
+    if (CAST(a)->isBranch == CAST(b)->isBranch)
+      return (strcmp(CAST(a)->text, CAST(b)->text));
+    if (CAST(a)->isBranch)
+      return (-1);
+    return (1);
+}
+#undef CAST
+
+
 static void
 listDirectoryOnColumn(WMFilePanel *panel, int column, char *path)
 {
@@ -522,17 +589,17 @@ listDirectoryOnColumn(WMFilePanel *panel, int column, char *path)
 	    isDirectory = S_ISDIR(stat_buf.st_mode);
 	    
 	    if (filterFileName(panel, dentry->d_name, isDirectory))
-		WMAddSortedBrowserItem(bPtr, column, dentry->d_name, 
-					  isDirectory);
+		WMInsertBrowserItem(bPtr, column, -1, dentry->d_name, isDirectory);
 	}
     }
+    WMSortBrowserColumnWithComparer(bPtr, column, comparer);
 
     closedir(dir);
 }
 
 
 static void
-fillColumn(WMBrowser *bPtr, int column)
+fillColumn(WMBrowserDelegate *self, WMBrowser *bPtr, int column, WMList *list)
 {
     char *path;
     WMFilePanel *panel;
@@ -545,7 +612,7 @@ fillColumn(WMBrowser *bPtr, int column)
 
     panel = WMGetHangedData(bPtr);
     listDirectoryOnColumn(panel, column, path);
-    free(path);
+    wfree(path);
 }
 
 
@@ -561,15 +628,238 @@ browserClick(WMBrowser *bPtr, WMFilePanel *panel)
     int col = WMGetBrowserSelectedColumn(bPtr);
     WMListItem *item = WMGetBrowserSelectedItemInColumn(bPtr, col);
 
-    panel->flags.ignoreTextChangeNotification = 1;
-
     if (!item || item->isBranch)
 	WMSetTextFieldText(panel->fileField, NULL);
     else {
 	WMSetTextFieldText(panel->fileField, item->text);
     }
+}
 
-    panel->flags.ignoreTextChangeNotification = 0;
+
+static void
+showError(WMScreen *scr, WMWindow *owner, char *s, char *file)
+{
+    char *errStr;
+
+    if (file) {
+        errStr = wmalloc(strlen(file)+strlen(s));
+        sprintf(errStr, s, file);
+    } else {
+        errStr = wstrdup(s);
+    }
+    WMRunAlertPanel(scr, owner, "Error", errStr, "OK", NULL, NULL);
+    wfree(errStr);
+}
+
+
+static void
+createDir(WMButton *bPre, WMFilePanel *panel)
+{
+    char *directory_name;
+    char *directory;
+    char *file;
+    char *s;
+    WMScreen *scr = WMWidgetScreen(panel->win);
+    WMInputPanel *_panel;
+
+    _panel = WMCreateInputPanel(scr, panel->win,
+            "Create Directory", "Enter directory name", "", "OK", "Cancel");
+    scr->modalView = W_VIEW(_panel->win);
+    WMMapWidget(_panel->win);
+    scr->modal = 1;
+    while (!_panel->done || WMScreenPending(scr)) {
+        XEvent event;
+        WMNextEvent(scr->display, &event);
+        WMHandleEvent(&event);
+    }
+    scr->modal = 0;
+
+    if (_panel->result == WAPRDefault)
+        directory_name = WMGetTextFieldText(_panel->text);
+    else {
+        WMDestroyInputPanel(_panel);
+        return;
+    }
+
+    WMDestroyInputPanel(_panel);
+
+    directory = getCurrentFileName(panel);
+    {
+        char *s = strrchr(directory,'/');
+        if (s) s[1] = 0;
+    }
+
+    if (directory_name[0] == '/') {
+        directory[0] = 0;
+    } else {
+        while ((s = strstr(directory,"//"))) {
+            int i;
+            for (i = 2;s[i] == '/';i++);
+            strcpy(s, &s[i-1]);
+        }
+        if ((s = strrchr(directory, '/')) && !s[1]) s[0] = 0;
+    }
+    while ((s = strstr(directory_name,"//"))) {
+        int i;
+        for (i = 2;s[i] == '/';i++);
+        strcpy(s, &s[i-1]);
+    }
+    if ((s = strrchr(directory_name, '/')) && !s[1]) s[0] = 0;
+
+    file = wmalloc(strlen(directory_name)+strlen(directory)+1);
+    sprintf(file, "%s/%s", directory, directory_name);
+    while ((s = strstr(file,"//"))) {
+        int i;
+        for (i = 2;s[i] == '/';i++);
+        strcpy(s, &s[i-1]);
+    }
+
+    if (mkdir(file,0xfff) != 0) {
+        switch (errno) {
+            case EACCES:
+                showError(scr, panel->win, "Permission denied.", NULL);
+                break;
+            case EEXIST:
+                showError(scr, panel->win, "'%s' already existes.", file);
+                break;
+            case ENOENT:
+                showError(scr, panel->win, "Path does not exist.", NULL);
+        }
+    }
+    else WMSetFilePanelDirectory(panel, file);
+
+    wfree(directory_name);
+    wfree(directory);
+    wfree(file);
+
+}
+
+static void
+deleteFile(WMButton *bPre, WMFilePanel *panel)
+{
+    char *file;
+    char *buffer, *s;
+    struct stat filestat;
+    WMScreen *scr = WMWidgetScreen(panel->win);
+
+    file = getCurrentFileName(panel);
+
+    while ((s = strstr(file,"//"))) {
+        int i;
+        for (i = 2;s[i] == '/';i++);
+        strcpy(s, &s[i-1]);
+    }
+    if (strlen(file) > 1 && (s = strrchr(file, '/')) && !s[1]) s[0] = 0;
+
+    if (stat(file,&filestat)) {
+        switch (errno) {
+            case ENOENT:
+                showError(scr, panel->win, "'%s' does not exist.", file);
+                break;
+            case EACCES:
+                showError(scr, panel->win, "Permission denied.", NULL);
+                break;
+            case ENOMEM:
+                showError(scr, panel->win,
+                        "Insufficient memory available.", NULL);
+                break;
+            case EROFS:
+                showError(scr, panel->win,
+                        "'%s' is on a read-only filesystem.", file);
+                break;
+            default:
+                showError(scr, panel->win, "Can not delete '%s'.", file);
+        }
+        wfree(file);
+        return;
+    } else if (S_ISDIR(filestat.st_mode)) {
+        buffer = wmalloc(strlen(file)+20);
+        sprintf(buffer,"Delete directory %s ?",file);
+    } else {
+        buffer = wmalloc(strlen(file)+15);
+        sprintf(buffer,"Delete file %s ?",file);
+    }
+
+    if (!WMRunAlertPanel(WMWidgetScreen(panel->win), panel->win,
+                "Warning", buffer, "OK", "Cancel", NULL)) {
+        if (S_ISDIR(filestat.st_mode)) {
+            if (rmdir(file) != 0) {
+                switch (errno) {
+                    case EACCES:
+                        showError(scr, panel->win, "Permission denied.", NULL);
+                        break;
+                    case ENOENT:
+                        showError(scr, panel->win, "Directory '%s' does not exist.", file);
+                        break;
+                    case ENOTEMPTY:
+                        showError(scr, panel->win, "Directory '%s' is not empty.", file);
+                        break;
+                    case EBUSY:
+                        showError(scr, panel->win, "Directory '%s' is busy.", file);
+                        break;
+                    default:
+                        showError(scr, panel->win, "Can not delete '%s'.", file);
+                }
+            } else {
+                char *s = strrchr(file,'/');
+                if (s) s[0] = 0;
+                WMSetFilePanelDirectory(panel, file);
+            }
+        } else if (remove(file) != 0) {
+            switch (errno) {
+                case EISDIR:
+                    showError(scr, panel->win, "'%s' is a directory.", file);
+                    break;
+                case ENOENT:
+                    showError(scr, panel->win, "'%s' does not exist.", file);
+                    break;
+                case EACCES:
+                    showError(scr, panel->win, "Permission denied.", NULL);
+                    break;
+                case ENOMEM:
+                    showError(scr, panel->win,
+                            "Insufficient memory available.", NULL);
+                    break;
+                case EROFS:
+                    showError(scr, panel->win,
+                            "'%s' is on a read-only filesystem.", file);
+                    break;
+                default:
+                    showError(scr, panel->win, "Can not delete '%s'.", file);
+            }
+        } else {
+            char *s = strrchr(file,'/');
+            if (s) s[1] = 0;
+            WMSetFilePanelDirectory(panel, file);
+        }
+    }
+    wfree(buffer);
+    wfree(file);
+}
+
+static void
+goUnmount(WMButton *bPtr, WMFilePanel *panel)
+{
+}
+
+
+static void
+goFloppy(WMButton *bPtr, WMFilePanel *panel)
+{
+    char *file;
+    struct stat filestat;
+    WMScreen *scr = WMWidgetScreen(panel->win);
+
+    file = FLOPPY_PATH;
+    if (stat(FLOPPY_PATH,&filestat)) {
+        showError(scr, panel->win, "An error occured browsing '%s'.", file);
+        return;
+    } else if (!S_ISDIR(filestat.st_mode)) {
+        showError(scr, panel->win, "'%s' is not a directory.", file);
+        return;
+    }
+    
+    WMSetFilePanelDirectory(panel, FLOPPY_PATH);
 }
 
 
@@ -607,11 +897,16 @@ handleEvents(XEvent *event, void *data)
             WMResizeWidget(pPtr->fileField, newWidth-60-10, 24);
             WMMoveWidget(pPtr->nameLabel, 7, newHeight-(PHEIGHT-282));
             WMMoveWidget(pPtr->fileField, 60, newHeight-(PHEIGHT-278));
-            WMMoveWidget(pPtr->okButton, newWidth-(PWIDTH-230),
-                         newHeight-(PHEIGHT-325));
-            WMMoveWidget(pPtr->cancelButton, newWidth-(PWIDTH-140),
-                         newHeight-(PHEIGHT-325));
-            WMMoveWidget(pPtr->homeButton, 55, newHeight-(PHEIGHT-325));
+            WMMoveWidget(pPtr->okButton, newWidth-(PWIDTH-245),
+                    newHeight-(PHEIGHT-325));
+            WMMoveWidget(pPtr->cancelButton, newWidth-(PWIDTH-165),
+                    newHeight-(PHEIGHT-325));
+
+            WMMoveWidget(pPtr->trashcanButton, 7, newHeight-(PHEIGHT-325));
+            WMMoveWidget(pPtr->createDirButton, 37, newHeight-(PHEIGHT-325));
+            WMMoveWidget(pPtr->homeButton, 67, newHeight-(PHEIGHT-325));
+            WMMoveWidget(pPtr->disketteButton, 97, newHeight-(PHEIGHT-325));
+            WMMoveWidget(pPtr->unmountButton, 127, newHeight-(PHEIGHT-325));
 
 	    newColumnCount = (newWidth - 14) / 140;
 	    WMSetBrowserMaxVisibleColumns(pPtr->browser, newColumnCount);
@@ -640,8 +935,8 @@ getCurrentFileName(WMFilePanel *panel)
         } else
            strcpy(tmp, file);
 
-	free(file);
-	free(path);
+	wfree(file);
+	wfree(path);
 	return tmp;
     } else {
 	return path;
@@ -659,7 +954,7 @@ validOpenFile(WMFilePanel *panel)
 
     if (file[0] != '\0')
         haveFile = 1;
-    free(file);
+    wfree(file);
 
     col = WMGetBrowserSelectedColumn(panel->browser);
     item = WMGetBrowserSelectedItemInColumn(panel->browser, col);
@@ -698,10 +993,10 @@ buttonClick(WMButton *bPtr, WMFilePanel *panel)
 		WMRunAlertPanel(WMWidgetScreen(panel->win), panel->win,
 				"Error", "File does not exist.",
 				"Ok", NULL, NULL);
-		free(file);
+		wfree(file);
 		return;
 	    }
-	    free(file);
+	    wfree(file);
 	}
 	panel->flags.canceled = 0;
     } else
