@@ -31,6 +31,8 @@ typedef struct _Panel {
     WMFrame *frame;
     char *sectionName;
 
+    char *description;
+
     CallbackRec callbacks;
     WMWindow *win;
     
@@ -73,6 +75,7 @@ typedef struct _Panel {
     char capturing;		       /* shortcut capture */
     char unsaved;		       /* if there are unsaved changes */
     char dontSave;
+    char scrolledBrowser;
 } _Panel;
 
 
@@ -101,23 +104,25 @@ enum {
 
 enum {
     CpExec = 0,
-	CpArrange = 1,
-	CpHide = 2,
-	CpShow = 3,
-	CpExit = 4,
-	CpShutdown = 5,
-	CpRestart = 6,
-	CpRestartWM = 7,
-	CpSaveSession = 8,
-	CpClearSession = 9,
-	CpRefresh = 10,
-	CpInfo = 11,
-	CpLegal = 12
+	CpShExec = 1,
+	CpArrange = 2,
+	CpHide = 3,
+	CpShow = 4,
+	CpExit = 5,
+	CpShutdown = 6,
+	CpRestart = 7,
+	CpRestartWM = 8,
+	CpSaveSession = 9,
+	CpClearSession = 10,
+	CpRefresh = 11,
+	CpInfo = 12,
+	CpLegal = 13
 };
 
 enum {
     TNothing,
 	TExec,
+	TShExec,
 	TSimpleCommand,
 	TRestart,
 	TRestartWM,
@@ -130,6 +135,19 @@ enum {
 
 static void showData(_Panel *panel);
 
+static void fillBrowser(WMBrowserDelegate *self, WMBrowser *bPtr, int column,
+		       WMList *list);
+
+static void scrolledBrowser(WMBrowserDelegate *self, WMBrowser *sender);
+
+static WMBrowserDelegate browserDelegate = {
+    NULL, /* data */
+	fillBrowser, /* createRowsForColumn */
+	NULL, /* titleOfColumn */
+	scrolledBrowser, /* didScroll */
+	NULL  /* willScroll */
+};
+
 
 static Bool
 isMenu(proplist_t item)
@@ -326,6 +344,36 @@ getItemCommand(proplist_t item)
 }
 
 
+
+static proplist_t
+getSubmenuInColumn(_Panel *panel, int column)
+{
+    proplist_t parent;
+    proplist_t submenu;
+    WMList *list;
+    int r;
+
+    if (column == 0) {
+	return panel->menu;
+    }
+    if (column >= WMGetBrowserNumberOfColumns(panel->browser))
+	return NULL;
+
+    list = WMGetBrowserListInColumn(panel->browser, column - 1);
+    assert(list != NULL);
+
+    r = WMGetListSelectedItemRow(list);
+
+    parent = getSubmenuInColumn(panel, column - 1);
+
+    assert(parent != NULL);
+
+    submenu = PLGetArrayElement(parent, r + 1);
+
+    return submenu;
+}
+
+
 static void
 updateForItemType(_Panel *panel, int type)
 {
@@ -344,7 +392,8 @@ updateForItemType(_Panel *panel, int type)
     } else {
 	WMMapWidget(panel->shoF);
     }
-    if (type==TExec || type==TRestart || type==TExternalMenu) {
+    if (type==TExec || type == TShExec || type==TRestart 
+	|| type==TExternalMenu) {
 	WMMapWidget(panel->proF);
     } else {
 	WMUnmapWidget(panel->proF);
@@ -372,7 +421,9 @@ updateForItemType(_Panel *panel, int type)
     if (type == TRestart) {
 	WMSetFrameTitle(panel->proF, _("Window Manager"));
     } else if (type == TExternalMenu) {
-	WMSetFrameTitle(panel->proF, _("Program to open files"));
+	WMSetFrameTitle(panel->proF, _("Program to Open Files"));
+    } else if (type == TShExec) {
+	WMSetFrameTitle(panel->proF, _("Command to Execute"));
     } else {
 	WMSetFrameTitle(panel->proF, _("Program to Run"));
     }
@@ -387,7 +438,7 @@ getItemOfSelectedEntry(WMBrowser *bPtr)
     int i;
 
     i = WMGetBrowserSelectedColumn(bPtr);
-    menu = (proplist_t)WMGetHangedData(WMGetBrowserListInColumn(bPtr, i));
+    menu = getSubmenuInColumn((_Panel*)WMGetHangedData(bPtr), i);
 
     i = WMGetBrowserSelectedRowInColumn(bPtr, i);
     item = PLGetArrayElement(menu, i+1);
@@ -406,8 +457,8 @@ performCommand(WMWidget *w, void *data)
     int column;
     int row;
     static int cmdIndex=0;
-    WMList *list;
     char *title = NULL;
+    Bool removed = False;
 
     column = WMGetBrowserFirstVisibleColumn(panel->browser);
     if (pop == panel->cmd2P) {
@@ -417,8 +468,7 @@ performCommand(WMWidget *w, void *data)
     if (column >= WMGetBrowserNumberOfColumns(panel->browser))
 	return;
 
-    list = WMGetBrowserListInColumn(panel->browser, column);
-    menu = WMGetHangedData(list);
+    menu = getSubmenuInColumn(panel, column);
     
     row = WMGetBrowserSelectedRowInColumn(panel->browser, column);
 
@@ -454,6 +504,7 @@ performCommand(WMWidget *w, void *data)
      case CRemove:
 	if (row < 0)
 	    return;
+	panel->scrolledBrowser = 0;
 	WMRemoveBrowserItem(panel->browser, column, row);
 	menuItem = PLGetArrayElement(menu, row+1);
 	if (strcmp(getItemCommand(menuItem), "WORKSPACE_MENU")==0) {
@@ -465,6 +516,12 @@ performCommand(WMWidget *w, void *data)
 	updateForItemType(panel, TNothing);
 	panel->editedItem = NULL;
 	panel->unsaved = 1;
+
+	if (pop == panel->cmd1P && !panel->scrolledBrowser) {
+	    WMSetTextFieldText(panel->tit2T, NULL);
+	}
+
+	removed = True;
 	return;
      case CCut:
 	if (row < 0)
@@ -476,7 +533,6 @@ performCommand(WMWidget *w, void *data)
 	}
 	if (panel->itemClipboard)
 	    PLRelease(panel->itemClipboard);
-
 	WMRemoveBrowserItem(panel->browser, column, row);
 	menuItem = PLGetArrayElement(menu, row+1);
 	PLRemoveArrayElement(menu, row+1);
@@ -487,6 +543,7 @@ performCommand(WMWidget *w, void *data)
 	WMSetPopUpButtonItemEnabled(panel->cmd1P, CPaste, True);
 	WMSetPopUpButtonItemEnabled(panel->cmd2P, CPaste, True);
 	panel->unsaved = 1;
+	removed = True;
 	return;	
      case CCopy:
 	if (row < 0)
@@ -516,6 +573,21 @@ performCommand(WMWidget *w, void *data)
 	break;
     }
 
+
+    {
+	int i, j;
+	WMList *list;
+	
+	list = WMGetBrowserListInColumn(panel->browser, 0);
+	for (i = 0; i < WMGetListNumberOfRows(list); i++) {
+	    WMListItem *item =WMGetListItem(list, i);
+	    if (!item)
+		printf("%i empty\n", i);
+	    else
+		puts(item->text);
+	}
+	
+    }
     if (row>=0) row++;
     WMInsertBrowserItem(panel->browser, column, row, title, isMenu(menuItem));
     if (row<0)
@@ -523,7 +595,15 @@ performCommand(WMWidget *w, void *data)
     else
 	PLInsertArrayElement(menu, menuItem, row+1);
     free(title);
+    
+    
     panel->unsaved = 1;
+    
+    if (removed) {
+	if (pop == panel->cmd1P) {
+	    WMSetTextFieldText(panel->tit2T, NULL);
+	}
+    }
 }
 
 
@@ -533,7 +613,7 @@ browserClick(WMWidget *w, void *data)
     _Panel *panel = (_Panel*)data;
     proplist_t item;
     char *command;
-
+    
     /* stop shortcut capture */
     panel->capturing = 0;
 
@@ -583,6 +663,10 @@ browserClick(WMWidget *w, void *data)
 	WMSetTextFieldText(panel->proT, getItemParameter(item));
 	WMSetPopUpButtonSelectedItem(panel->cmdP, CpExec);
 	updateForItemType(panel, TExec);
+    } else if (strcmp(command, "SHEXEC")==0) {
+	WMSetTextFieldText(panel->proT, getItemParameter(item));
+	WMSetPopUpButtonSelectedItem(panel->cmdP, CpShExec);
+	updateForItemType(panel, TShExec);
     } else if (strcmp(command, "WORKSPACE_MENU")==0) {
 	updateForItemType(panel, TWSMenu);
     } else if (strcmp(command, "EXIT")==0) {
@@ -621,29 +705,23 @@ browserClick(WMWidget *w, void *data)
 }
 
 
+
 static void
-fillBrowserColumn(WMBrowser *bPtr, int column)
+fillBrowser(WMBrowserDelegate *self, WMBrowser *bPtr, int column, WMList *list)
 {
     _Panel *panel = (_Panel*)WMGetHangedData(bPtr);
     proplist_t menuItem;
     proplist_t menuList = NULL;
     int i;
 
-    if (column > 0) {
-	menuList = getItemOfSelectedEntry(panel->browser);
 
-	WMHangData(WMGetBrowserListInColumn(bPtr, column), menuList);
+    menuList = getSubmenuInColumn(panel, column);
+    assert(menuList != NULL);
 
-	if (column > WMGetBrowserFirstVisibleColumn(bPtr))
-	    WMSetTextFieldText(panel->tit2T, getItemTitle(menuList));
-	else
-	    WMSetTextFieldText(panel->tit1T, getItemTitle(menuList));
+    if (column > WMGetBrowserFirstVisibleColumn(bPtr)) {
+	WMSetTextFieldText(panel->tit2T, getItemTitle(menuList));
     } else {
-	menuList = panel->menu;
-
-	WMHangData(WMGetBrowserListInColumn(bPtr, column), menuList);
-
-	WMSetTextFieldText(panel->tit1T, getItemTitle(panel->menu));
+	WMSetTextFieldText(panel->tit1T, getItemTitle(menuList));
     }
 
     for (i=1; i<PLGetNumberOfElements(menuList); i++) {
@@ -667,6 +745,7 @@ changedItem(void *observerData, WMNotification *notification)
     char *command;
     char *str;
 
+
     if (!item)
 	return;
 
@@ -698,7 +777,8 @@ changedItem(void *observerData, WMNotification *notification)
 	    litem->text = str;
 	
 	    WMRedisplayWidget(list);
-	} else if (strcmp(command, "EXEC")==0 
+	} else if (strcmp(command, "EXEC")==0
+		   || strcmp(command, "SHEXEC")==0
 		   || strcmp(command, "RESTART")==0) {
 	    if (t == panel->proT) {
 		str = WMGetTextFieldText(t);
@@ -741,13 +821,8 @@ changedTitle(void *observerData, WMNotification *notification)
     column = WMGetBrowserFirstVisibleColumn(panel->browser);
     if (panel->tit2T == t)
 	column++;
-
-    list = WMGetBrowserListInColumn(panel->browser, column);
-    if (!list) {
-	return;
-    }
     
-    menu = (proplist_t)WMGetHangedData(list);
+    menu = getSubmenuInColumn(panel, column);
     if (!menu)
 	return;
 
@@ -791,6 +866,15 @@ changedCommand(WMWidget *w, void *data)
 	    updateForItemType(panel, TExec);
 	}
 	break;
+     case CpShExec:
+	if (strcmp(getItemCommand(panel->editedItem), "SHEXEC")!=0) {
+	    changeItemCommand(panel->editedItem, "SHEXEC");
+	    tmp = WMGetTextFieldText(panel->proT);
+	    changeItemParameter(panel->editedItem, tmp);
+	    free(tmp);
+	    updateForItemType(panel, TShExec);
+	}
+	break;
      case CpArrange:
 	if (strcmp(getItemCommand(panel->editedItem), "ARRANGE_ICONS")!=0) {
 	    changeItemCommand(panel->editedItem, "ARRANGE_ICONS");
@@ -814,7 +898,11 @@ changedCommand(WMWidget *w, void *data)
 	    changeItemCommand(panel->editedItem, "EXIT");
 	    updateForItemType(panel, TExit);
 	}
-	/* fall through */
+	if (WMGetButtonSelected(panel->noconfirmB))
+	    changeItemParameter(panel->editedItem, "QUICK");
+	else
+	    changeItemParameter(panel->editedItem, "");
+	break;
      case CpShutdown:
 	if (strcmp(getItemCommand(panel->editedItem), "SHUTDOWN")!=0) {
 	    changeItemCommand(panel->editedItem, "SHUTDOWN");
@@ -955,24 +1043,22 @@ captureClick(WMWidget *w, void *data)
 
 
 static void
-scrolledBrowser(void *observerData, WMNotification *notification)
+scrolledBrowser(WMBrowserDelegate *self, WMBrowser *sender)
 {
-    _Panel *panel = (_Panel*)observerData;
+    _Panel *panel = (_Panel*)self->data;
     int column;
-    WMList *list;
     proplist_t item;
 
-    column = WMGetBrowserFirstVisibleColumn(panel->browser);
-    
-    list = WMGetBrowserListInColumn(panel->browser, column);
-    item = WMGetHangedData(list);
+    column = WMGetBrowserFirstVisibleColumn(sender);
+
+    item = getSubmenuInColumn(panel, column);
     WMSetTextFieldText(panel->tit1T, getItemTitle(item));
 
-    list = WMGetBrowserListInColumn(panel->browser, column+1);
-    if (list) {
-	item = WMGetHangedData(list);
+    item = getSubmenuInColumn(panel, column + 1);
+    if (item) {
 	WMSetTextFieldText(panel->tit2T, getItemTitle(item));
     }
+    panel->scrolledBrowser = 1;
 }
 
 
@@ -1071,12 +1157,14 @@ createPanel(_Panel *p)
     WMSetBrowserTitled(panel->browser, False);
     WMResizeWidget(panel->browser, 295, 160);
     WMMoveWidget(panel->browser, 15, 65);
-    WMSetBrowserFillColumnProc(panel->browser, fillBrowserColumn);
+    
+    browserDelegate.data = panel;
+    
+    WMSetBrowserDelegate(panel->browser, &browserDelegate);
     WMHangData(panel->browser, panel);
     WMSetBrowserPathSeparator(panel->browser, "\r");
     WMSetBrowserAction(panel->browser, browserClick, panel);
-    WMAddNotificationObserver(scrolledBrowser, panel, 
-			      WMBrowserDidScrollNotification, panel->browser);
+
     /**/
 
     panel->labF = WMCreateFrame(panel->frame);
@@ -1101,12 +1189,13 @@ createPanel(_Panel *p)
     WMResizeWidget(panel->cmdP, 170, 20);
     WMMoveWidget(panel->cmdP, 10, 20);
     WMAddPopUpButtonItem(panel->cmdP, _("Run Program"));
+    WMAddPopUpButtonItem(panel->cmdP, _("Execute Shell Command"));
     WMAddPopUpButtonItem(panel->cmdP, _("Arrange Icons"));
     WMAddPopUpButtonItem(panel->cmdP, _("Hide Others"));
     WMAddPopUpButtonItem(panel->cmdP, _("Show All Windows"));
     WMAddPopUpButtonItem(panel->cmdP, _("Exit WindowMaker"));
     WMAddPopUpButtonItem(panel->cmdP, _("Exit X Session"));
-    WMAddPopUpButtonItem(panel->cmdP, _("Start window manager"));
+    WMAddPopUpButtonItem(panel->cmdP, _("Start Window Manager"));
     WMAddPopUpButtonItem(panel->cmdP, _("Restart WindowMaker"));
     WMAddPopUpButtonItem(panel->cmdP, _("Save Session"));
     WMAddPopUpButtonItem(panel->cmdP, _("Clear Session"));
@@ -1262,7 +1351,8 @@ postProcessMenu(proplist_t menu)
 	    if (strcmp(cmd, "RESTARTW")==0) {
 		changeItemCommand(item, "RESTART");
 		removeParameter(item);
-	    } else if (strcmp(cmd, "EXEC")==0 || strcmp(cmd, "OPEN_MENU")==0) {
+	    } else if (strcmp(cmd, "EXEC")==0 || strcmp(cmd, "SHEXEC")==0 || 
+		       strcmp(cmd, "OPEN_MENU")==0) {
 		/* do nothing */
 	    } else if (strcmp(cmd, "RESTART")==0 || strcmp(cmd, "SHUTDOWN")==0
 		       || strcmp(cmd, "EXIT")==0) {
@@ -1312,7 +1402,6 @@ getDefaultMenu(_Panel *panel, int *hasWSMenu)
 	}
     }
 
-    free(gspath);
     free(menuPath);
 
     if (menu) {
@@ -1326,7 +1415,7 @@ getDefaultMenu(_Panel *panel, int *hasWSMenu)
 }
 
 
-static void 
+static void
 showData(_Panel *panel)
 {
     char *gspath;
@@ -1335,12 +1424,11 @@ showData(_Panel *panel)
     int hasWSMenu = 0;
 
     gspath = wusergnusteppath();
-    
+
     menuPath = wmalloc(strlen(gspath)+32);
     strcpy(menuPath, gspath);
-    free(gspath);
     strcat(menuPath, "/Defaults/WMRootMenu");
-    
+
     menu = PLGetProplistWithPath(menuPath);
     pmenu = NULL;
 
@@ -1369,7 +1457,7 @@ showData(_Panel *panel)
 
     if (menu)
         PLRelease(menu);
-    
+
     if (panel->itemClipboard) {
 	PLRelease(panel->itemClipboard);
 	panel->itemClipboard = NULL;
@@ -1419,6 +1507,8 @@ InitMenu(WMScreen *scr, WMWindow *win)
 
     panel->sectionName = _("Applications Menu Definition");
 
+    panel->description = _("Edit the menu for launching applications.");
+
     panel->win = win;
 
     panel->callbacks.createWidgets = createPanel;
