@@ -25,20 +25,21 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
 #include <ctype.h>
 
 #include <X11/Xlib.h>
 
+
 #include <WINGs.h>
 #include <WUtil.h>
 
 #include "WPrefs.h"
 
-//#include "TexturePanel.h"
+#include "TexturePanel.h"
 
-typedef struct _TexturePanel TexturePanel;
 
-#define MAX_SECTION_PARTS 4
+#define MAX_SECTION_PARTS 5
 
 typedef struct _TexturePanel {
     WMWindow *win;
@@ -99,6 +100,7 @@ typedef struct _TexturePanel {
     WMButton *dispB;
     WMPopUpButton *arrP;
 
+    RImage *image;
     char *imageFile;
 
     /*****/
@@ -118,6 +120,9 @@ typedef struct _TexturePanel {
 
     int currentType;
 
+    
+    proplist_t pathList;
+
 } _TexturePanel;
 
 
@@ -193,6 +198,59 @@ updateGradButtons(TexturePanel *panel)
 }
 
 
+
+static void
+updateTGradImage(TexturePanel *panel)
+{
+    RImage *image, *gradient;
+    WMPixmap *pixmap;
+    RColor from;
+    RColor to;
+    WMColor *color;
+
+    if (!panel->image)
+	return;
+    
+    color = WMGetColorWellColor(panel->tcol1W);
+    from.red = WMRedComponentOfColor(color)>>8;
+    from.green = WMGreenComponentOfColor(color)>>8;
+    from.blue = WMBlueComponentOfColor(color)>>8;
+
+    color = WMGetColorWellColor(panel->tcol2W);
+    to.red = WMRedComponentOfColor(color)>>8;
+    to.green = WMGreenComponentOfColor(color)>>8;
+    to.blue = WMBlueComponentOfColor(color)>>8;
+
+    if (panel->image->width < 141 || panel->image->height < 91) {
+	image = RMakeTiledImage(panel->image, 141, 91);
+    } else {
+	image = RCloneImage(panel->image);
+    }
+
+    if (WMGetButtonSelected(panel->dirhB)) {
+	gradient = RRenderGradient(image->width, image->height, &from, &to,
+				   RHorizontalGradient);
+    } else if (WMGetButtonSelected(panel->dirvB)) {
+	gradient = RRenderGradient(image->width, image->height, &from, &to,
+				   RVerticalGradient);
+    } else {
+	gradient = RRenderGradient(image->width, image->height, &from, &to,
+				   RDiagonalGradient);
+    }
+
+    RCombineImagesWithOpaqueness(image, gradient,
+				 WMGetSliderValue(panel->topaS));
+    RDestroyImage(gradient);
+    pixmap = WMCreatePixmapFromRImage(WMWidgetScreen(panel->win),
+				      image, 128);
+
+    WMSetLabelImage(panel->imageL, pixmap);
+    WMReleasePixmap(pixmap);
+    WMResizeWidget(panel->imageL, image->width, image->height);
+    RDestroyImage(image);
+}
+
+
 static void
 updateSGradButtons(TexturePanel *panel)
 {
@@ -235,43 +293,6 @@ updateSGradButtons(TexturePanel *panel)
 }
 
 
-static void
-changeTypeCallback(WMWidget *w, void *data)
-{
-    TexturePanel *panel = (TexturePanel*)data;
-    int newType;
-    int i;
-
-    newType = WMGetPopUpButtonSelectedItem(w);
-    if (newType == panel->currentType)
-	return;
-
-    if (panel->currentType >= 0) {
-	for (i = 0; i < MAX_SECTION_PARTS; i++) {
-	    if (panel->sectionParts[panel->currentType][i] == NULL)
-		break;
-	    WMUnmapWidget(panel->sectionParts[panel->currentType][i]);
-	}
-    }
-
-    for (i = 0; i < MAX_SECTION_PARTS; i++) {
-	if (panel->sectionParts[newType][i] == NULL)
-	    break;
-	WMMapWidget(panel->sectionParts[newType][i]);
-    }
-    panel->currentType = newType;
-
-    switch (newType) {
-     case TYPE_SGRADIENT:
-	updateSGradButtons(panel);
-	break;
-     case TYPE_GRADIENT:
-	updateGradButtons(panel);
-	break;
-    }
-}
-
-
 /*********** Gradient ************/
 
 static void
@@ -367,7 +388,7 @@ static void
 sliderChangeCallback(WMWidget *w, void *data)
 {
     TexturePanel *panel = (TexturePanel*)data;
-    RHSVColor hsv, *hsvp;
+    RHSVColor hsv;
     int row, rows;
     WMListItem *item;
     RColor **colors;
@@ -491,6 +512,8 @@ gradAddCallback(WMWidget *w, void *data)
     updateGradButtons(panel);
 
     sliderChangeCallback(panel->ghueS, panel);
+
+    WMSetButtonEnabled(panel->okB, WMGetListNumberOfRows(panel->gcolL) > 1);
 }
 
 
@@ -540,6 +563,8 @@ gradDeleteCallback(WMWidget *w, void *data)
     updateGradButtons(panel);
 
     gradClickCallback(panel->gcolL, panel);
+
+    WMSetButtonEnabled(panel->okB, WMGetListNumberOfRows(panel->gcolL) > 1);
 }
 
 
@@ -552,60 +577,119 @@ colorWellObserver(void *self, WMNotification *n)
 }
 
 
+
+
+static void
+opaqChangeCallback(WMWidget *w, void *data)
+{
+    TexturePanel *panel = (TexturePanel*)data;
+
+    updateTGradImage(panel);
+}
+
 /****************** Image ******************/
 
+static void
+updateImage(TexturePanel *panel, char *path)
+{
+    WMScreen *scr = WMWidgetScreen(panel->win);
+    RImage *image;
+    WMPixmap *pixmap;
+    WMSize size;
+
+    if (path) {
+	image = RLoadImage(WMScreenRContext(scr), path, 0);
+	if (!image) {
+	    char *message;
+	    
+	    message = wstrappend(_("Could not load the selected file: "),
+				 (char*)RMessageForError(RErrorCode));
+
+	    WMRunAlertPanel(scr, panel->win, _("Error"), message,
+			    _("OK"), NULL, NULL);
+
+	    if (!panel->image)
+		WMSetButtonEnabled(panel->okB, False);
+	    return;
+	}
+
+	WMSetButtonEnabled(panel->okB, True);
+
+	if (panel->image)
+	    RDestroyImage(panel->image);
+	panel->image = image;
+    } else {
+	image = panel->image;
+    }
+    
+    if (WMGetPopUpButtonSelectedItem(panel->typeP) == TYPE_PIXMAP) {
+	if (image) {
+	    pixmap = WMCreatePixmapFromRImage(scr, image, 128);
+
+	    size = WMGetPixmapSize(pixmap);
+	    WMSetLabelImage(panel->imageL, pixmap);
+	    WMResizeWidget(panel->imageL, size.width, size.height);
+	
+	    WMReleasePixmap(pixmap);
+	}
+    } else {
+	updateTGradImage(panel);
+    }
+}
+
+
 static void
 browseImageCallback(WMWidget *w, void *data)
 {
     TexturePanel *panel = (TexturePanel*)data;
     WMOpenPanel *opanel;
     WMScreen *scr = WMWidgetScreen(w);
+    static char *ipath = NULL;
 
     opanel = WMGetOpenPanel(scr);
     WMSetFilePanelCanChooseDirectories(opanel, False);
     WMSetFilePanelCanChooseFiles(opanel, True);
 
-    if (WMRunModalFilePanelForDirectory(opanel, /*panel->win*/NULL, "/home",
+    if (!ipath)
+	ipath = wstrdup(wgethomedir());
+    
+    if (WMRunModalFilePanelForDirectory(opanel, panel->win, ipath,
 					"Open Image", NULL)) {
-	char *path;
+	char *path, *fullpath;
+	char *tmp, *tmp2;
 
-	path = WMGetFilePanelFileName(opanel);
-	if (!path)
+	tmp = WMGetFilePanelFileName(opanel);
+	if (!tmp)
 	    return;
+	fullpath = tmp;
 
-	if (!RGetImageFileFormat(path)) {
+	free(ipath);
+	ipath = fullpath;
+
+	path = wstrdup(fullpath);
+
+	tmp2 = strrchr(fullpath, '/');
+	if (tmp2)
+	    tmp2++;
+
+	tmp = wfindfileinarray(panel->pathList, tmp2);
+
+	if (tmp) {
+	    if (strcmp(fullpath, tmp)==0) {
+		free(path);
+		path = tmp2;
+	    }
+	    free(tmp);
+	}
+
+	if (!RGetImageFileFormat(fullpath)) {
 	    WMRunAlertPanel(scr, panel->win, _("Error"),
 			    _("The selected file does not contain a supported image."),
 			    _("OK"), NULL, NULL);
 	    free(path);
 	} else {
-	    RImage *image, *scaled;
-	    WMPixmap *pixmap;
-	    WMSize size;
-
-	    image = RLoadImage(WMScreenRContext(scr), path, 0);
-	    if (!image) {
-		char *message;
-
-		message = wstrappend(_("Could not load the selected file: "),
-				     (char*)RMessageForError(RErrorCode));
-
-		WMRunAlertPanel(scr, panel->win, _("Error"), message,
-				_("OK"), NULL, NULL);
-		free(message);
-		free(path);
-		return;
-	    }
-
-	    pixmap = WMCreatePixmapFromRImage(scr, image, 128);
-	    RDestroyImage(image);
-
-	    size = WMGetPixmapSize(pixmap);
-	    WMSetLabelImage(panel->imageL, pixmap);
-	    WMResizeWidget(panel->imageL, size.width, size.height);
-
-	    WMReleasePixmap(pixmap);
-
+	    updateImage(panel, fullpath);
+	    free(panel->imageFile);
 	    panel->imageFile = path;
 
 	    WMSetTextFieldText(panel->imageT, path);
@@ -628,6 +712,54 @@ buttonCallback(WMWidget *w, void *data)
 }
 
 
+
+static void
+changeTypeCallback(WMWidget *w, void *data)
+{
+    TexturePanel *panel = (TexturePanel*)data;
+    int newType;
+    int i;
+
+    newType = WMGetPopUpButtonSelectedItem(w);
+    if (newType == panel->currentType)
+	return;
+
+    if (panel->currentType >= 0) {
+	for (i = 0; i < MAX_SECTION_PARTS; i++) {
+	    if (panel->sectionParts[panel->currentType][i] == NULL)
+		break;
+	    WMUnmapWidget(panel->sectionParts[panel->currentType][i]);
+	}
+    }
+
+    for (i = 0; i < MAX_SECTION_PARTS; i++) {
+	if (panel->sectionParts[newType][i] == NULL)
+	    break;
+	WMMapWidget(panel->sectionParts[newType][i]);
+    }
+    panel->currentType = newType;
+
+    switch (newType) {
+     case TYPE_SGRADIENT:
+	updateSGradButtons(panel);
+	WMSetButtonEnabled(panel->okB, True);
+	break;
+     case TYPE_GRADIENT:
+	updateGradButtons(panel);
+	WMSetButtonEnabled(panel->okB, WMGetListNumberOfRows(panel->gcolL)>1);
+	break;
+     case TYPE_TGRADIENT:
+     case TYPE_PIXMAP:
+	updateImage(panel, NULL);
+	WMSetButtonEnabled(panel->okB, panel->image!=NULL);
+	break;
+     default:
+	WMSetButtonEnabled(panel->okB, True);
+	break;
+    }
+}
+
+
 /*
  *--------------------------------------------------------------------------
  * Public functions
@@ -643,6 +775,12 @@ DestroyTexturePanel(TexturePanel *panel)
 void
 ShowTexturePanel(TexturePanel *panel)
 {
+    Display *dpy = WMScreenDisplay(WMWidgetScreen(panel->win));
+    Screen *scr = DefaultScreenOfDisplay(dpy);
+
+    WMSetWindowInitialPosition(panel->win,
+			 (WidthOfScreen(scr)-WMWidgetWidth(panel->win))/2,
+			 (HeightOfScreen(scr)-WMWidgetHeight(panel->win))/2);
     WMMapWidget(panel->win);
 }
 
@@ -699,10 +837,10 @@ SetTexturePanelTexture(TexturePanel *panel, char *name, proplist_t texture)
 
 	p = PLGetArrayElement(texture, 1);
 	if (!p) {
-	    goto bad_texture;
+	    str = "black";
+	} else {
+	    str = PLGetString(p);
 	}
-
-	str = PLGetString(p);
 	color = WMCreateNamedColor(scr, str, False);
 
 	WMSetColorWellColor(panel->defcW, color);
@@ -717,9 +855,10 @@ SetTexturePanelTexture(TexturePanel *panel, char *name, proplist_t texture)
 
 	p = PLGetArrayElement(texture, 1);
 	if (!p) {
-	    goto bad_texture;
+	    str = "black";
+	} else {
+	    str = PLGetString(p);
 	}
-	str = PLGetString(p);
 	color = WMCreateNamedColor(scr, str, False);
 
 	WMSetColorWellColor(panel->tcol1W, color);
@@ -728,9 +867,10 @@ SetTexturePanelTexture(TexturePanel *panel, char *name, proplist_t texture)
 
 	p = PLGetArrayElement(texture, 2);
 	if (!p) {
-	    goto bad_texture;
+	    str = "black";
+	} else {
+	    str = PLGetString(p);
 	}
-	str = PLGetString(p);
 	color = WMCreateNamedColor(scr, str, False);
 
 	WMSetColorWellColor(panel->tcol2W, color);
@@ -742,11 +882,62 @@ SetTexturePanelTexture(TexturePanel *panel, char *name, proplist_t texture)
     } else if (strcasecmp(type, "thgradient")==0
 	       || strcasecmp(type, "tvgradient")==0
 	       || strcasecmp(type, "tdgradient")==0) {
+	int i;
 
 	WMSetPopUpButtonSelectedItem(panel->typeP, TYPE_TGRADIENT);
 
-	/****** TODO: setador de textura apartir desse */
 	gradient = type[1];
+
+	WMSetTextFieldText(panel->imageT,
+			   PLGetString(PLGetArrayElement(texture, 1)));
+	if (panel->imageFile)
+	    free(panel->imageFile);
+	panel->imageFile = wstrdup(PLGetString(PLGetArrayElement(texture, 1)));
+
+
+	i = 180;
+	sscanf(PLGetString(PLGetArrayElement(texture, 2)), "%i", &i);
+	WMSetSliderValue(panel->topaS, i);
+
+	p = PLGetArrayElement(texture, 3);
+	if (!p) {
+	    str = "black";
+	} else {
+	    str = PLGetString(p);
+	}
+	color = WMCreateNamedColor(scr, str, False);
+
+	WMSetColorWellColor(panel->tcol1W, color);
+
+	WMReleaseColor(color);
+
+	p = PLGetArrayElement(texture, 4);
+	if (!p) {
+	    str = "black";
+	} else {
+	    str = PLGetString(p);
+	}
+	color = WMCreateNamedColor(scr, str, False);
+
+	WMSetColorWellColor(panel->tcol2W, color);
+
+	WMReleaseColor(color);
+	
+	WMSetTextFieldText(panel->imageT,
+			   PLGetString(PLGetArrayElement(texture, 1)));
+	
+    if (panel->imageFile)
+        free(panel->imageFile);
+    if ((panel->imageFile = wfindfileinarray(panel->pathList,
+                PLGetString(PLGetArrayElement(texture, 1)))) != NULL) {
+
+        panel->image = RLoadImage(WMScreenRContext(scr), panel->imageFile, 0);
+        updateTGradImage(panel);
+
+        updateSGradButtons(panel);
+    } else wwarning("could not load file '%s': %s", panel->imageFile,
+		     RMessageForError(RErrorCode));
+
     /*...............................................*/
     } else if (strcasecmp(type, "mhgradient")==0
 	       || strcasecmp(type, "mvgradient")==0
@@ -763,10 +954,10 @@ SetTexturePanelTexture(TexturePanel *panel, char *name, proplist_t texture)
 
 	p = PLGetArrayElement(texture, 1);
 	if (!p) {
-	    goto bad_texture;
+	    str = "black";
+	} else {
+	    str = PLGetString(p);
 	}
-
-	str = PLGetString(p);
 	color = WMCreateNamedColor(scr, str, False);
 
 	WMSetColorWellColor(panel->defcW, color);
@@ -779,9 +970,10 @@ SetTexturePanelTexture(TexturePanel *panel, char *name, proplist_t texture)
 
 	    p = PLGetArrayElement(texture, i);
 	    if (!p) {
-		goto bad_texture;
+		str = "black";
+	    } else {
+		str = PLGetString(p);
 	    }
-	    str = PLGetString(p);
 
 	    XParseColor(WMScreenDisplay(scr), WMScreenRContext(scr)->cmap,
 			str, &xcolor);
@@ -817,10 +1009,26 @@ SetTexturePanelTexture(TexturePanel *panel, char *name, proplist_t texture)
 	 case 'M':
 	    WMSetPopUpButtonSelectedItem(panel->arrP, PTYPE_MAXIMIZE);
 	    break;
+	 default:
 	 case 'T':
 	    WMSetPopUpButtonSelectedItem(panel->arrP, PTYPE_TILE);
 	    break;
 	}
+
+	WMSetTextFieldText(panel->imageT,
+			   PLGetString(PLGetArrayElement(texture, 1)));
+	
+	if (panel->imageFile)
+	    free(panel->imageFile);
+	panel->imageFile = wfindfileinarray(panel->pathList,
+			    PLGetString(PLGetArrayElement(texture, 1)));
+
+	color = WMCreateNamedColor(scr,
+			   PLGetString(PLGetArrayElement(texture, 2)), False);
+	WMSetColorWellColor(panel->defcW, color);
+	WMReleaseColor(color);
+
+	updateImage(panel, panel->imageFile);
     }
 
     changeTypeCallback(panel->typeP, panel);
@@ -835,6 +1043,7 @@ SetTexturePanelTexture(TexturePanel *panel, char *name, proplist_t texture)
 	 case 'V':
 	    WMPerformButtonClick(panel->dirvB);
 	    break;
+	 default:
 	 case 'D':
 	    WMPerformButtonClick(panel->dirdB);
 	    break;
@@ -903,7 +1112,7 @@ GetTexturePanelTexture(TexturePanel *panel)
 					   PLMakeString(str), NULL);
 	    break;
 	 case PTYPE_TILE:
-	    prop = PLMakeArrayFromElements(PLMakeString("tpixmap"), 
+	    prop = PLMakeArrayFromElements(PLMakeString("tpixmap"),
 					   PLMakeString(panel->imageFile),
 					   PLMakeString(str), NULL);
 	    break;
@@ -911,6 +1120,39 @@ GetTexturePanelTexture(TexturePanel *panel)
 	free(str);
 	break;
 
+     case TYPE_TGRADIENT:
+	color = WMGetColorWellColor(panel->tcol1W);
+	str = WMGetColorRGBDescription(color);
+
+	color = WMGetColorWellColor(panel->tcol2W);
+	str2 = WMGetColorRGBDescription(color);
+
+	sprintf(buff, "%i", WMGetSliderValue(panel->topaS));
+
+	if (WMGetButtonSelected(panel->dirdB)) {
+	    prop = PLMakeArrayFromElements(PLMakeString("tdgradient"),
+					   PLMakeString(panel->imageFile),
+					   PLMakeString(buff),
+					   PLMakeString(str),
+					   PLMakeString(str2), NULL);
+	} else 	if (WMGetButtonSelected(panel->dirvB)) {
+	    prop = PLMakeArrayFromElements(PLMakeString("tvgradient"),
+					   PLMakeString(panel->imageFile),
+					   PLMakeString(buff),
+					   PLMakeString(str),
+					   PLMakeString(str2), NULL);
+	} else {
+	    prop = PLMakeArrayFromElements(PLMakeString("thgradient"),
+					   PLMakeString(panel->imageFile),
+					   PLMakeString(buff),
+					   PLMakeString(str),
+					   PLMakeString(str2), NULL);
+	}
+	free(str);
+	free(str2);
+	break;
+
+
      case TYPE_SGRADIENT:
 	color = WMGetColorWellColor(panel->tcol1W);
 	str = WMGetColorRGBDescription(color);
@@ -972,10 +1214,17 @@ GetTexturePanelTexture(TexturePanel *panel)
 
 
 
+void
+SetTexturePanelPixmapPath(TexturePanel *panel, proplist_t array)
+{
+    panel->pathList = array;
+}
+
+
 
 TexturePanel*
 CreateTexturePanel(WMWindow *keyWindow)
-//CreateTexturePanel(WMScreen *scr)
+/*CreateTexturePanel(WMScreen *scr)*/
 {
     TexturePanel *panel;
     WMScreen *scr = WMWidgetScreen(keyWindow);
@@ -997,6 +1246,7 @@ CreateTexturePanel(WMWindow *keyWindow)
 
     WMResizeWidget(panel->win, 325, 423);
     WMSetWindowTitle(panel->win, _("Texture Panel"));
+    WMSetWindowCloseAction(panel->win, buttonCallback, panel);
 
 
     /* texture name */
@@ -1017,7 +1267,7 @@ CreateTexturePanel(WMWindow *keyWindow)
     WMMoveWidget(panel->typeP, 15, 65);
     WMAddPopUpButtonItem(panel->typeP, _("Solid Color"));
     WMAddPopUpButtonItem(panel->typeP, _("Gradient Texture"));
-    WMAddPopUpButtonItem(panel->typeP, _("Simle Gradient Texture"));
+    WMAddPopUpButtonItem(panel->typeP, _("Simple Gradient Texture"));
     WMAddPopUpButtonItem(panel->typeP, _("Textured Gradient"));
     WMAddPopUpButtonItem(panel->typeP, _("Image Texture"));
     WMSetPopUpButtonSelectedItem(panel->typeP, 0);
@@ -1188,6 +1438,9 @@ CreateTexturePanel(WMWindow *keyWindow)
     WMMoveWidget(panel->topaS, 15, 20);
     WMSetSliderMaxValue(panel->topaS, 255);
     WMSetSliderValue(panel->topaS, 200);
+    WMSetSliderContinuous(panel->topaS, False);
+    WMSetSliderAction(panel->topaS, opaqChangeCallback, panel);
+    
     WMMapSubwidgets(panel->topaF);
 
     {
@@ -1244,10 +1497,11 @@ CreateTexturePanel(WMWindow *keyWindow)
     WMSetButtonText(panel->browB, _("Browse..."));
     WMSetButtonAction(panel->browB, browseImageCallback, panel);
 
-    panel->dispB = WMCreateCommandButton(panel->imageF);
+/*    panel->dispB = WMCreateCommandButton(panel->imageF);
     WMResizeWidget(panel->dispB, 90, 24);
     WMMoveWidget(panel->dispB, 190, 80);
     WMSetButtonText(panel->dispB, _("Show"));
+ */
 
     panel->arrP = WMCreatePopUpButton(panel->imageF);
     WMResizeWidget(panel->arrP, 90, 20);
@@ -1337,7 +1591,7 @@ testOKButton(WMWidget *self, void *data)
     RImage *image;
     
     TexturePanel *panel = (TexturePanel*)data;
-//    test = GetTexturePanelTextureString(panel);
+    /* test = GetTexturePanelTextureString(panel); */
     
     wwarning(test);
     
@@ -1347,7 +1601,7 @@ testOKButton(WMWidget *self, void *data)
     XMapRaised(dpy, win);
     XFlush(dpy);
     
-  //  image = RenderTexturePanelTexture(panel, 250, 250);
+    /* image = RenderTexturePanelTexture(panel, 250, 250); */
     
     RConvertImage(WMScreenRContext(WMWidgetScreen(panel->okB)), image, &pix);
     
