@@ -2,6 +2,7 @@
  *  Window Maker window manager
  * 
  *  Copyright (c) 1997, 1998 Alfredo K. Kojima
+ *  Copyright (c) 1998       Dan Pascu
  * 
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -29,6 +30,9 @@
 #include "WindowMaker.h"
 #include "menu.h"
 #include "window.h"
+#ifdef USER_MENU
+#include "usermenu.h"
+#endif /* USER_MENU */
 #include "icon.h"
 #include "appicon.h"
 #include "application.h"
@@ -107,10 +111,10 @@ findDockIconFor(WDock *dock, Window main_window)
 {
     WAppIcon *aicon = NULL;
 
-    aicon = wDockFindIconFor(dock, main_window);
+    aicon = wDockFindIconForWindow(dock, main_window);
     if (!aicon) {
         wDockTrackWindowLaunch(dock, main_window);
-        aicon = wDockFindIconFor(dock, main_window);
+        aicon = wDockFindIconForWindow(dock, main_window);
     }
     return aicon;
 }
@@ -291,6 +295,9 @@ wApplicationCreate(WScreen *scr, Window main_window)
 	leader->main_window = main_window;
     }
     wapp->menu = wAppMenuGet(scr, main_window);
+#ifdef USER_MENU
+    if (!wapp->menu) wapp->menu = wUserMenuGet(scr, wapp->main_window_desc);
+#endif /* USER_MENU */
 
 
     /*
@@ -353,9 +360,6 @@ wApplicationCreate(WScreen *scr, Window main_window)
     
     if (wapp->app_icon) {
         wapp->app_icon->main_window = main_window;
-#ifdef WMSOUND
-        wSoundServerGrab(wapp->app_icon->wm_class, main_window);
-#endif
     }
 
 #ifndef REDUCE_APPICONS
@@ -372,7 +376,7 @@ wApplicationCreate(WScreen *scr, Window main_window)
 
         if (clip && clip->attract_icons && wDockFindFreeSlot(clip, &x, &y)) {
             wapp->app_icon->attracted = 1;
-            if (!clip->keep_attracted && !wapp->app_icon->icon->shadowed) {
+            if (!wapp->app_icon->icon->shadowed) {
                 wapp->app_icon->icon->shadowed = 1;
                 wapp->app_icon->icon->force_paint = 1;
                 /* We don't do an wAppIconPaint() here because it's in
@@ -409,6 +413,7 @@ wApplicationCreate(WScreen *scr, Window main_window)
 	scr->wapp_list->prev = wapp;
     scr->wapp_list = wapp;
 
+    
 #ifdef WMSOUND
     wSoundPlay(WMSOUND_APPSTART);
 #endif
@@ -432,7 +437,7 @@ wApplicationDestroy(WApplication *wapp)
 
     if (!wapp)
       return;
-    
+  
     wapp->refcount--;
     if (wapp->refcount>0)
       return;
@@ -458,9 +463,7 @@ wApplicationDestroy(WApplication *wapp)
     XDeleteContext(dpy, wapp->main_window, wAppWinContext);
     wAppMenuDestroy(wapp->menu);
     if (wapp->app_icon) {
-        if (wapp->app_icon->docked
-            && (!wapp->app_icon->attracted || 
-		wapp->app_icon->dock->keep_attracted)) {
+        if (wapp->app_icon->docked && !wapp->app_icon->attracted) {
 #ifdef REDUCE_APPICONS
 	    if (napps == 0) {
 #endif
