@@ -3,6 +3,7 @@
  *  Window Maker window manager
  * 
  *  Copyright (c) 1997, 1998 Alfredo K. Kojima
+ *  Copyright (c) 1999       Dan Pascu
  * 
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -35,6 +36,12 @@
 #include <dirent.h>
 #include <limits.h>
 
+#include <signal.h>
+#ifdef __FreeBSD__
+#include <sys/signal.h>
+#endif
+
+
 #ifndef PATH_MAX
 #define PATH_MAX DEFAULT_PATH_MAX
 #endif
@@ -177,7 +184,10 @@ typedef struct IconPanel {
     
     WMList *dirList;
     WMList *iconList;
+    WMFont *normalfont;
     
+    WMButton *previewButton;
+
     WMLabel *iconView;
     
     WMLabel *fileLabel;
@@ -190,6 +200,7 @@ typedef struct IconPanel {
 #endif
     short done;
     short result;
+    short preview;
 } IconPanel;
 
 
@@ -201,6 +212,9 @@ listPixmaps(WScreen *scr, WMList *lPtr, char *path)
     DIR *dir;
     char pbuf[PATH_MAX+16];
     char *apath;
+    IconPanel *panel = WMGetHangedData(lPtr);
+
+    panel->preview = False;
     
     apath = wexpandpath(path);
     dir = opendir(apath);
@@ -236,12 +250,14 @@ listPixmaps(WScreen *scr, WMList *lPtr, char *path)
 	
 	if (statb.st_mode & (S_IRUSR|S_IRGRP|S_IROTH)
 	    && statb.st_mode & (S_IFREG|S_IFLNK)) {
-	    WMAddSortedListItem(lPtr, dentry->d_name);
+	    WMAddListItem(lPtr, dentry->d_name);
 	}
     }
+    WMSortListItems(lPtr);
 
     closedir(dir);
     free(apath);
+    panel->preview = True;
 }
 
 
@@ -259,23 +275,15 @@ setViewedImage(IconPanel *panel, char *file)
     pixmap = WMCreateBlendedPixmapFromFile(WMWidgetScreen(panel->win),
 					   file, &color);
     if (!pixmap) {
-	char *msg;
-	char *tmp;
-
 	WMSetButtonEnabled(panel->okButton, False);
 	    
-	tmp = _("Could not load image file ");
-	msg = wmalloc(strlen(tmp)+strlen(file)+6);
-	strcpy(msg, tmp);
-	strcat(msg, file);
-	
-	wMessageDialog(panel->scr, _("Error"), msg, _("OK"), NULL, NULL);
-	free(msg);
+	WMSetLabelText(panel->iconView, _("Could not load image file "));
 	
 	WMSetLabelImage(panel->iconView, NULL);
     } else {
 	WMSetButtonEnabled(panel->okButton, True);
-	
+
+	WMSetLabelText(panel->iconView, NULL);
 	WMSetLabelImage(panel->iconView, pixmap);
 	WMReleasePixmap(pixmap);
     }
@@ -337,7 +345,7 @@ listIconPaths(WMList *lPtr)
 	/* do not sort, because the order implies the order of
 	 * directories searched */
 	if (access(tmp, X_OK)==0)
-	    WMAddListItem(lPtr, tmp);
+	    WMAddListItem(lPtr, path);
 	free(tmp);
     } while ((path=strtok(NULL, ":"))!=NULL);
 
@@ -345,6 +353,94 @@ listIconPaths(WMList *lPtr)
 }
 
 
+void
+drawIconProc(WMList *lPtr, int index, Drawable d, char *text,
+	     int state, WMRect *rect) 
+{
+    IconPanel *panel = WMGetHangedData(lPtr);
+    GC gc = panel->scr->draw_gc;
+    GC copygc = panel->scr->copy_gc;
+    char *buffer, *dirfile;
+    WMPixmap *pixmap;
+    WMColor *blackcolor;
+    WMColor *whitecolor;
+    WMSize size;
+    WMScreen *wmscr=WMWidgetScreen(panel->win);
+    int width;
+
+    if(!panel->preview) return;
+
+    width = rect->size.width;
+
+    blackcolor = WMBlackColor(wmscr);
+    whitecolor = WMWhiteColor(wmscr);
+
+    dirfile = wexpandpath(WMGetListSelectedItem(panel->dirList)->text);
+    buffer = wmalloc(strlen(dirfile)+strlen(text)+4);
+    sprintf(buffer, "%s/%s", dirfile, text);
+    free(dirfile);
+
+    pixmap = WMCreatePixmapFromFile(WMWidgetScreen(panel->win), buffer);
+    free(buffer);
+    if (!pixmap) {
+        WMRemoveListItem(lPtr, index);
+        return;
+    }
+
+    XClearArea(dpy, d, rect->pos.x, rect->pos.y, width, rect->size.height, 
+	       False);
+    XSetClipMask(dpy, gc, None);
+    /*
+    XDrawRectangle(dpy, d, WMColorGC(whitecolor), rect->pos.x + 5, 
+		   rect->pos.y +5, width - 10, 54);
+     */
+    XDrawLine(dpy, d, WMColorGC(whitecolor), rect->pos.x,
+	      rect->pos.y+rect->size.height-1, rect->pos.x+width, 
+	      rect->pos.y+rect->size.height-1);
+   
+    
+    if (state&WLDSSelected) {
+        XFillRectangle(dpy, d, WMColorGC(whitecolor), rect->pos.x,
+		       rect->pos.y, width, rect->size.height);
+    }
+
+    size = WMGetPixmapSize(pixmap);
+
+    XSetClipMask(dpy, copygc, WMGetPixmapMaskXID(pixmap));
+    XSetClipOrigin(dpy, copygc, rect->pos.x + (width-size.width)/2,
+		   rect->pos.y+2);
+    XCopyArea(dpy, WMGetPixmapXID(pixmap), d, copygc, 0, 0,
+	      size.width>100?100:size.width, size.height>64?64:size.height, 
+	      rect->pos.x + (width-size.width)/2, rect->pos.y+2);
+
+    {
+        int i,j;
+	int fheight = WMFontHeight(panel->normalfont);
+	int tlen = strlen(text);
+	int twidth = WMWidthOfString(panel->normalfont, text, tlen);
+	int ofx, ofy;
+	
+	ofx = rect->pos.x + (width - twidth)/2;
+	ofy = rect->pos.y + 64 - fheight;
+	
+        for(i=-1;i<2;i++)
+	    for(j=-1;j<2;j++)
+		WMDrawString(wmscr, d, WMColorGC(whitecolor), 
+			     panel->normalfont, ofx+i, ofy+j,
+			     text, tlen);
+
+	WMDrawString(wmscr, d, WMColorGC(blackcolor), panel->normalfont, 
+		     ofx, ofy, text, tlen);
+    }
+
+    WMReleasePixmap(pixmap);
+    /* I hope it is better to do not use cache / on my box it is fast nuff */
+    XFlush(dpy);
+
+    WMReleaseColor(blackcolor);
+    WMReleaseColor(whitecolor);
+}
+
 
 static void
 buttonCallback(void *self, void *clientData)
@@ -359,6 +455,14 @@ buttonCallback(void *self, void *clientData)
     } else if (bPtr==panel->cancelButton) {
 	panel->done = True;
 	panel->result = False;
+    } else if (bPtr==panel->previewButton) {
+        /**** Previewer ****/
+        WMSetButtonEnabled(bPtr, False);
+        WMSetListUserDrawItemHeight(panel->iconList, 68);
+        WMSetListUserDrawProc(panel->iconList, drawIconProc);
+        WMRedisplayWidget(panel->iconList);
+        /* for draw proc to access screen/gc */
+        /*** end preview ***/
     }
 #if 0
     else if (bPtr==panel->chooseButton) {
@@ -397,6 +501,7 @@ wIconChooserDialog(WScreen *scr, char **file, char *instance, char *class)
     WMResizeWidget(panel->win, 450, 280);
     
     boldFont = WMBoldSystemFontOfSize(scr->wmscreen, 12);
+    panel->normalfont = WMSystemFontOfSize(WMWidgetScreen(panel->win), 12);
     
     panel->dirLabel = WMCreateLabel(panel->win);
     WMResizeWidget(panel->dirLabel, 200, 20);
@@ -437,13 +542,22 @@ wIconChooserDialog(WScreen *scr, char **file, char *instance, char *class)
     WMResizeWidget(panel->iconList, 140, 170);
     WMMoveWidget(panel->iconList, 215, 30); 
     WMSetListAction(panel->iconList, listCallback, panel);
-   
+
+    WMHangData(panel->iconList,panel);
+
+    panel->previewButton = WMCreateCommandButton(panel->win);
+    WMResizeWidget(panel->previewButton, 75, 26);
+    WMMoveWidget(panel->previewButton, 365, 130);
+    WMSetButtonText(panel->previewButton, _("Preview"));
+    WMSetButtonAction(panel->previewButton, buttonCallback, panel);
+
     panel->iconView = WMCreateLabel(panel->win);
     WMResizeWidget(panel->iconView, 75, 75);
-    WMMoveWidget(panel->iconView, 365, 60);
-    WMSetLabelImagePosition(panel->iconView, WIPImageOnly);
+    WMMoveWidget(panel->iconView, 365, 40);
+    WMSetLabelImagePosition(panel->iconView, WIPOverlaps);
     WMSetLabelRelief(panel->iconView, WRSunken);
-    
+    WMSetLabelTextAlignment(panel->iconView, WACenter);
+
     panel->fileLabel = WMCreateLabel(panel->win);
     WMResizeWidget(panel->fileLabel, 80, 20);
     WMMoveWidget(panel->fileLabel, 10, 210);
@@ -483,13 +597,13 @@ wIconChooserDialog(WScreen *scr, char **file, char *instance, char *class)
     {
 	char *tmp;
 
-	tmp = malloc((instance ? strlen(instance) : 0)
+	tmp = wmalloc((instance ? strlen(instance) : 0)
 		     + (class ? strlen(class) : 0) + 32);
 
 	if (tmp && (instance || class))
 	    sprintf(tmp, "%s [%s.%s]", _("Icon Chooser"), instance, class);
 	else
-	    tmp = _("Icon Chooser");
+	    strcpy(tmp, _("Icon Chooser"));
 
 	wwin = wManageInternalWindow(scr, parent, None, tmp,
 				     (scr->scr_width - 450)/2,
@@ -536,6 +650,8 @@ wIconChooserDialog(WScreen *scr, char **file, char *instance, char *class)
 	*file = NULL;
     }
 
+    WMReleaseFont(panel->normalfont);
+
     WMUnmapWidget(panel->win);
 
     WMDestroyWidget(panel->win);    
@@ -588,8 +704,8 @@ typedef struct {
 
 
 #define COPYRIGHT_TEXT  \
-     "Copyright \xa9 1997~1999 Alfredo K. Kojima <kojima@windowmaker.org>\n"\
-     "Copyright \xa9 1998,1999 Dan Pascu <dan@windowmaker.org>"
+     "Copyright \xa9 1997~2000 Alfredo K. Kojima <kojima@windowmaker.org>\n"\
+     "Copyright \xa9 1998~2000 Dan Pascu <dan@windowmaker.org>"
 
  
 
@@ -848,14 +964,12 @@ handleLogoPush(XEvent *event, void *data)
 "             .....XmnbN:q&Bo....             ",
 "                 ............                "};
     static char *msgs[] = {
-	"Sloppy focus is a *?#@",
-	    "Repent! Sloppy focus users will burn in hell!!!",
 	    "Have a nice day!"
     };
 
 
     clicks++;
-    if (!panel->timer && !broken && clicks > 2) {
+    if (!panel->timer && !broken && clicks > 0) {
 	char *file;
 	char *path;
 
@@ -961,8 +1075,10 @@ wShowInfoPanel(WScreen *scr)
 
 
     if (thePanel) {
-	wRaiseFrame(thePanel->wwin->frame->core);
-	wSetFocusTo(scr, thePanel->wwin);
+	if (thePanel->scr == scr) {
+	    wRaiseFrame(thePanel->wwin->frame->core);
+	    wSetFocusTo(scr, thePanel->wwin);
+	}
 	return;
     }
     
@@ -1040,7 +1156,7 @@ wShowInfoPanel(WScreen *scr)
     WMSetLabelTextAlignment(panel->copyrL, WALeft);
     WMSetLabelText(panel->copyrL, COPYRIGHT_TEXT);
     /* we want the (c) character in the helvetica font */
-    font = WMCreateFontInDefaultEncoding(scr->wmscreen, HELVETICA10_FONT);
+    font = WMCreateNormalFont(scr->wmscreen, HELVETICA10_FONT);
     if (font) {
 	WMSetLabelFont(panel->copyrL, font);
     }
@@ -1133,6 +1249,9 @@ wShowInfoPanel(WScreen *scr)
 
     WSETUFLAG(wwin, no_closable, 0);
     WSETUFLAG(wwin, no_close_button, 0);
+#ifdef XKB_BUTTON_HINT
+    wFrameWindowHideButton(wwin->frame, WFF_LANGUAGE_BUTTON);
+#endif
     wWindowUpdateButtonImages(wwin);
     wFrameWindowShowButton(wwin->frame, WFF_RIGHT_BUTTON);
     wwin->frame->on_click_right = destroyInfoPanel;
@@ -1142,7 +1261,7 @@ wShowInfoPanel(WScreen *scr)
     panel->wwin = wwin;
 
     thePanel = panel;
-#ifndef SILLYNESS
+#ifdef SILLYNESS
     if (InitXThing(panel->scr)) {	
 	panel->timer = WMAddTimerHandler(100, logoPushCallback, panel);
 	panel->cycle = 0;
@@ -1212,8 +1331,10 @@ wShowLegalPanel(WScreen *scr)
     WWindow *wwin;
 
     if (legalPanel) {
-	wRaiseFrame(legalPanel->wwin->frame->core);
-	wSetFocusTo(scr, legalPanel->wwin);
+	if (legalPanel->scr == scr) {
+	    wRaiseFrame(legalPanel->wwin->frame->core);
+	    wSetFocusTo(scr, legalPanel->wwin);
+	}
 	return;
     }
 
@@ -1247,6 +1368,9 @@ wShowLegalPanel(WScreen *scr)
     WSETUFLAG(wwin, no_close_button, 0);
     wWindowUpdateButtonImages(wwin);
     wFrameWindowShowButton(wwin->frame, WFF_RIGHT_BUTTON);
+#ifdef XKB_BUTTON_HINT
+    wFrameWindowHideButton(wwin->frame, WFF_LANGUAGE_BUTTON);
+#endif
     wwin->frame->on_click_right = destroyLegalPanel;
     
     panel->wwin = wwin;
@@ -1258,3 +1382,422 @@ wShowLegalPanel(WScreen *scr)
     legalPanel = panel;
 }
 
+
+/*
+ ***********************************************************************
+ * Crashing Dialog Panel
+ ***********************************************************************
+ */
+
+extern WDDomain *WDWindowAttributes;
+
+
+typedef struct _CrashPanel {
+    WMWindow *win;            /* main window */
+
+    WMLabel *iconL;           /* application icon */
+    WMLabel *nameL;           /* title of panel */
+
+    WMFrame *sepF;            /* separator frame */
+
+    WMLabel *noteL;           /* Title of note */
+    WMLabel *note2L;          /* body of note with what happened */
+
+    WMFrame *whatF;           /* "what to do next" frame */
+    WMPopUpButton *whatP;     /* action selection popup button */
+
+    WMButton *okB;            /* ok button */
+
+    Bool done;                /* if finished with this dialog */
+    int action;               /* what to do after */
+
+    KeyCode retKey;
+
+} CrashPanel;
+
+
+static void
+handleKeyPress(XEvent *event, void *clientData)
+{
+    CrashPanel *panel = (CrashPanel*)clientData;
+
+    if (event->xkey.keycode == panel->retKey) {
+	WMPerformButtonClick(panel->okB);
+    }
+}
+
+
+static void
+okButtonCallback(void *self, void *clientData)
+{
+    CrashPanel *panel = (CrashPanel*)clientData;
+
+    panel->done = True;
+}
+
+
+static void
+setCrashAction(void *self, void *clientData)
+{
+    WMPopUpButton *pop = (WMPopUpButton*)self;
+    CrashPanel *panel = (CrashPanel*)clientData;
+
+    panel->action = WMGetPopUpButtonSelectedItem(pop);
+}
+
+
+static WMPixmap*
+getWindowMakerIconImage(WMScreen *scr)
+{
+    proplist_t dict, key, option, value=NULL;
+    WMPixmap *pix=NULL;
+    char *path;
+
+    PLSetStringCmpHook(NULL);
+
+    key = PLMakeString("Logo.WMPanel");
+    option = PLMakeString("Icon");
+
+    dict = PLGetDictionaryEntry(WDWindowAttributes->dictionary, key);
+
+    if (dict) {
+        value = PLGetDictionaryEntry(dict, option);
+    }
+
+    PLRelease(key);
+    PLRelease(option);
+
+    PLSetStringCmpHook(StringCompareHook);
+
+    if (value && PLIsString(value)) {
+        path = FindImage(wPreferences.icon_path, PLGetString(value));
+
+        if (path) {
+            RImage *image;
+
+            image = RLoadImage(WMScreenRContext(scr), path, 0);
+            if (image) {
+                pix = WMCreatePixmapFromRImage(scr, image, 0);
+                RDestroyImage(image);
+            }
+            free(path);
+        }
+    }
+
+    return pix;
+}
+
+
+#define PWIDTH	295
+#define PHEIGHT	345
+
+
+int
+wShowCrashingDialogPanel(int whatSig)
+{
+    CrashPanel *panel;
+    WMScreen *scr;
+    WMFont *font;
+    WMPixmap *logo;
+    int screen_no, scr_width, scr_height;
+    int action;
+    char buf[256];
+
+    panel = wmalloc(sizeof(CrashPanel));
+    memset(panel, 0, sizeof(CrashPanel));
+
+    screen_no = DefaultScreen(dpy);
+    scr_width = WidthOfScreen(ScreenOfDisplay(dpy, screen_no));
+    scr_height = HeightOfScreen(ScreenOfDisplay(dpy, screen_no));
+
+    scr = WMCreateScreen(dpy, screen_no);
+    if (!scr) {
+        wsyserror(_("cannot open connection for crashing dialog panel. Aborting."));
+	return WMAbort;
+    }
+
+    panel->retKey = XKeysymToKeycode(dpy, XK_Return);
+
+    panel->win = WMCreateWindow(scr, "crashingDialog");
+    WMResizeWidget(panel->win, PWIDTH, PHEIGHT);
+    WMMoveWidget(panel->win, (scr_width - PWIDTH)/2, (scr_height - PHEIGHT)/2);
+
+    logo = getWindowMakerIconImage(scr);
+    if (logo) {
+        panel->iconL = WMCreateLabel(panel->win);
+        WMResizeWidget(panel->iconL, 64, 64);
+        WMMoveWidget(panel->iconL, 10, 10);
+        WMSetLabelImagePosition(panel->iconL, WIPImageOnly);
+        WMSetLabelImage(panel->iconL, logo);
+    }
+
+    panel->nameL = WMCreateLabel(panel->win);
+    WMResizeWidget(panel->nameL, 190, 18);
+    WMMoveWidget(panel->nameL, 80, 35);
+    WMSetLabelTextAlignment(panel->nameL, WALeft);
+    font = WMBoldSystemFontOfSize(scr, 18);
+    WMSetLabelFont(panel->nameL, font);
+    WMReleaseFont(font);
+    WMSetLabelText(panel->nameL, _("Fatal error"));
+
+    panel->sepF = WMCreateFrame(panel->win);
+    WMResizeWidget(panel->sepF, PWIDTH+4, 2);
+    WMMoveWidget(panel->sepF, -2, 80);
+
+    panel->noteL = WMCreateLabel(panel->win);
+    WMResizeWidget(panel->noteL, PWIDTH-20, 40);
+    WMMoveWidget(panel->noteL, 10, 90);
+    WMSetLabelTextAlignment(panel->noteL, WAJustified);
+#ifdef SYS_SIGLIST_DECLARED
+    sprintf(buf, _("Window Maker received signal %i\n(%s)."),
+            whatSig, sys_siglist[whatSig]);
+#else
+    sprintf(buf, _("Window Maker received signal %i."), whatSig);
+#endif
+    WMSetLabelText(panel->noteL, buf);
+
+    panel->note2L = WMCreateLabel(panel->win);
+    WMResizeWidget(panel->note2L, PWIDTH-20, 100);
+    WMMoveWidget(panel->note2L, 10, 130);
+    WMSetLabelTextAlignment(panel->note2L, WALeft);
+    WMSetLabelText(panel->note2L,
+                   _(" This fatal error occured probably due to a bug."
+                   " Please fill the included BUGFORM and "
+                   "report it to bugs@windowmaker.org."));
+
+
+    panel->whatF = WMCreateFrame(panel->win);
+    WMResizeWidget(panel->whatF, PWIDTH-20, 50);
+    WMMoveWidget(panel->whatF, 10, 240);
+    WMSetFrameTitle(panel->whatF, _("What do you want to do now?"));
+
+    panel->whatP = WMCreatePopUpButton(panel->whatF);
+    WMResizeWidget(panel->whatP, PWIDTH-20-70, 20);
+    WMMoveWidget(panel->whatP, 35, 20);
+    WMSetPopUpButtonPullsDown(panel->whatP, False);
+    WMSetPopUpButtonText(panel->whatP, _("Select action"));
+    WMAddPopUpButtonItem(panel->whatP, _("Abort and leave a core file"));
+    WMAddPopUpButtonItem(panel->whatP, _("Restart Window Maker"));
+    WMAddPopUpButtonItem(panel->whatP, _("Start alternate window manager"));
+    WMSetPopUpButtonAction(panel->whatP, setCrashAction, panel);
+    WMSetPopUpButtonSelectedItem(panel->whatP, WMRestart);
+    panel->action = WMRestart;
+
+    WMMapSubwidgets(panel->whatF);
+
+    panel->okB = WMCreateCommandButton(panel->win);
+    WMResizeWidget(panel->okB, 80, 26);
+    WMMoveWidget(panel->okB, 205, 309);
+    WMSetButtonText(panel->okB, _("OK"));
+    WMSetButtonImage(panel->okB, WMGetSystemPixmap(scr, WSIReturnArrow));
+    WMSetButtonAltImage(panel->okB, WMGetSystemPixmap(scr, WSIHighlightedReturnArrow));
+    WMSetButtonImagePosition(panel->okB, WIPRight);
+    WMSetButtonAction(panel->okB, okButtonCallback, panel);
+
+    panel->done = 0;
+
+    WMCreateEventHandler(WMWidgetView(panel->win), KeyPressMask,
+                         handleKeyPress, panel);
+
+    WMRealizeWidget(panel->win);
+    WMMapSubwidgets(panel->win);
+
+    WMMapWidget(panel->win);
+
+    XSetInputFocus(dpy, WMWidgetXID(panel->win), RevertToParent, CurrentTime);
+
+    while (!panel->done) {
+        XEvent event;
+
+        WMNextEvent(dpy, &event);
+        WMHandleEvent(&event);
+    }
+
+    action = panel->action;
+
+    WMUnmapWidget(panel->win);
+    WMDestroyWidget(panel->win);
+    free(panel);
+
+    return action;
+}
+
+
+
+
+/*****************************************************************************
+ *			About GNUstep Panel
+ *****************************************************************************/
+
+
+static void
+drawGNUstepLogo(Display *dpy, Drawable d, int width, int height,
+		unsigned long blackPixel, unsigned long whitePixel)
+{
+    GC gc;
+    XGCValues gcv;
+    XRectangle rects[3];
+
+    gcv.foreground = blackPixel;
+    gc = XCreateGC(dpy, d, GCForeground, &gcv);
+
+    XFillArc(dpy, d, gc, width/45, height/45, 
+	     width - 2*width/45, height - 2*height/45, 0, 360*64);
+
+    rects[0].x = 0;
+    rects[0].y = 37*height/45;
+    rects[0].width = width/3;
+    rects[0].height = height - rects[0].y;
+
+    rects[1].x = rects[0].width;
+    rects[1].y = height/2;
+    rects[1].width = width - 2*width/3;
+    rects[1].height = height - rects[1].y;
+
+    rects[2].x = 2*width/3;
+    rects[2].y = height - 37*height/45;
+    rects[2].width = width/3;
+    rects[2].height = height - rects[2].y;
+
+    XSetClipRectangles(dpy, gc, 0, 0, rects, 3, Unsorted);
+    XFillRectangle(dpy, d, gc, 0, 0, width, height);
+
+    XSetForeground(dpy, gc, whitePixel);
+    XFillArc(dpy, d, gc, width/45, height/45, 
+	     width - 2*width/45, height - 2*height/45, 0, 360*64);
+
+    XFreeGC(dpy, gc);
+}
+
+
+
+#define GNUSTEP_TEXT \
+	"Window Maker is part of the GNUstep project.\n"\
+	"The GNUstep project aims to create a free\n"\
+	"implementation of the OpenStep(tm) specification\n"\
+	"which is a object-oriented framework for\n"\
+	"creating advanced graphical, multi-platform\n"\
+	"applications. Additionally, a development and\n"\
+	"user desktop enviroment will be created on top\n"\
+	"of the framework. For more information about\n"\
+	"GNUstep, please visit: www.gnustep.org"
+
+
+typedef struct {
+    WScreen *scr;
+    
+    WWindow *wwin;
+
+    WMWindow *win;
+
+    WMLabel *gstepL;
+    WMLabel *textL;
+} GNUstepPanel;
+
+
+static GNUstepPanel *gnustepPanel = NULL;
+
+static void
+destroyGNUstepPanel(WCoreWindow *foo, void *data, XEvent *event)
+{
+    WMUnmapWidget(gnustepPanel->win);
+
+    WMDestroyWidget(gnustepPanel->win);    
+
+    wUnmanageWindow(gnustepPanel->wwin, False, False);
+
+    free(gnustepPanel);
+
+    gnustepPanel = NULL;
+}
+
+
+void
+wShowGNUstepPanel(WScreen *scr)
+{
+    GNUstepPanel *panel;
+    Window parent;
+    WWindow *wwin;
+    WMPixmap *pixmap;
+    WMColor *color;
+
+    if (gnustepPanel) {
+	if (gnustepPanel->scr == scr) {
+	    wRaiseFrame(gnustepPanel->wwin->frame->core);
+	    wSetFocusTo(scr, gnustepPanel->wwin);
+	}
+	return;
+    }
+
+    panel = wmalloc(sizeof(GNUstepPanel));
+
+    panel->scr = scr;
+
+    panel->win = WMCreateWindow(scr->wmscreen, "About GNUstep");
+    WMResizeWidget(panel->win, 325, 200);
+
+    pixmap = WMCreatePixmap(scr->wmscreen, 130, 130, 
+			    WMScreenDepth(scr->wmscreen), True);
+
+    color = WMCreateNamedColor(scr->wmscreen, "gray50", True);
+
+    drawGNUstepLogo(dpy, WMGetPixmapXID(pixmap), 130, 130,
+		    WMColorPixel(color), scr->white_pixel);
+
+    WMReleaseColor(color);
+
+    XSetForeground(dpy, scr->mono_gc, 0);
+    XFillRectangle(dpy, WMGetPixmapMaskXID(pixmap), scr->mono_gc, 0, 0, 
+		   130, 130);
+    drawGNUstepLogo(dpy, WMGetPixmapMaskXID(pixmap), 130, 130, 1, 1);
+
+    panel->gstepL = WMCreateLabel(panel->win);
+    WMResizeWidget(panel->gstepL, 285, 64);
+    WMMoveWidget(panel->gstepL, 20, 0);
+    WMSetLabelTextAlignment(panel->gstepL, WARight);
+    WMSetLabelText(panel->gstepL, "GNUstep");
+    {
+	WMFont *font = WMBoldSystemFontOfSize(scr->wmscreen, 24);
+
+	WMSetLabelFont(panel->gstepL, font);
+	WMReleaseFont(font);
+    }
+
+    panel->textL = WMCreateLabel(panel->win);
+    WMResizeWidget(panel->textL, 275, 130);
+    WMMoveWidget(panel->textL, 30, 50);
+    WMSetLabelTextAlignment(panel->textL, WARight);
+    WMSetLabelImagePosition(panel->textL, WIPOverlaps);
+    WMSetLabelText(panel->textL, GNUSTEP_TEXT);
+    WMSetLabelImage(panel->textL, pixmap);
+
+    WMReleasePixmap(pixmap);
+
+    WMRealizeWidget(panel->win);
+    WMMapSubwidgets(panel->win);
+
+    parent = XCreateSimpleWindow(dpy, scr->root_win, 0, 0, 325, 200, 0, 0, 0);
+
+    XReparentWindow(dpy, WMWidgetXID(panel->win), parent, 0, 0);
+
+    wwin = wManageInternalWindow(scr, parent, None, "About GNUstep",
+				 (scr->scr_width - 325)/2,
+				 (scr->scr_height - 200)/2, 325, 200);
+
+    WSETUFLAG(wwin, no_closable, 0);
+    WSETUFLAG(wwin, no_close_button, 0);
+    wWindowUpdateButtonImages(wwin);
+    wFrameWindowShowButton(wwin->frame, WFF_RIGHT_BUTTON);
+#ifdef XKB_BUTTON_HINT
+    wFrameWindowHideButton(wwin->frame, WFF_LANGUAGE_BUTTON);
+#endif
+    wwin->frame->on_click_right = destroyGNUstepPanel;
+    
+    panel->wwin = wwin;
+    
+    WMMapWidget(panel->win);
+
+    wWindowMap(wwin);
+
+    gnustepPanel = panel;
+}
