@@ -3,6 +3,7 @@
  *  Window Maker window manager
  * 
  *  Copyright (c) 1997, 1998 Alfredo K. Kojima
+ *  Copyright (c) 1998       Dan Pascu
  * 
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -43,6 +44,7 @@ typedef struct WDock {
 #define EXTENDED_DOCK  WM_CLIP
     int type;
 
+    WMagicNumber auto_expand_magic;
     WMagicNumber auto_collapse_magic;
     WMagicNumber auto_raise_magic;
     WMagicNumber auto_lower_magic;
@@ -54,7 +56,6 @@ typedef struct WDock {
     unsigned int mapped:1;
     unsigned int lowered:1;
     unsigned int attract_icons:1;      /* If clip should attract app-icons */
-    unsigned int keep_attracted:1;     /* if keep them when application exits*/
 
     unsigned int lclip_button_pushed:1;
     unsigned int rclip_button_pushed:1;
@@ -77,7 +78,7 @@ void wDockShowIcons(WDock *dock);
 void wDockLower(WDock *dock);
 void wDockRaise(WDock *dock);
 void wDockRaiseLower(WDock *dock);
-void wDockSaveState(WScreen *scr);
+void wDockSaveState(WScreen *scr, proplist_t old_state);
 
 Bool wDockAttachIcon(WDock *dock, WAppIcon *icon, int x, int y);
 Bool wDockSnapIcon(WDock *dock, WAppIcon *icon, int req_x, int req_y,
@@ -86,7 +87,7 @@ Bool wDockFindFreeSlot(WDock *dock, int *req_x, int *req_y);
 void wDockDetach(WDock *dock, WAppIcon *icon);
 
 void wDockTrackWindowLaunch(WDock *dock, Window window);
-WAppIcon *wDockFindIconFor(WDock *dock, Window window);
+WAppIcon *wDockFindIconForWindow(WDock *dock, Window window);
 void wDockDoAutoLaunch(WDock *dock, int workspace);
 void wDockLaunchWithState(WDock *dock, WAppIcon *btn, WSavedState *state);
 
@@ -107,4 +108,10 @@ void wClipUpdateForWorkspaceChange(WScreen *scr, int workspace);
 
 RImage *wClipMakeTile(WScreen *scr, RImage *normalTile);
 
+#define WO_FAILED          0
+#define WO_NOT_APPLICABLE  1
+#define WO_SUCCESS         2
+
+int wClipMakeIconOmnipresent(WAppIcon *aicon, int omnipresent);
+
 #endif
