@@ -23,6 +23,9 @@
 
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
+#ifdef KEEP_XKB_LOCK_STATUS     
+#include <X11/XKBlib.h>         
+#endif /* KEEP_XKB_LOCK_STATUS */
 
 #include <stdlib.h>
 #include <string.h>
@@ -60,9 +63,13 @@ static void updateTitlebar(WFrameWindow *fwin);
 
 WFrameWindow*
 wFrameWindowCreate(WScreen *scr, int wlevel, int x, int y, 
-		   int width, int height, int flags,
+		   int width, int height, int *clearance, int flags,
 		   WTexture **title_texture, WTexture **resize_texture,
-		   unsigned long *color, GC *gc, WFont **font)
+		   unsigned long *color,
+#ifdef DRAWSTRING_PLUGIN
+           int function_offset,
+#endif
+           GC *gc, WMFont **font)
 {
     WFrameWindow *fwin;
     
@@ -76,11 +83,20 @@ wFrameWindowCreate(WScreen *scr, int wlevel, int x, int y,
     fwin->title_texture = title_texture;
     fwin->resizebar_texture = resize_texture;
     fwin->title_pixel = color;
+    fwin->title_clearance = clearance;
+#ifdef DRAWSTRING_PLUGIN
+    fwin->drawstring_proc_offset = function_offset;
+#endif
     fwin->title_gc = gc;
     fwin->font = font;
-    
+#ifdef KEEP_XKB_LOCK_STATUS
+    fwin->languagemode = XkbGroup1Index;
+    fwin->last_languagemode = XkbGroup2Index;
+#endif
+
     fwin->core = wCoreCreateTopLevel(scr, x, y, width, height, 
-				     FRAME_BORDER_WIDTH);
+				     (flags & WFF_BORDER) 
+				     ? FRAME_BORDER_WIDTH : 0);
     if (wPreferences.use_saveunders) {
 	unsigned long vmask;
 	XSetWindowAttributes attribs;
@@ -122,7 +138,7 @@ wFrameWindowUpdateBorders(WFrameWindow *fwin, int flags)
 	height = fwin->core->height - fwin->top_width - fwin->bottom_width;
 
     if (flags & WFF_TITLEBAR)
-	theight = (*fwin->font)->height + TITLEBAR_EXTRA_HEIGHT;
+	theight = WMFontHeight(*fwin->font) + (*fwin->title_clearance + TITLEBAR_EXTEND_SPACE) * 2;
     else
 	theight = 0;
     
@@ -144,6 +160,16 @@ wFrameWindowUpdateBorders(WFrameWindow *fwin, int flags)
 		if (fwin->left_button) {
 		    wCoreConfigure(fwin->left_button, 0, 0, bsize, bsize);
 		}
+#ifdef XKB_BUTTON_HINT
+		if (fwin->language_button)
+		    if (fwin->flags.hide_left_button || !fwin->left_button
+			|| fwin->flags.lbutton_dont_fit) {
+			wCoreConfigure(fwin->language_button, 0, 0, bsize, bsize);
+		    } else {
+			wCoreConfigure(fwin->language_button, bsize, 0, bsize, bsize);
+		    }
+#endif
+
 		
 		if (fwin->right_button) {
 		    wCoreConfigure(fwin->right_button, width-bsize+1,
@@ -154,6 +180,13 @@ wFrameWindowUpdateBorders(WFrameWindow *fwin, int flags)
 		    wCoreConfigure(fwin->left_button, 3, (theight-bsize)/2,
 				   bsize, bsize);
 		}
+        
+#ifdef XKB_BUTTON_HINT
+		if (fwin->language_button) {
+		    wCoreConfigure(fwin->language_button, 6 + bsize, (theight-bsize)/2,
+				   bsize, bsize);
+		}
+#endif
 		
 		if (fwin->right_button) {
 		    wCoreConfigure(fwin->right_button, width-bsize-3,
@@ -165,15 +198,24 @@ wFrameWindowUpdateBorders(WFrameWindow *fwin, int flags)
 	    /* we had a titlebar, but now we don't need it anymore */
 	    for (i=0; i < (fwin->flags.single_texture ? 1 : 3); i++) {
 		FREE_PIXMAP(fwin->title_back[i]);
-                if (wPreferences.new_style) {
+		if (wPreferences.new_style) {
 		    FREE_PIXMAP(fwin->lbutton_back[i]);
 		    FREE_PIXMAP(fwin->rbutton_back[i]);
-                }
+#ifdef XKB_BUTTON_HINT
+		    FREE_PIXMAP(fwin->languagebutton_back[i]);
+#endif
+		}
 	    }
 	    if (fwin->left_button)
 		wCoreDestroy(fwin->left_button);
 	    fwin->left_button = NULL;
 	    
+#ifdef XKB_BUTTON_HINT
+	    if (fwin->language_button)
+		wCoreDestroy(fwin->language_button);
+	    fwin->language_button = NULL;
+#endif
+
 	    if (fwin->right_button)
 		wCoreDestroy(fwin->right_button);
 	    fwin->right_button = NULL;
@@ -191,33 +233,61 @@ wFrameWindowUpdateBorders(WFrameWindow *fwin, int flags)
 	    
 	    fwin->flags.titlebar = 1;
 	    fwin->titlebar = wCoreCreate(fwin->core, 0, 0, width+1, theight);
-	    
+
 	    if (flags & WFF_LEFT_BUTTON) {
 		fwin->flags.left_button = 1;
-                if (wPreferences.new_style) {
-                    fwin->left_button = wCoreCreate(fwin->core, 0, 0,
-                                                    bsize, bsize);
+		if (wPreferences.new_style) {
+		    fwin->left_button = wCoreCreate(fwin->core, 0, 0,
+						    bsize, bsize);
 		    if (width < theight*4) {
 			fwin->flags.lbutton_dont_fit = 1;
 		    } else {
 			XMapRaised(dpy, fwin->left_button->window);
 		    }
-                } else {
+		} else {
 		    fwin->left_button = 
 			wCoreCreate(fwin->titlebar, 3, (theight-bsize)/2,
 				    bsize, bsize);
-
+		    
 		    XSetWindowBackground(dpy, fwin->left_button->window,
 					 scr->widget_texture->normal.pixel);
-
+		    
 		    if (width < theight*3) {
 			fwin->flags.lbutton_dont_fit = 1;
 		    } else {
 			XMapRaised(dpy, fwin->left_button->window);
 		    }
-                }
+		}
 	    }
-	    
+        
+#ifdef XKB_BUTTON_HINT
+	    if (flags & WFF_LANGUAGE_BUTTON) {
+		fwin->flags.language_button = 1;
+		if (wPreferences.new_style) {
+		    fwin->language_button = wCoreCreate(fwin->core,
+							bsize, 0, bsize, bsize);
+		    
+		    if (width < theight*4) {
+			fwin->flags.languagebutton_dont_fit = 1;
+		    } else {
+			XMapRaised(dpy, fwin->language_button->window);
+		    }
+		} else {
+		    fwin->language_button = 
+			wCoreCreate(fwin->titlebar, bsize + 6, (theight-bsize)/2,
+				    bsize, bsize); 
+		    
+		    XSetWindowBackground(dpy, fwin->language_button->window,
+					 scr->widget_texture->normal.pixel); 
+		    
+		    if (width < theight*3) {
+			fwin->flags.languagebutton_dont_fit = 1;
+		    } else {
+			XMapRaised(dpy, fwin->language_button->window);
+		    }
+		}
+	    }
+#endif
 	    
 	    if (flags & WFF_RIGHT_BUTTON) {
 		fwin->flags.right_button = 1;
@@ -249,14 +319,10 @@ wFrameWindowUpdateBorders(WFrameWindow *fwin, int flags)
 	}
     }
     checkTitleSize(fwin);
-
-    if (flags & WFF_RESIZEBAR) {
-	fwin->bottom_width = RESIZEBAR_HEIGHT;
-    } else {
-	fwin->bottom_width = 0;
-    } 
     
     if (flags & WFF_RESIZEBAR) {
+	fwin->bottom_width = RESIZEBAR_HEIGHT;
+	
 	if (!fwin->resizebar) {
 	    fwin->flags.resizebar = 1;
 	    fwin->resizebar = wCoreCreate(fwin->core, 0,
@@ -280,6 +346,8 @@ wFrameWindowUpdateBorders(WFrameWindow *fwin, int flags)
 	    }
 	}
     } else {
+	fwin->bottom_width = 0;
+	
 	if (fwin->resizebar) {
 	    fwin->bottom_width = 0;
 	    wCoreDestroy(fwin->resizebar);
@@ -293,6 +361,13 @@ wFrameWindowUpdateBorders(WFrameWindow *fwin, int flags)
 			   height + fwin->top_width + fwin->bottom_width);
     }
     
+    
+    if (flags & WFF_BORDER) {
+	XSetWindowBorderWidth(dpy, fwin->core->window, FRAME_BORDER_WIDTH);
+    } else {
+	XSetWindowBorderWidth(dpy, fwin->core->window, 0);
+    }
+    
     /* setup object descriptors */
     
     if (fwin->titlebar) {
@@ -316,6 +391,16 @@ wFrameWindowUpdateBorders(WFrameWindow *fwin, int flags)
 	fwin->left_button->descriptor.handle_mousedown = buttonMouseDown;
     }
     
+#ifdef XKB_BUTTON_HINT
+    if (fwin->language_button) {
+	fwin->language_button->descriptor.handle_expose = handleButtonExpose;
+	fwin->language_button->descriptor.parent = fwin;
+	fwin->language_button->descriptor.parent_type = WCLASS_FRAME;
+	fwin->language_button->descriptor.handle_mousedown = buttonMouseDown;
+    }
+#endif
+
+    
     if (fwin->right_button) {
 	fwin->right_button->descriptor.parent = fwin;
 	fwin->right_button->descriptor.parent_type = WCLASS_FRAME;
@@ -336,6 +421,11 @@ wFrameWindowDestroy(WFrameWindow *fwin)
     if (fwin->left_button)
 	wCoreDestroy(fwin->left_button);
     
+#ifdef XKB_BUTTON_HINT
+    if (fwin->language_button)
+    wCoreDestroy(fwin->language_button);
+#endif
+    
     if (fwin->right_button)
 	wCoreDestroy(fwin->right_button);
     
@@ -356,6 +446,9 @@ wFrameWindowDestroy(WFrameWindow *fwin)
 	FREE_PIXMAP(fwin->title_back[i]);
         if (wPreferences.new_style) {
 	    FREE_PIXMAP(fwin->lbutton_back[i]);
+#ifdef XKB_BUTTON_HINT
+	    FREE_PIXMAP(fwin->languagebutton_back[i]);
+#endif
 	    FREE_PIXMAP(fwin->rbutton_back[i]);
         }
     }
@@ -383,7 +476,7 @@ updateTitlebar(WFrameWindow *fwin)
     int x, w;
     int theight;
     
-    theight = (*fwin->font)->height + TITLEBAR_EXTRA_HEIGHT;
+	theight = WMFontHeight(*fwin->font) + (*fwin->title_clearance + TITLEBAR_EXTEND_SPACE) * 2;
 
     x = 0;
     w = fwin->core->width + 1;
@@ -392,11 +485,50 @@ updateTitlebar(WFrameWindow *fwin)
 	if (fwin->flags.hide_left_button || !fwin->left_button
 	    || fwin->flags.lbutton_dont_fit) {
 	    x = 0;
+#ifdef XKB_BUTTON_HINT
+        if(fwin->language_button)
+            wCoreConfigure(fwin->language_button, 0, 0,
+                    fwin->language_button->width,
+                    fwin->language_button->width);
+#endif
 	} else {
+#ifdef XKB_BUTTON_HINT
+        if(fwin->language_button)
+            wCoreConfigure(fwin->language_button, fwin->left_button->width, 0,
+                    fwin->language_button->width,
+                    fwin->language_button->width);
+#endif
 	    x = fwin->left_button->width;
 	    w -= fwin->left_button->width;
 	}
+#ifdef XKB_BUTTON_HINT
+	if (fwin->flags.hide_language_button || !fwin->language_button
+	    || fwin->flags.languagebutton_dont_fit) {
+    } else {
+	    x += fwin->language_button->width;
+	    w -= fwin->language_button->width;
     }
+#endif
+    }
+#ifdef XKB_BUTTON_HINT
+    else {
+        int bsize = theight - 7;
+        if (fwin->flags.hide_left_button || !fwin->left_button
+            || fwin->flags.lbutton_dont_fit) {
+            if(fwin->language_button)
+                wCoreConfigure(fwin->language_button, 3, (theight-bsize)/2,
+                        fwin->language_button->width,
+                        fwin->language_button->width);
+        }
+        else {
+            if(fwin->language_button)
+                wCoreConfigure(fwin->language_button,
+                        6 + fwin->left_button->width, (theight-bsize)/2,
+                        fwin->language_button->width,
+                        fwin->language_button->width);
+        }
+    }
+#endif
     
     if (wPreferences.new_style) {
 	if (!fwin->flags.hide_right_button && fwin->right_button
@@ -424,13 +556,24 @@ wFrameWindowHideButton(WFrameWindow *fwin, int flags)
 	XUnmapWindow(dpy, fwin->left_button->window);
 	fwin->flags.hide_left_button = 1;
     }
+    
+#ifdef XKB_BUTTON_HINT
+    if ((flags & WFF_LANGUAGE_BUTTON) && fwin->language_button) {
+	XUnmapWindow(dpy, fwin->language_button->window);
+    fwin->flags.hide_language_button = 1;
+    }
+#endif
 
     if (fwin->titlebar) {
 	if (wPreferences.new_style) {
 	    updateTitlebar(fwin);
 	} else {
+#ifdef XKB_BUTTON_HINT
+	    updateTitlebar(fwin);
+#else
 	    XClearWindow(dpy, fwin->titlebar->window);
 	    wFrameWindowPaint(fwin);
+#endif
 	}
 	checkTitleSize(fwin);
     }
@@ -448,6 +591,17 @@ wFrameWindowShowButton(WFrameWindow *fwin, int flags)
 	
 	fwin->flags.hide_right_button = 0;
     }
+
+#ifdef XKB_BUTTON_HINT
+    if ((flags & WFF_LANGUAGE_BUTTON) && fwin->language_button
+	&& fwin->flags.hide_language_button) {
+	
+	if (!fwin->flags.languagebutton_dont_fit)
+	    XMapWindow(dpy, fwin->language_button->window);
+	
+	fwin->flags.hide_language_button = 0;
+    }
+#endif
     
     if ((flags & WFF_LEFT_BUTTON) && fwin->left_button
 	&& fwin->flags.hide_left_button) {
@@ -472,21 +626,33 @@ wFrameWindowShowButton(WFrameWindow *fwin, int flags)
 
 
 static void
+#ifdef XKB_BUTTON_HINT
+renderTexture(WScreen *scr, WTexture *texture, int width, int height, 
+	      int bwidth, int bheight, int left, int language, int right,
+	      Pixmap *title, Pixmap *lbutton, Pixmap *languagebutton, Pixmap *rbutton)
+#else
 renderTexture(WScreen *scr, WTexture *texture, int width, int height, 
 	      int bwidth, int bheight, int left, int right,
 	      Pixmap *title, Pixmap *lbutton, Pixmap *rbutton)
+#endif
 {
     RImage *img;
     RImage *limg, *rimg, *mimg;
+#ifdef XKB_BUTTON_HINT
+    RImage *timg;
+#endif
     int x, w;
-    
+
     *title = None;
     *lbutton = None;
     *rbutton = None;
+#ifdef XKB_BUTTON_HINT
+    *languagebutton = None;
+#endif
 
     img = wTextureRenderImage(texture, width, height, WREL_FLAT);
     if (!img) {
-        wwarning(_("could not render gradient: %s"), RMessageForError(RErrorCode));
+        wwarning(_("could not render texture: %s"), RMessageForError(RErrorCode));
         return;
     }
     
@@ -498,6 +664,13 @@ renderTexture(WScreen *scr, WTexture *texture, int width, int height,
 	
 	x = 0;
 	w = img->width;
+    
+#ifdef XKB_BUTTON_HINT
+    if (language) {
+	    timg = RGetSubImage(img, bwidth * left, 0, bwidth, bheight);
+	} else
+	    timg = NULL;
+#endif
 	
 	if (limg) {
 	    RBevelImage(limg, RBEV_RAISED2);
@@ -508,6 +681,18 @@ renderTexture(WScreen *scr, WTexture *texture, int width, int height,
 	    w -= limg->width;
 	    RDestroyImage(limg);
 	}
+
+#ifdef XKB_BUTTON_HINT
+    if (timg) {
+	    RBevelImage(timg, RBEV_RAISED2);
+	    if (!RConvertImage(scr->rcontext, timg, languagebutton)) {
+		wwarning(_("error rendering image:%s"), RMessageForError(RErrorCode));
+	    }
+	    x += timg->width;
+	    w -= timg->width;
+	    RDestroyImage(timg);
+    }
+#endif
 	
 	if (right) {
 	    rimg = RGetSubImage(img, width - bwidth, 0,	bwidth, bheight);
@@ -550,6 +735,59 @@ renderTexture(WScreen *scr, WTexture *texture, int width, int height,
 }
 
 
+static void
+renderResizebarTexture(WScreen *scr, WTexture *texture, int width, int height,
+		       int cwidth, Pixmap *pmap)
+{
+    RImage *img;
+    RColor light;
+    RColor dark;
+
+    *pmap = None;
+
+    img = wTextureRenderImage(texture, width, height, WREL_FLAT);
+    if (!img) {
+        wwarning(_("could not render texture: %s"),
+		 RMessageForError(RErrorCode));
+        return;
+    }
+
+    light.alpha = 0;
+    light.red = light.green = light.blue = 80;
+
+    dark.alpha = 0;
+    dark.red = dark.green = dark.blue = 40;
+
+    ROperateLine(img, RSubtractOperation, 0, 0, width-1, 0, &dark);
+    ROperateLine(img, RAddOperation, 0, 1, width-1, 1, &light);
+
+    ROperateLine(img, RSubtractOperation, cwidth, 2, cwidth, height-1, &dark);
+    ROperateLine(img, RAddOperation, cwidth+1, 2, cwidth+1, height-1, &light);
+
+    if (width > 1)
+	ROperateLine(img, RSubtractOperation, width-cwidth-2, 2, 
+		     width-cwidth-2, height-1, &dark);
+    ROperateLine(img, RAddOperation, width-cwidth-1, 2, width-cwidth-1, 
+		 height-1, &light);
+
+#ifdef SHADOW_RESIZEBAR
+    ROperateLine(img, RAddOperation, 0, 1, 0, height-1, &light);
+    ROperateLine(img, RSubtractOperation, width-1, 1, width-1, height-1,
+		 &dark);
+    ROperateLine(img, RSubtractOperation, 0, height-1, width-1, height-1,
+		 &dark);
+#endif /* SHADOW_RESIZEBAR */
+
+
+    if (!RConvertImage(scr->rcontext, img, pmap)) {
+	wwarning(_("error rendering image: %s"), RMessageForError(RErrorCode));
+    }
+
+    RDestroyImage(img);
+}
+
+
+
 static void
 updateTexture(WFrameWindow *fwin)
 {
@@ -565,6 +803,13 @@ updateTexture(WFrameWindow *fwin)
 		if (fwin->left_button && fwin->lbutton_back[i])
 		    XSetWindowBackgroundPixmap(dpy, fwin->left_button->window,
 					       fwin->lbutton_back[i]);
+
+#ifdef XKB_BUTTON_HINT
+		if (fwin->language_button && fwin->languagebutton_back[i]) {
+		    XSetWindowBackgroundPixmap(dpy, fwin->language_button->window,
+					       fwin->languagebutton_back[i]);
+        }
+#endif
 		
 		if (fwin->right_button && fwin->rbutton_back[i])
 		    XSetWindowBackgroundPixmap(dpy, fwin->right_button->window,
@@ -577,6 +822,11 @@ updateTexture(WFrameWindow *fwin)
 		if (fwin->left_button)
 		    XSetWindowBackground(dpy, fwin->left_button->window,
 					 pixel);
+#ifdef XKB_BUTTON_HINT
+		if (fwin->language_button)
+		    XSetWindowBackground(dpy, fwin->language_button->window,
+					 pixel);
+#endif
 		if (fwin->right_button)
 		    XSetWindowBackground(dpy, fwin->right_button->window,
 					 pixel);
@@ -588,17 +838,17 @@ updateTexture(WFrameWindow *fwin)
 	    XClearWindow(dpy, fwin->left_button->window);
 	    handleButtonExpose(&fwin->left_button->descriptor, NULL);
 	}
+#ifdef XKB_BUTTON_HINT
+       if (fwin->language_button) {
+           XClearWindow(dpy, fwin->language_button->window);
+           handleButtonExpose(&fwin->language_button->descriptor, NULL);
+       }
+#endif
 	if (fwin->right_button) {
 	    XClearWindow(dpy, fwin->right_button->window);
 	    handleButtonExpose(&fwin->right_button->descriptor, NULL);
 	}
     }
-    
-    if (fwin->resizebar) {
-	XSetWindowBackground(dpy, fwin->resizebar->window, 
-			     fwin->resizebar_texture[0]->solid.normal.pixel);
-	XClearWindow(dpy, fwin->resizebar->window);
-    }
 }
 
 
@@ -607,44 +857,95 @@ static void
 remakeTexture(WFrameWindow *fwin, int state)
 {
     Pixmap pmap, lpmap, rpmap;
+#ifdef XKB_BUTTON_HINT
+    Pixmap tpmap;
+#endif
 
     if (fwin->title_texture[state] && fwin->titlebar) {
 	FREE_PIXMAP(fwin->title_back[state]);
 	if (wPreferences.new_style) {
 	    FREE_PIXMAP(fwin->lbutton_back[state]);
 	    FREE_PIXMAP(fwin->rbutton_back[state]);
+#ifdef XKB_BUTTON_HINT
+	    FREE_PIXMAP(fwin->languagebutton_back[state]);
+#endif
 	}
 
 	if (fwin->title_texture[state]->any.type!=WTEX_SOLID) {
 	    int left, right;
+#ifdef XKB_BUTTON_HINT
+        int language;
+#endif
 	    int width;
 
 	    /* eventually surrounded by if new_style */
 	    left = fwin->left_button && !fwin->flags.hide_left_button
 		&& !fwin->flags.lbutton_dont_fit;
+#ifdef XKB_BUTTON_HINT
+	    language = fwin->language_button && !fwin->flags.hide_language_button
+		&& !fwin->flags.languagebutton_dont_fit;
+#endif
 	    right = fwin->right_button && !fwin->flags.hide_right_button
 		&& !fwin->flags.rbutton_dont_fit;
-	    
+
 	    width = fwin->core->width+1;
-	    
+
+#ifdef XKB_BUTTON_HINT
+	    renderTexture(fwin->screen_ptr, fwin->title_texture[state],
+			  width, fwin->titlebar->height,
+			  fwin->titlebar->height, fwin->titlebar->height,
+			  left, language, right, &pmap, &lpmap, &tpmap, &rpmap);
+#else
 	    renderTexture(fwin->screen_ptr, fwin->title_texture[state],
 			  width, fwin->titlebar->height,
 			  fwin->titlebar->height, fwin->titlebar->height,
 			  left, right, &pmap, &lpmap, &rpmap);
-	    
+#endif
+
 	    fwin->title_back[state] = pmap;
 	    if (wPreferences.new_style) {
 		fwin->lbutton_back[state] = lpmap;
 		fwin->rbutton_back[state] = rpmap;
+#ifdef XKB_BUTTON_HINT
+		fwin->languagebutton_back[state] = tpmap;
+#endif
 	    }
 	}
     }
+    if (fwin->resizebar_texture && fwin->resizebar_texture[0] 
+	&& fwin->resizebar && state == 0) {
+
+	FREE_PIXMAP(fwin->resizebar_back[0]);
+
+	if (fwin->resizebar_texture[0]->any.type!=WTEX_SOLID) {
+
+	    renderResizebarTexture(fwin->screen_ptr,
+				   fwin->resizebar_texture[0],
+				   fwin->resizebar->width,
+				   fwin->resizebar->height,
+				   fwin->resizebar_corner_width,
+				   &pmap);
+
+	    fwin->resizebar_back[0] = pmap;
+	}
+	
+	/* this part should be in updateTexture() */
+	if (fwin->resizebar_texture[0]->any.type!=WTEX_SOLID) {
+	    XSetWindowBackgroundPixmap(dpy, fwin->resizebar->window,
+				       fwin->resizebar_back[0]);
+	} else {
+	    XSetWindowBackground(dpy, fwin->resizebar->window,
+				 fwin->resizebar_texture[0]->solid.normal.pixel);
+	}
+	XClearWindow(dpy, fwin->resizebar->window);
+    }
 }
 
 
 void
 wFrameWindowPaint(WFrameWindow *fwin)
 {
+    WScreen *scr = fwin->screen_ptr;
 
     if (fwin->flags.is_client_window_frame)
 	fwin->flags.justification = wPreferences.title_justification;
@@ -685,7 +986,8 @@ wFrameWindowPaint(WFrameWindow *fwin)
 		   WREL_RAISED);
     }
     
-    if (fwin->resizebar	&& !fwin->flags.repaint_only_titlebar) {
+    if (fwin->resizebar	&& !fwin->flags.repaint_only_titlebar
+	&& fwin->resizebar_texture[0]->any.type == WTEX_SOLID) {
 	Window win;
 	int w, h;
 	int cw;
@@ -719,11 +1021,11 @@ wFrameWindowPaint(WFrameWindow *fwin)
     
     
     if (fwin->title && fwin->titlebar && !fwin->flags.repaint_only_resizebar) {
-	int x, w;
-	int lofs = 6, rofs = 6;
-	int titlelen;
-	char *title;
-	int allButtons = 1;
+        int x, w;
+        int lofs = 6, rofs = 6;
+        int titlelen;
+        char *title;
+        int allButtons = 1;
 
 
         if (!wPreferences.new_style) {
@@ -732,6 +1034,14 @@ wFrameWindowPaint(WFrameWindow *fwin)
 		lofs += fwin->left_button->width + 3;
 	    else
 		allButtons = 0;
+
+#ifdef XKB_BUTTON_HINT
+	    if (fwin->language_button && !fwin->flags.hide_language_button
+		&& !fwin->flags.languagebutton_dont_fit)
+		lofs += fwin->language_button->width;
+	    else
+		allButtons = 0;
+#endif
 	    
 	    if (fwin->right_button && !fwin->flags.hide_right_button
 		&& !fwin->flags.rbutton_dont_fit)
@@ -739,10 +1049,16 @@ wFrameWindowPaint(WFrameWindow *fwin)
 	    else
 		allButtons = 0;
         }
+
+#ifdef XKB_BUTTON_HINT
+        fwin->languagebutton_image =
+            scr->b_pixmaps[WBUT_XKBGROUP1 + fwin->languagemode];
+#endif
+
 	title = ShrinkString(*fwin->font, fwin->title,
 			     fwin->titlebar->width - lofs - rofs);
 	titlelen = strlen(title);
-	w = wTextWidth((*fwin->font)->font, title, titlelen);
+	w = WMWidthOfString(*fwin->font, title, titlelen);
 
 	switch (fwin->flags.justification) {
 	 case WTJ_LEFT:
@@ -761,31 +1077,39 @@ wFrameWindowPaint(WFrameWindow *fwin)
 	    break;
 	}
 
-#ifdef TITLE_TEXT_SHADOW
-       if(wPreferences.title_shadow){
-          int shadowx,shadowy;
-          XSetForeground(dpy, *fwin->title_gc, 
-              fwin->title_pixel[fwin->flags.state+3]);
-           for(shadowx=0;shadowx<TITLE_TEXT_SHADOW_WIDTH;shadowx++)
-           for(shadowy=0;shadowy<TITLE_TEXT_SHADOW_HEIGHT;shadowy++)
-               wDrawString(fwin->titlebar->window, *fwin->font,
-                       *fwin->title_gc,
-                       x + shadowx + TITLE_TEXT_SHADOW_X_OFFSET, 
-                       (*fwin->font)->y + TITLEBAR_EXTRA_HEIGHT/2
-                        + shadowy + TITLE_TEXT_SHADOW_Y_OFFSET, title,
-                       titlelen);
-       }
-#endif /* TITLE_TEXT_SHADOW */
 
 	XSetForeground(dpy, *fwin->title_gc, 
 		       fwin->title_pixel[fwin->flags.state]);
-	    
-	wDrawString(fwin->titlebar->window, *fwin->font,
-		    *fwin->title_gc, x, 
-		    (*fwin->font)->y + TITLEBAR_EXTRA_HEIGHT/2, title,
-		    titlelen);
+	
+#ifdef DRAWSTRING_PLUGIN
+    if (scr->drawstring_func[fwin->flags.state + fwin->drawstring_proc_offset]) {
+        scr->drawstring_func[fwin->flags.state + fwin->drawstring_proc_offset]->
+            proc.drawString(scr->drawstring_func[fwin->flags.state 
+                    + fwin->drawstring_proc_offset]->arg,
+                    fwin->titlebar->window, *fwin->title_gc,
+                    *fwin->font, x, *fwin->title_clearance + TITLEBAR_EXTEND_SPACE,
+                    fwin->titlebar->width, fwin->top_width, title, titlelen);
+    } else {
+        WMDrawString(scr->wmscreen, fwin->titlebar->window, 
+                *fwin->title_gc, *fwin->font, x, *fwin->title_clearance + TITLEBAR_EXTEND_SPACE, 
+                title, titlelen);
+    }
+#else
+    WMDrawString(scr->wmscreen, fwin->titlebar->window, 
+            *fwin->title_gc, *fwin->font, x, *fwin->title_clearance + TITLEBAR_EXTEND_SPACE, 
+            title, titlelen);
+#endif /* DRAWSTRING_PLUGIN */
 
 	free(title);
+	
+	if (fwin->left_button)
+	    handleButtonExpose(&fwin->left_button->descriptor, NULL);
+	if (fwin->right_button)
+	    handleButtonExpose(&fwin->right_button->descriptor, NULL);
+#ifdef XKB_BUTTON_HINT
+	if (fwin->language_button)
+	    handleButtonExpose(&fwin->language_button->descriptor, NULL);
+#endif 
     }
 }
 
@@ -834,6 +1158,24 @@ reconfigure(WFrameWindow *fwin, int x, int y, int width, int height,
 		fwin->flags.lbutton_dont_fit = 0;
 	    }
 	}
+    
+#ifdef XKB_BUTTON_HINT
+	if (fwin->language_button) {
+	    if (width < fwin->top_width*k && !fwin->flags.languagebutton_dont_fit) {
+		
+		if (!fwin->flags.hide_language_button) {
+		    XUnmapWindow(dpy, fwin->language_button->window);
+		}
+		fwin->flags.languagebutton_dont_fit = 1;
+	    } else if (width >= fwin->top_width*k && fwin->flags.languagebutton_dont_fit) {
+		
+		if (!fwin->flags.hide_language_button) {
+		    XMapWindow(dpy, fwin->language_button->window);
+		}
+		fwin->flags.languagebutton_dont_fit = 0;
+	    }
+	}
+#endif
 	
 	if (fwin->right_button) {
 	    if (width < fwin->top_width*2 && !fwin->flags.rbutton_dont_fit) {
@@ -933,6 +1275,16 @@ wFrameWindowUpdatePushButton(WFrameWindow *fwin, Bool pushed)
 #endif /* OLWM_HINTS */
 
 
+#ifdef XKB_BUTTON_HINT
+void
+wFrameWindowUpdateLanguageButton(WFrameWindow *fwin)
+{
+    paintButton(fwin->language_button, fwin->title_texture[fwin->flags.state],
+		fwin->title_pixel[fwin->flags.state],
+		fwin->languagebutton_image, True);
+}
+#endif /* XKB_BUTTON_HINT */
+
 
 /*********************************************************************/
 
@@ -973,13 +1325,18 @@ checkTitleSize(WFrameWindow *fwin)
 	if (fwin->left_button && !fwin->flags.hide_left_button
 	    && !fwin->flags.lbutton_dont_fit)
 	    width -= fwin->left_button->width + 3;
+
+#ifdef XKB_BUTTON_HINT
+	if (fwin->language_button && !fwin->flags.hide_language_button
+	    && !fwin->flags.languagebutton_dont_fit)
+	    width -= fwin->language_button->width + 3;
+#endif
 	
 	if (fwin->right_button && !fwin->flags.hide_right_button
 	    && !fwin->flags.rbutton_dont_fit)
 	    width -= fwin->right_button->width + 3;
     }
-    if (wTextWidth((*fwin->font)->font, fwin->title, 
-		    strlen(fwin->title)) > width) {
+    if (WMWidthOfString(*fwin->font, fwin->title, strlen(fwin->title)) > width) {
 	fwin->flags.incomplete_title = 1;
     } else {
 	fwin->flags.incomplete_title = 0;
@@ -1086,12 +1443,21 @@ handleButtonExpose(WObjDescriptor *desc, XEvent *event)
     WFrameWindow *fwin = (WFrameWindow*)desc->parent;
     WCoreWindow *button = (WCoreWindow*)desc->self;
 
+#ifdef XKB_BUTTON_HINT
+    if (button == fwin->language_button) {
+        if (wPreferences.modelock){
+            paintButton(button, fwin->title_texture[fwin->flags.state],
+                    fwin->title_pixel[fwin->flags.state],
+                    fwin->languagebutton_image, False);
+        }
+    } else 
+#endif
     if (button == fwin->left_button) {
-	paintButton(button, fwin->title_texture[fwin->flags.state],
-		    fwin->title_pixel[fwin->flags.state],
-		    fwin->lbutton_image, False);
+        paintButton(button, fwin->title_texture[fwin->flags.state],
+                fwin->title_pixel[fwin->flags.state],
+                fwin->lbutton_image, False);
     } else {
-	Bool pushed = False;
+        Bool pushed = False;
 
 #ifdef OLWM_HINTS
 	if (fwin->flags.right_button_pushed_in)
@@ -1156,6 +1522,12 @@ buttonMouseDown(WObjDescriptor *desc, XEvent *event)
     } else {
 	image = fwin->rbutton_image;
     } 
+#ifdef XKB_BUTTON_HINT
+    if (button == fwin->language_button) {
+    if (!wPreferences.modelock) return;
+	image = fwin->languagebutton_image;
+    }
+#endif
     
     pixel = fwin->title_pixel[fwin->flags.state];
     texture = fwin->title_texture[fwin->flags.state];
@@ -1197,6 +1569,13 @@ buttonMouseDown(WObjDescriptor *desc, XEvent *event)
 	    if (fwin->on_click_right)
 		(*fwin->on_click_right)(button, fwin->child, &ev);
 	}
+#ifdef XKB_BUTTON_HINT
+	else if (button == fwin->language_button) {
+	    if (fwin->on_click_language)
+		(*fwin->on_click_language)(button, fwin->child, &ev);
+	}
+#endif
+
     }
 }
 
