@@ -57,6 +57,49 @@ static void miniwindowMouseDown(WObjDescriptor *desc, XEvent *event);
 static void miniwindowDblClick(WObjDescriptor *desc, XEvent *event);
 
 
+/****** Notification Observers ******/
+
+static void
+appearanceObserver(void *self, WMNotification *notif)
+{
+    WIcon *icon = (WIcon*)self;
+    int flags = (int)WMGetNotificationClientData(notif);
+
+    if (flags & WTextureSettings) {
+	icon->force_paint = 1;
+    }
+    if (flags & WFontSettings) {
+	icon->force_paint = 1;
+    }
+    /*
+    if (flags & WColorSettings) {
+    }
+     */
+
+    wIconPaint(icon);
+
+    /* so that the appicon expose handlers will paint the appicon specific 
+     * stuff */
+    XClearArea(dpy, icon->core->window, 0, 0, icon->core->width, 
+	       icon->core->height, True);
+}
+
+
+static void
+tileObserver(void *self, WMNotification *notif)
+{
+    WIcon *icon = (WIcon*)self;
+
+    icon->force_paint = 1;
+    wIconPaint(icon);
+
+    XClearArea(dpy, icon->core->window, 0, 0, 1, 1, True);
+}
+
+/************************************/
+
+
+
 INLINE static void
 getSize(Drawable d, unsigned int *w, unsigned int *h, unsigned int *dep)
 {
@@ -138,7 +181,11 @@ wIconCreate(WWindow *wwin)
     wIconUpdate(icon);
     
     XFlush(dpy);
-    
+
+    WMAddNotificationObserver(appearanceObserver, icon, 
+			      WNIconAppearanceSettingsChanged, icon);
+    WMAddNotificationObserver(tileObserver, icon, 
+			      WNIconTileSettingsChanged, icon);
     return icon;
 }
 
@@ -191,6 +238,11 @@ wIconCreateWithIconFile(WScreen *scr, char *iconfile, int tile)
     
     wIconUpdate(icon);
 
+    WMAddNotificationObserver(appearanceObserver, icon, 
+			      WNIconAppearanceSettingsChanged, icon);
+    WMAddNotificationObserver(tileObserver, icon, 
+			      WNIconTileSettingsChanged, icon);
+
     return icon;
 }
 
@@ -202,6 +254,8 @@ wIconDestroy(WIcon *icon)
     WCoreWindow *core = icon->core;
     WScreen *scr = core->screen_ptr;
 
+    WMRemoveNotificationObserver(icon);
+
     if (icon->handlerID)
 	WMDeleteTimerHandler(icon->handlerID);
 
@@ -254,7 +308,7 @@ makeIcon(WScreen *scr, RImage *icon, int titled, int shadowed, int tileType)
     Pixmap pixmap;
     int x, y, sx, sy;
     unsigned w, h;
-    int theight = scr->icon_title_font->height;
+    int theight = WMFontHeight(scr->icon_title_font);
     
     if (tileType == TILE_NORMAL)
 	tile = RCloneImage(scr->icon_tile);
@@ -345,7 +399,17 @@ wIconValidateIconSize(WScreen *scr, RImage *icon)
     
     if (!icon)
 	return NULL;
-    
+#ifndef DONT_SCALE_ICONS
+    if (wPreferences.icon_size != 64) {
+	w = wPreferences.icon_size * icon->width / 64;
+	h = wPreferences.icon_size * icon->height / 64;
+
+	tmp = RScaleImage(icon, w, h);
+	RDestroyImage(icon);
+	icon = tmp;
+    }
+#endif
+#if 0
     if (icon->width > wPreferences.icon_size
         || icon->height > wPreferences.icon_size) {
 	if (icon->width > icon->height) {
@@ -359,6 +423,7 @@ wIconValidateIconSize(WScreen *scr, RImage *icon)
 	RDestroyImage(icon);
 	icon = tmp;
     }
+#endif
 
     return icon;
 }
@@ -551,7 +616,7 @@ void
 wIconUpdate(WIcon *icon)
 {
     WScreen *scr = icon->core->screen_ptr;
-    int title_height = scr->icon_title_font->height;
+    int title_height = WMFontHeight(scr->icon_title_font);
     WWindow *wwin = icon->owner;
 
     assert(scr->icon_tile!=NULL);
@@ -759,17 +824,17 @@ wIconPaint(WIcon *icon)
 	int l;
 	int w;
 
-	tmp = ShrinkString(scr->icon_title_font, icon->icon_name, 
+	tmp = ShrinkString(scr->icon_title_font, icon->icon_name,
 			   wPreferences.icon_size-4);
-	w = wTextWidth(scr->icon_title_font->font, tmp, l=strlen(tmp));
+	w = WMWidthOfString(scr->icon_title_font, tmp, l=strlen(tmp));
 
 	if (w > icon->core->width - 4)
 	  x = (icon->core->width - 4) - w;
 	else
 	  x = (icon->core->width - w)/2;
 
- 	wDrawString(icon->core->window, scr->icon_title_font, gc, 
- 		      x, 1 + scr->icon_title_font->y, tmp, l);
+	WMDrawString(scr->wmscreen, icon->core->window, gc,
+		     scr->icon_title_font, x, 1, tmp, l);
 	free(tmp);
     }
 
