@@ -3,6 +3,7 @@
  *  Window Maker window manager
  * 
  *  Copyright (c) 1997, 1998 Alfredo K. Kojima
+ *  Copyright (c) 1997, 1998 Dan Pascu
  * 
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -30,6 +31,7 @@
 #include <stdio.h>
 #include <unistd.h>
 #include <ctype.h>
+#include <nana.h>
 
 #include "WindowMaker.h"
 #include "wcore.h"
@@ -81,12 +83,57 @@ static void menuTitleMouseDown(WCoreWindow *sender, void *data, XEvent *event);
 
 static void menuCloseClick(WCoreWindow *sender, void *data, XEvent *event);
 
+static void updateTexture(WMenu *menu);
 
+#ifndef LITE
+static int saveMenuRecurs(proplist_t menus, WScreen *scr, WMenu *menu);
+static int restoreMenuRecurs(WScreen *scr, proplist_t menus, WMenu *menu, char *path);
+#endif /* !LITE */
 
 static void selectEntry(WMenu *menu, int entry_no);
 static void closeCascade(WMenu *menu);
 
 
+/****** Notification Observers ******/
+
+static void
+appearanceObserver(void *self, WMNotification *notif)
+{
+    WMenu *menu = (WMenu*)self;
+    int flags = (int)WMGetNotificationClientData(notif);
+
+    if (!menu->flags.realized)
+	return;
+    
+    if (WMGetNotificationName(notif) == WNMenuAppearanceSettingsChanged) {
+	if (flags & WFontSettings) {
+	    menu->flags.realized = 0;
+	    wMenuRealize(menu);
+	}
+	if (flags & WTextureSettings) {
+	    if (!menu->flags.brother)
+		updateTexture(menu);
+	}
+	if (flags & (WTextureSettings|WColorSettings)) {
+	    wMenuPaint(menu);
+	}
+    } else if (menu->flags.titled) {
+
+	if (flags & WFontSettings) {
+	    menu->flags.realized = 0;
+	    wMenuRealize(menu);
+	}
+	if (flags & WTextureSettings) {
+	    menu->frame->flags.need_texture_remake = 1;
+	}
+	if (flags & (WColorSettings|WTextureSettings)) {
+	    wFrameWindowPaint(menu->frame);
+	}
+    }
+}
+
+/************************************/
+
 
 /*
  *----------------------------------------------------------------------
@@ -117,17 +164,21 @@ wMenuCreate(WScreen *screen, char *title, int main_menu)
     tmp = (main_menu ? WMMainMenuLevel : WMSubmenuLevel);
 #endif
 
-    flags = WFF_SINGLE_STATE;
+    flags = WFF_SINGLE_STATE|WFF_BORDER;
     if (title) {
 	flags |= WFF_TITLEBAR|WFF_RIGHT_BUTTON;
 	menu->flags.titled = 1;
     }
     menu->frame =
-	wFrameWindowCreate(screen, tmp, 8, 2, 1, 1, flags,
+	wFrameWindowCreate(screen, tmp, 8, 2, 1, 1, &wPreferences.menu_title_clearance, flags,
 			   screen->menu_title_texture, NULL,
-			   screen->menu_title_pixel, &screen->menu_title_gc, 
-			   &screen->menu_title_font);
-
+                           screen->menu_title_pixel,
+#ifdef DRAWSTRING_PLUGIN
+			   W_STRING_MTITLE,
+#endif
+			   &screen->menu_title_gc,
+                           &screen->menu_title_font);
+			   
     menu->frame->core->descriptor.parent = menu;
     menu->frame->core->descriptor.parent_type = WCLASS_MENU;
     menu->frame->core->descriptor.handle_mousedown = menuMouseDown;
@@ -139,7 +190,7 @@ wMenuCreate(WScreen *screen, char *title, int main_menu)
     }
 
     menu->frame->flags.justification = WTJ_LEFT;
-    
+
     menu->frame->rbutton_image = screen->b_pixmaps[WBUT_CLOSE];
 
     menu->entry_no = 0;
@@ -185,7 +236,12 @@ wMenuCreate(WScreen *screen, char *title, int main_menu)
 	menu->brother->flags.brother = 1;
 	menu->brother->brother = menu;
     }
-    
+    WMAddNotificationObserver(appearanceObserver, menu,
+			      WNMenuAppearanceSettingsChanged, menu);
+
+    WMAddNotificationObserver(appearanceObserver, menu,
+			      WNMenuTitleAppearanceSettingsChanged, menu);
+
 
     return menu;
 }
@@ -222,39 +278,6 @@ insertEntry(WMenu *menu, WMenuEntry *entry, int index)
 }
 
 
-void
-wMenuRefresh(WMenu *menu, int flags)
-{
-    int i;
-    
-    if (flags & MR_TEXT_BACK) {
-	menu->frame->flags.need_texture_remake = 1;
-    }
-    
-    if (flags & (MR_RESIZED|MR_TITLE_TEXT)) {
-	menu->flags.realized = 0;
-    }
-
-    wMenuRealize(menu);
-
-    if (menu->flags.titled)
-	wFrameWindowPaint(menu->frame);
-    
-    if (!menu->flags.brother) {
-	if (menu->brother)
-	    wMenuRefresh(menu->brother, flags);
-
-	for (i=0; i < menu->cascade_no; i++) {
-	    if (!menu->cascades[i]->flags.brother)
-		wMenuRefresh(menu->cascades[i], flags);
-	    else
-		wMenuRefresh(menu->cascades[i]->brother, flags);
-	}
-    }
-}
-
-
-
 WMenuEntry*
 wMenuInsertCallback(WMenu *menu, int index, char *text,
 		    void (*callback)(WMenu *menu, WMenuEntry *entry), 
@@ -262,29 +285,15 @@ wMenuInsertCallback(WMenu *menu, int index, char *text,
 {
     WMenuEntry *entry;
 
-#ifdef DEBUG
-    if (!menu) {
-	printf("Passed NULL as menu parameter to wMenuAddCallback() \n");
-	return NULL;
-    }    
-#endif
-
-    assert(menu->flags.brother==0);
     menu->flags.realized = 0;
     menu->brother->flags.realized = 0;
     
     /* reallocate array if it's too small */
     if (menu->entry_no >= menu->alloced_entries) {
 	void *tmp;
-#ifdef DEBUG
-	puts("doing wrealloc()");
-#endif	
+
 	tmp = wrealloc(menu->entries,
 		       sizeof(WMenuEntry)*(menu->alloced_entries+5));
-	if (tmp==NULL) {
-	    wwarning(_("wrealloc() failed while trying to add menu item"));
-	    return NULL;
-	}
 
 	menu->entries = tmp;
 	menu->alloced_entries += 5;
@@ -426,6 +435,90 @@ wMenuRemoveItem(WMenu *menu, int index)
 }
 
 
+static Pixmap
+renderTexture(WMenu *menu)
+{
+    RImage *img;
+    Pixmap pix;
+    int i;
+    RColor light;
+    RColor dark;
+    RColor mid;
+    WScreen *scr = menu->menu->screen_ptr;
+    WTexture *texture = scr->menu_item_texture;
+
+    if (wPreferences.menu_style == MS_NORMAL) {
+	img = wTextureRenderImage(texture, menu->menu->width, 
+				  menu->entry_height, WREL_MENUENTRY);
+    } else {
+	img = wTextureRenderImage(texture, menu->menu->width, 
+				  menu->menu->height+1, WREL_MENUENTRY);
+    }
+    if (!img) {
+	wwarning(_("could not render texture: %s"), 
+		 RMessageForError(RErrorCode));
+
+	return None;
+    }
+
+    if (wPreferences.menu_style == MS_SINGLE_TEXTURE) {
+	light.alpha = 0;
+	light.red = light.green = light.blue = 80;
+    
+	dark.alpha = 255;
+	dark.red = dark.green = dark.blue = 0;
+    
+	mid.alpha = 0;
+	mid.red = mid.green = mid.blue = 40;
+    
+	for (i = 1; i < menu->entry_no; i++) {	    
+	    ROperateLine(img, RSubtractOperation, 0, i*menu->entry_height-2,
+			 menu->menu->width-1, i*menu->entry_height-2, &mid);
+	    
+	    RDrawLine(img, 0, i*menu->entry_height-1,
+		      menu->menu->width-1, i*menu->entry_height-1, &dark);
+
+	    ROperateLine(img, RAddOperation, 0, i*menu->entry_height,
+			 menu->menu->width-1, i*menu->entry_height, 
+			 &light);
+	}
+    }
+    if (!RConvertImage(scr->rcontext, img, &pix)) {
+	wwarning(_("error rendering image:%s"), RMessageForError(RErrorCode));
+    }
+    RDestroyImage(img);
+
+    return pix;
+}
+
+
+static void
+updateTexture(WMenu *menu)
+{
+    WScreen *scr = menu->menu->screen_ptr;
+
+    /* setup background texture */
+    if (scr->menu_item_texture->any.type != WTEX_SOLID) {
+	if (!menu->flags.brother) {
+	    FREE_PIXMAP(menu->menu_texture_data);
+
+	    menu->menu_texture_data = renderTexture(menu);
+
+	    XSetWindowBackgroundPixmap(dpy, menu->menu->window,
+				       menu->menu_texture_data);
+	    XClearWindow(dpy, menu->menu->window);
+
+	    XSetWindowBackgroundPixmap(dpy, menu->brother->menu->window,
+				       menu->menu_texture_data);
+	    XClearWindow(dpy, menu->brother->menu->window);
+	}
+    } else {
+	XSetWindowBackground(dpy, menu->menu->window,
+			     scr->menu_item_texture->any.color.pixel);
+	XClearWindow(dpy, menu->menu->window);
+    }
+}
+
 
 void 
 wMenuRealize(WMenu *menu)
@@ -443,22 +536,22 @@ wMenuRealize(WMenu *menu)
 	brother_done = 0;
     }
 
-    flags = WFF_SINGLE_STATE;
+    flags = WFF_SINGLE_STATE|WFF_BORDER;
     if (menu->flags.titled)
 	flags |= WFF_TITLEBAR|WFF_RIGHT_BUTTON;
 
     wFrameWindowUpdateBorders(menu->frame, flags);
 
     if (menu->flags.titled) {
-	twidth = wTextWidth(scr->menu_title_font->font, menu->frame->title,
-			     strlen(menu->frame->title));
+	twidth = WMWidthOfString(scr->menu_title_font, menu->frame->title,
+				 strlen(menu->frame->title));
         theight = menu->frame->top_width;
-        twidth += theight + theight/2 + (wPreferences.new_style ? 8 : 0);
+        twidth += theight + (wPreferences.new_style ? 16 : 8);
     } else {
 	twidth = 0;
 	theight = 0;
     }
-    eheight = scr->menu_entry_font->height + 6;
+    eheight = WMFontHeight(scr->menu_entry_font) + 6 + wPreferences.menu_text_clearance * 2;
     menu->entry_height = eheight;
     mrwidth = 0;
     mwidth = 0;
@@ -467,7 +560,7 @@ wMenuRealize(WMenu *menu)
 
 	/* search widest text */
 	text = menu->entries[i]->text;
-	width = wTextWidth(scr->menu_entry_font->font, text, strlen(text))+12;
+	width = WMWidthOfString(scr->menu_entry_font, text, strlen(text))+10;
 
 	if (menu->entries[i]->flags.indicator) {
 	    width += MENU_INDICATOR_SPACE;
@@ -479,10 +572,10 @@ wMenuRealize(WMenu *menu)
 	/* search widest text on right */
 	text = menu->entries[i]->rtext;
 	if (text)
-	    rwidth = wTextWidth(scr->menu_entry_font->font, text,
-				 strlen(text)) + 5;
+	    rwidth = WMWidthOfString(scr->menu_entry_font, text, strlen(text))
+		+ 5;
 	else if (menu->entries[i]->cascade>=0)
-	    rwidth = eheight;
+	    rwidth = 16;
 	else
 	    rwidth = 4;
 
@@ -494,34 +587,14 @@ wMenuRealize(WMenu *menu)
     if (mwidth < twidth)
 	mwidth = twidth;
 
+    
     wCoreConfigure(menu->menu, 0, theight, mwidth, menu->entry_no*eheight -1);
 
     wFrameWindowResize(menu->frame, mwidth, menu->entry_no*eheight-1
 		       + menu->frame->top_width + menu->frame->bottom_width);
 
 
-    /* setup background texture */
-    if (scr->menu_item_texture->any.type != WTEX_SOLID) {
-	if (!menu->flags.brother) {
-	    FREE_PIXMAP(menu->menu_texture_data);
-
-	    wTextureRender(scr, scr->menu_item_texture,
-			   &menu->menu_texture_data, menu->menu->width,
-			   menu->entry_height, WREL_MENUENTRY);
-
-	    XSetWindowBackgroundPixmap(dpy, menu->menu->window,
-				       menu->menu_texture_data);
-	    XClearWindow(dpy, menu->menu->window);
-
-	    XSetWindowBackgroundPixmap(dpy, menu->brother->menu->window,
-				       menu->menu_texture_data);
-	    XClearWindow(dpy, menu->brother->menu->window);
-	}
-    } else {
-	XSetWindowBackground(dpy, menu->menu->window,
-			     scr->menu_item_texture->any.color.pixel);
-	XClearWindow(dpy, menu->menu->window);
-    }
+    updateTexture(menu);
 
     menu->flags.realized = 1;
 
@@ -537,6 +610,8 @@ wMenuDestroy(WMenu *menu, int recurse)
 {
     int i;
 
+    WMRemoveNotificationObserver(menu);
+    
     /* remove any pending timers */
     if (menu->timer)
 	WMDeleteTimerHandler(menu->timer);
@@ -556,6 +631,12 @@ wMenuDestroy(WMenu *menu, int recurse)
 
 	    if (menu->entries[i]->rtext)
 	      free(menu->entries[i]->rtext);
+#ifdef USER_MENU
+
+	    if (menu->entries[i]->instances){
+		  PLRelease(menu->entries[i]->instances);
+		}
+#endif /* USER_MENU */
 
 	    if (menu->entries[i]->free_cdata && menu->entries[i]->clientdata) {
 		(*menu->entries[i]->free_cdata)(menu->entries[i]->clientdata);
@@ -595,31 +676,48 @@ wMenuDestroy(WMenu *menu, int recurse)
 }
 
 
+#define F_NORMAL	0
+#define F_TOP		1
+#define F_BOTTOM	2
+#define F_NONE		3
+
 static void
-drawFrame(WScreen *scr, Window win, int y, int w, int h)
+drawFrame(WScreen *scr, Window win, int y, int w, int h, int type)
 {
     XSegment segs[2];
+    int i;
 
-    segs[0].x1 = 0;
-    segs[0].y1 = y;
-    segs[0].x2 = w-1;
-    segs[0].y2 = y;
-    segs[1].x1 = 0;
-    segs[1].y1 = y;
-    segs[1].x2 = 0;
-    segs[1].y2 = y + h - 2;
-    XDrawSegments(dpy, win, scr->menu_item_auxtexture->light_gc, segs, 2);
+    i = 0;
+    segs[i].x1 = segs[i].x2 = w-1;
+    segs[i].y1 = y;
+    segs[i].y2 = y + h - 1;
+    i++;
+    if (type != F_TOP && type != F_NONE) {
+	segs[i].x1 = 1;
+	segs[i].y1 = segs[i].y2 = y + h-2;
+	segs[i].x2 = w-1;
+	i++;
+    }
+    XDrawSegments(dpy, win, scr->menu_item_auxtexture->dim_gc, segs, i);
 
-    XDrawLine(dpy, win, scr->menu_item_auxtexture->dark_gc, 0, y+h-1,
-	      w-1, y+h-1);
+    i = 0;
+    segs[i].x1 = 0;
+    segs[i].y1 = y;
+    segs[i].x2 = 0;
+    segs[i].y2 = y + h - 1;
+    i++;
+    if (type != F_BOTTOM && type != F_NONE) {
+	segs[i].x1 = 0;
+	segs[i].y1 = y;
+	segs[i].x2 = w-1;
+	segs[i].y2 = y;
+	i++;
+    }
+    XDrawSegments(dpy, win, scr->menu_item_auxtexture->light_gc, segs, i);
 
-    segs[0].x1 = 1;
-    segs[0].y1 = segs[0].y2 = y + h-2;
-    segs[0].x2 = w-1;
-    segs[1].x1 = segs[1].x2 = w-1;
-    segs[1].y1 = y + 1;
-    segs[1].y2 = y + h-2;
-    XDrawSegments(dpy, win, scr->menu_item_auxtexture->dim_gc, segs, 2);
+    if (type != F_TOP && type != F_NONE)
+	XDrawLine(dpy, win, scr->menu_item_auxtexture->dark_gc, 0, y+h-1,
+		  w-1, y+h-1);
 }
 
 
@@ -627,6 +725,7 @@ static void
 paintEntry(WMenu *menu, int index, int selected)
 {
     int x, y, w, h, tw;
+    int type;
     GC light, dim, dark, textGC;
     WScreen *scr=menu->frame->screen_ptr;
     Window win = menu->menu->window;
@@ -641,17 +740,28 @@ paintEntry(WMenu *menu, int index, int selected)
     dim = scr->menu_item_auxtexture->dim_gc;
     dark = scr->menu_item_auxtexture->dark_gc;
 
+    if (wPreferences.menu_style == MS_FLAT && menu->entry_no > 1) {
+	if (index == 0)
+	    type = F_TOP;
+	else if (index == menu->entry_no - 1)
+	    type = F_BOTTOM;
+	else
+	    type = F_NONE;
+    } else {
+	type = F_NORMAL;
+    }
+    
     /* paint background */
     if (selected) {
 	XSetForeground(dpy, scr->select_menu_gc, scr->select_pixel);
 	XFillRectangle(dpy, win, scr->select_menu_gc, 1, y+1, w-2, h-3);
 	if (scr->menu_item_texture->any.type == WTEX_SOLID)
-	    drawFrame(scr, win, y, w, h);
+	    drawFrame(scr, win, y, w, h, type);
     } else { 
 	if (scr->menu_item_texture->any.type == WTEX_SOLID) {
 	    XClearArea(dpy, win, 0, y + 1, w - 1, h - 3, False);
 	    /* draw the frame */
-	    drawFrame(scr, win, y, w, h);
+	    drawFrame(scr, win, y, w, h, type);
 	} else {
 	    XClearArea(dpy, win, 0, y, w, h, False);
 	}
@@ -673,9 +783,8 @@ paintEntry(WMenu *menu, int index, int selected)
     if (entry->flags.indicator)
 	x += MENU_INDICATOR_SPACE + 2;
 
-    wDrawString(win, scr->menu_entry_font, 
-		  textGC, x, 3+y+scr->menu_entry_font->y, entry->text, 
-		  strlen(entry->text));
+    WMDrawString(scr->wmscreen, win, textGC, scr->menu_entry_font, 
+		 x, 3 + y + wPreferences.menu_text_clearance, entry->text, strlen(entry->text));
 
     if (entry->cascade>=0) {
 	/* draw the cascade indicator */
@@ -728,15 +837,15 @@ paintEntry(WMenu *menu, int index, int selected)
     /* draw right text */
     
     if (entry->rtext && entry->cascade<0) {
-	tw = wTextWidth(scr->menu_entry_font->font, entry->rtext,
-			 strlen(entry->rtext));
+	tw = WMWidthOfString(scr->menu_entry_font, entry->rtext,
+			     strlen(entry->rtext));
 
-	wDrawString(win, scr->menu_entry_font, textGC, w-6-tw, 
-		    3+y+scr->menu_entry_font->y, entry->rtext,
-		    strlen(entry->rtext));
+	WMDrawString(scr->wmscreen, win, textGC, scr->menu_entry_font, w-6-tw, 
+		     y + 3 + wPreferences.menu_text_clearance, entry->rtext, strlen(entry->rtext));
     }
 }
 
+
 static void
 move_menus(WMenu *menu, int x, int y)
 {
@@ -1503,6 +1612,23 @@ getScrollAmount(WMenu *menu, int *hamount, int *vamount)
     
     getPointerPosition(scr, &xroot, &yroot);
 
+#ifdef VIRTUAL_DESKTOP
+    if (wPreferences.vedge_thickness) {
+        if (xroot <= wPreferences.vedge_thickness + 1 && menuX1 < wPreferences.vedge_thickness) {
+            /* scroll to the right */
+            *hamount = MIN(MENU_SCROLL_STEP, abs(menuX1));
+
+        } else if (xroot >= screenW-2-wPreferences.vedge_thickness && menuX2 > screenW-1-wPreferences.vedge_thickness) {
+            /* scroll to the left */
+            *hamount = MIN(MENU_SCROLL_STEP, abs(menuX2-screenW-1));
+
+            if (*hamount==0)
+                *hamount = 1;
+
+            *hamount = -*hamount;
+        }
+    } else
+#endif
 
     if (xroot <= 1 && menuX1 < 0) {
 	/* scroll to the right */
@@ -1626,6 +1752,23 @@ isPointNearBoder(WMenu *menu, int x, int y)
     return flag;
 }
 
+ 
+typedef struct _delay {
+    WWindow *wwin;
+    WMenu *menu;
+    int ox,oy;
+} _delay;
+
+
+static void
+_leaving(_delay *dl)
+{
+    wMenuMove(dl->menu, dl->ox, dl->oy, True);
+    dl->menu->jump_back=NULL;
+    dl->menu->menu->screen_ptr->flags.jump_back_pending = 0;
+    free(dl);
+}
+
 
 void
 wMenuScroll(WMenu *menu, XEvent *event)
@@ -1638,10 +1781,10 @@ wMenuScroll(WMenu *menu, XEvent *event)
     int old_frame_x = omenu->frame_x;
     int old_frame_y = omenu->frame_y;
     XEvent ev;
+    
+    if (omenu->jump_back)
+        WMDeleteTimerWithClientData(omenu->jump_back);
 
-#ifdef DEBUG
-    puts("Entering menu Scroll");
-#endif
 
     if ((/*omenu->flags.buttoned &&*/ !wPreferences.wrap_menus)
         || omenu->flags.app_menu) {
@@ -1661,9 +1804,9 @@ wMenuScroll(WMenu *menu, XEvent *event)
 
         WMNextEvent(dpy, &ev);
         switch (ev.type) {
-        case EnterNotify:
+	 case EnterNotify:
             WMHandleEvent(&ev);
-        case MotionNotify:
+	 case MotionNotify:
             x = (ev.type==MotionNotify) ? ev.xmotion.x_root : ev.xcrossing.x_root;
             y = (ev.type==MotionNotify) ? ev.xmotion.y_root : ev.xcrossing.y_root;
 
@@ -1726,12 +1869,19 @@ wMenuScroll(WMenu *menu, XEvent *event)
 	menu->timer = NULL;
     }
 
-    if (jump_back)
-        wMenuMove(omenu, old_frame_x, old_frame_y, True);
-
-#ifdef DEBUG
-    puts("Leaving menu Scroll");
-#endif
+    if (jump_back) {
+        _delay *delayer;
+        if (!omenu->jump_back) {
+            delayer=wmalloc(sizeof(_delay));
+            delayer->menu=omenu;
+            delayer->ox=old_frame_x;
+            delayer->oy=old_frame_y;
+            omenu->jump_back=delayer;
+            scr->flags.jump_back_pending = 1;
+        }
+        else delayer = omenu->jump_back;
+        WMAddTimerHandler(MENU_JUMP_BACK_DELAY,(WMCallback*)_leaving, delayer);
+    }
 }
 
 
@@ -1791,8 +1941,7 @@ menuMouseDown(WObjDescriptor *desc, XEvent *event)
         smenu = parentMenu(menu);
         old_frame_x = smenu->frame_x;
         old_frame_y = smenu->frame_y;
-    }
-    else if (event->xbutton.window == menu->frame->core->window) {
+    } else if (event->xbutton.window == menu->frame->core->window) {
         /* This is true if the menu was launched with right click on root window */
         delayed_select = 1;
         d_data.delayed_select = &delayed_select;
@@ -1800,6 +1949,11 @@ menuMouseDown(WObjDescriptor *desc, XEvent *event)
         d_data.magic = WMAddTimerHandler(wPreferences.dblclick_time,
                                         delaySelection, &d_data);
     }
+    
+    if (menu->flags.inside_handler) {
+	return;
+    }
+    menu->flags.inside_handler = 1;
 
     wRaiseFrame(menu->frame->core);
     
@@ -1814,7 +1968,7 @@ menuMouseDown(WObjDescriptor *desc, XEvent *event)
     }
 
     if (menu->flags.editing) {
-	return;
+	goto byebye;
     }
     entry_no = getEntryAt(menu, x, y);
     if (entry_no>=0) {
@@ -1823,9 +1977,9 @@ menuMouseDown(WObjDescriptor *desc, XEvent *event)
 	if (!close_on_exit && (bev->state & ControlMask) && smenu
 	    && entry->flags.editable) {
 	    editEntry(smenu, entry);
-	    return;
+	    goto byebye;
 	} else if (bev->state & ControlMask) {
-	    return;
+	    goto byebye;
 	}
 
 	if (entry->flags.enabled && entry->cascade>=0 && menu->cascades) {
@@ -1856,7 +2010,7 @@ menuMouseDown(WObjDescriptor *desc, XEvent *event)
     while (!done) {
 	int x, y;
 
-	XAllowEvents(dpy, SyncPointer, CurrentTime);
+	XAllowEvents(dpy, AsyncPointer|SyncPointer, CurrentTime);
 
 	WMMaskEvent(dpy, ExposureMask|ButtonMotionMask|ButtonReleaseMask
 		    |ButtonPressMask, &ev);
@@ -2048,8 +2202,7 @@ menuMouseDown(WObjDescriptor *desc, XEvent *event)
 	}
     }
 
-    if (((WMenu*)desc->parent)->flags.brother || close_on_exit
-	|| !smenu)
+    if (((WMenu*)desc->parent)->flags.brother || close_on_exit || !smenu)
 	closeCascade(desc->parent);
 
     /* close the cascade windows that should not remain opened */
@@ -2057,6 +2210,9 @@ menuMouseDown(WObjDescriptor *desc, XEvent *event)
 
     if (!wPreferences.wrap_menus)
         wMenuMove(parentMenu(desc->parent), old_frame_x, old_frame_y, True);
+
+byebye:    
+    ((WMenu*)desc->parent)->flags.inside_handler = 0;
 }
 
 
@@ -2077,14 +2233,15 @@ wMenuMove(WMenu *menu, int x, int y, int submenus)
 	
 	if (i>=0 && menu->cascades) {
 	    submenu = menu->cascades[i];
-	    if (submenu->flags.mapped && !submenu->flags.buttoned)
-	      if (wPreferences.align_menus) {
-		  wMenuMove(submenu, x + MENUW(menu), y, submenus);
-	      } else {
-		  wMenuMove(submenu, x+ MENUW(menu),
-			    y + submenu->entry_height*menu->selected_entry,
-			    submenus);
-	      }
+	    if (submenu->flags.mapped && !submenu->flags.buttoned) {
+		if (wPreferences.align_menus) {
+		    wMenuMove(submenu, x + MENUW(menu), y, submenus);
+		} else {
+		    wMenuMove(submenu, x+ MENUW(menu),
+			      y + submenu->entry_height*menu->selected_entry,
+			      submenus);
+		}
+	    }
 	}
     }
     if (submenus<0 && menu->parent!=NULL && menu->parent->flags.mapped &&
@@ -2154,10 +2311,6 @@ menuTitleMouseDown(WCoreWindow *sender, void *data, XEvent *event)
     int i, lower;
     Bool started;
 
-#ifdef DEBUG
-    printf("Moving menu\n");
-#endif
-
     /* can't touch the menu copy */
     if (menu->flags.brother)
 	return;
@@ -2241,9 +2394,6 @@ menuTitleMouseDown(WCoreWindow *sender, void *data, XEvent *event)
 	 case ButtonRelease:
 	    if (ev.xbutton.button != event->xbutton.button)
 		break;
-#ifdef DEBUG
-	    printf("End menu move\n");
-#endif
 	    XUngrabPointer(dpy, CurrentTime);
 	    return;
 
@@ -2287,46 +2437,47 @@ menuCloseClick(WCoreWindow *sender, void *data, XEvent *event)
 }
 
 
+static void
+saveMenuInfo(proplist_t dict, WMenu *menu, proplist_t key)
+{
+    proplist_t value, list;
+    char buffer[256];
+
+    sprintf(buffer, "%i,%i", menu->frame_x, menu->frame_y);
+    value = PLMakeString(buffer);
+    list = PLMakeArrayFromElements(value, NULL);
+    if (menu->flags.lowered)
+	PLAppendArrayElement(list, PLMakeString("lowered"));
+    PLInsertDictionaryEntry(dict, key, list);
+    PLRelease(value);
+    PLRelease(list);
+}
+
+
 void
 wMenuSaveState(WScreen *scr)
 {
-    proplist_t menus, key, value;
+    proplist_t menus, key;
     int save_menus = 0;
-    char buffer[256];
 
     menus = PLMakeDictionaryFromEntries(NULL, NULL, NULL);
 
 #ifndef LITE
-    if (scr->root_menu && scr->root_menu->flags.buttoned) {
-        sprintf(buffer, "%i,%i", scr->root_menu->frame_x,
-                                 scr->root_menu->frame_y);
-        key = PLMakeString("RootMenu");
-        value = PLMakeString(buffer);
-        PLInsertDictionaryEntry(menus, key, value);
-        PLRelease(key);
-        PLRelease(value);
+    if (scr->switch_menu && scr->switch_menu->flags.buttoned) {
+	key = PLMakeString("SwitchMenu");
+	saveMenuInfo(menus, scr->switch_menu, key);
+	PLRelease(key);
         save_menus = 1;
     }
 
-    if (scr->switch_menu && scr->switch_menu->flags.buttoned) {
-        sprintf(buffer, "%i,%i", scr->switch_menu->frame_x,
-                                 scr->switch_menu->frame_y);
-        key = PLMakeString("SwitchMenu");
-        value = PLMakeString(buffer);
-        PLInsertDictionaryEntry(menus, key, value);
-        PLRelease(key);
-        PLRelease(value);
-        save_menus = 1;
-    }
+    if (saveMenuRecurs(menus, scr, scr->root_menu))
+	save_menus = 1;
+
 #endif /* !LITE */
     if (scr->workspace_menu && scr->workspace_menu->flags.buttoned) {
-        sprintf(buffer, "%i,%i", scr->workspace_menu->frame_x,
-                                 scr->workspace_menu->frame_y);
         key = PLMakeString("WorkspaceMenu");
-        value = PLMakeString(buffer);
-        PLInsertDictionaryEntry(menus, key, value);
+	saveMenuInfo(menus, scr->workspace_menu, key);
         PLRelease(key);
-        PLRelease(value);
         save_menus = 1;
     }
     
@@ -2339,53 +2490,133 @@ wMenuSaveState(WScreen *scr)
 }
 
 
+#ifndef LITE
+
+static Bool
+getMenuPath(WMenu *menu, char *buffer, int bufSize)
+{
+    Bool ok = True;
+    int len = 0;
+
+    if (!menu->flags.titled || !menu->frame->title[0])
+	return False;
+
+    len = strlen(menu->frame->title);
+    if (len >= bufSize)
+	return False;
+
+    if (menu->parent) {
+	ok = getMenuPath(menu->parent, buffer, bufSize - len - 1);
+	if (!ok)
+	    return False;
+    }
+
+    strcat(buffer, "\\");
+    strcat(buffer, menu->frame->title);
+
+    return True;
+}
+
+
+static Bool
+saveMenuRecurs(proplist_t menus, WScreen *scr, WMenu *menu)
+{
+    proplist_t key;
+    int save_menus = 0, i;
+    char buffer[512];
+    Bool ok = True;
+
+    
+    if (menu->flags.brother)
+	menu = menu->brother;
+
+    if (menu->flags.buttoned && menu != scr->switch_menu) {
+
+	buffer[0] = '\0';
+	ok = getMenuPath(menu, buffer, 510);
+
+	if (ok) {
+	    key = PLMakeString(buffer);
+	    saveMenuInfo(menus, menu, key);
+	    PLRelease(key);	    
+	    save_menus = 1;
+	}
+    }
+
+    if (ok) {
+	for (i = 0; i < menu->cascade_no; i++) {
+	    if (saveMenuRecurs(menus, scr, menu->cascades[i]))
+		save_menus = 1;
+	}
+    }
+    return save_menus;
+}
+#endif /* !LITE */
+
+
 #define COMPLAIN(key) wwarning(_("bad value in menus state info:%s"), key)
 
 
+static Bool
+getMenuInfo(proplist_t info, int *x, int *y, Bool *lowered)
+{
+    proplist_t pos;
+    
+    *lowered = False;
+    
+    if (PLIsArray(info)) {
+	proplist_t flags;
+	pos = PLGetArrayElement(info, 0);
+	flags = PLGetArrayElement(info, 1);
+	if (flags != NULL && PLIsString(flags) && PLGetString(flags) != NULL
+	    && strcmp(PLGetString(flags), "lowered") == 0) {
+	    *lowered = True;
+	}
+    } else {
+	pos = info;
+    }
+
+    if (pos != NULL && PLIsString(pos)) {
+	if (sscanf(PLGetString(pos), "%i,%i", x, y)!=2)
+	    COMPLAIN("Position");
+    } else {
+	COMPLAIN("(position, flags...)");
+	return False;
+    }
+
+    return True;
+}
+
+
 static int
 restoreMenu(WScreen *scr, proplist_t menu, int which)
 {
-    int i, x, y;
+    int x, y;
+    Bool lowered = False;
     WMenu *pmenu = NULL;
-    
-    if (!menu)
-        return False;
-    
-    if (!PLIsString(menu)) {
-        COMPLAIN("Position");
-        return False;
-    }
 
-    if (sscanf(PLGetString(menu), "%i,%i", &x, &y)!=2)
-        COMPLAIN("Position");
+    if (!menu)
+	return False;
+    
+    if (!getMenuInfo(menu, &x, &y, &lowered))
+	return False;
+
 
 #ifndef LITE
-    if (which & WSS_ROOTMENU) {
-        OpenRootMenu(scr, x, y, False);
-        pmenu = scr->root_menu;
-    } else if (which & WSS_SWITCHMENU) {
+    if (which & WSS_SWITCHMENU) {
         OpenSwitchMenu(scr, x, y, False);
         pmenu = scr->switch_menu;
-    } else
-#endif /* !LITE */
-	if (which & WSS_WSMENU) {
-        OpenWorkspaceMenu(scr, x, y);
-        pmenu = scr->workspace_menu;
-        if (pmenu->parent) {
-            /* make parent map the copy in place of the original */
-            for (i=0; i<pmenu->parent->cascade_no; i++) {
-                if (pmenu->parent->cascades[i] == pmenu) {
-                    pmenu->parent->cascades[i] = pmenu->brother;
-                    break;
-                }
-            }
-        }
     }
+#endif /* !LITE */
 
     if (pmenu) {
         int width = MENUW(pmenu);
         int height = MENUH(pmenu);
         
+	if (lowered) {
+	    changeMenuLevels(pmenu, True);
+	}
+	
         x = (x < -width) ? 0 : x;
         x = (x > scr->scr_width) ? scr->scr_width - width : x;
         y = (y < 0) ? 0 : y;
@@ -2399,10 +2630,70 @@ restoreMenu(WScreen *scr, proplist_t menu, int which)
 }
 
 
+#ifndef LITE
+static int
+restoreMenuRecurs(WScreen *scr, proplist_t menus, WMenu *menu, char *path)
+{
+    proplist_t key, entry;
+    char buffer[512];
+    int i, x, y, res;
+    Bool lowered;
+
+    if (strlen(path) + strlen(menu->frame->title) > 510)
+	return False;
+
+    sprintf(buffer, "%s\\%s", path, menu->frame->title);
+    key = PLMakeString(buffer);
+    entry = PLGetDictionaryEntry(menus, key);
+    res = False;
+
+    if (entry && getMenuInfo(entry, &x, &y, &lowered)) {
+
+	if (!menu->flags.mapped) {
+	    int width = MENUW(menu);
+	    int height = MENUH(menu);
+
+	    wMenuMapAt(menu, x, y, False);
+
+	    if (menu->parent) {
+		/* make parent map the copy in place of the original */
+		for (i=0; i<menu->parent->cascade_no; i++) {
+		    if (menu->parent->cascades[i] == menu) {
+			menu->parent->cascades[i] = menu->brother;
+			break;
+		    }
+		}
+	    }
+	    if (lowered) {
+		changeMenuLevels(menu, True);
+	    }
+	    x = (x < -width) ? 0 : x;
+	    x = (x > scr->scr_width) ? scr->scr_width - width : x;
+	    y = (y < 0) ? 0 : y;
+	    y = (y > scr->scr_height) ? scr->scr_height - height : y;
+	    wMenuMove(menu, x, y, True);
+	    menu->flags.buttoned = 1;
+	    wFrameWindowShowButton(menu->frame, WFF_RIGHT_BUTTON);
+	    res = True;
+	}
+    }
+    
+    PLRelease(key);
+
+    for (i=0; i<menu->cascade_no; i++) {
+	if (restoreMenuRecurs(scr, menus, menu->cascades[i], buffer) != False)
+	    res = True;
+    }
+
+    return res;
+}
+#endif /* !LITE */
+
+
 void
 wMenuRestoreState(WScreen *scr)
 {
-    proplist_t menus, menu, key, rkey, skey, wkey;
+    proplist_t menus, menu, key, skey;
 
     key = PLMakeString("Menus");
     menus = PLGetDictionaryEntry(scr->session_state, key);
@@ -2413,19 +2704,20 @@ wMenuRestoreState(WScreen *scr)
 
     /* restore menus */
 
-    rkey = PLMakeString("RootMenu");
     skey = PLMakeString("SwitchMenu");
-    wkey = PLMakeString("WorkspaceMenu");
-    menu = PLGetDictionaryEntry(menus, rkey);
-    restoreMenu(scr, menu, WSS_ROOTMENU);
     menu = PLGetDictionaryEntry(menus, skey);
-    restoreMenu(scr, menu, WSS_SWITCHMENU);
-    menu = PLGetDictionaryEntry(menus, wkey);
-    restoreMenu(scr, menu, WSS_WSMENU);
-
-    PLRelease(rkey);
     PLRelease(skey);
-    PLRelease(wkey);
+    restoreMenu(scr, menu, WSS_SWITCHMENU);
+
+#ifndef LITE
+    if (!scr->root_menu) {
+        OpenRootMenu(scr, scr->scr_width*2, 0, False);
+        wMenuUnmap(scr->root_menu);
+    }
+    restoreMenuRecurs(scr, menus, scr->root_menu, "");
+#endif /* !LITE */
+
+    PLRelease(skey);
 }
 
 
@@ -2458,8 +2750,7 @@ OpenWorkspaceMenu(WScreen *scr, int x, int y)
                 wRaiseFrame(menu->frame->core);
 		wMenuMapCopyAt(menu, x, y);
             }
-        }
-        else {
+        } else {
             wMenuMapAt(menu, x, y, False);
         }
     }
