@@ -3,6 +3,7 @@
  *  Window Maker window manager
  * 
  *  Copyright (c) 1997, 1998 Alfredo K. Kojima
+ *  Copyright (c) 1998       Dan Pascu
  * 
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -24,7 +25,6 @@
 
 #ifndef LITE
 
-#include <assert.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -54,7 +54,6 @@
 #include "xmodifier.h"
 #include <proplist.h>
 
-#include "list.h"
 
 
 extern char *Locale;
@@ -113,6 +112,7 @@ static Shortcut *shortcutList = NULL;
  * REFRESH - forces the desktop to be repainted
  * EXIT [QUICK] - exit the window manager [without confirmation]
  * EXEC <program> - execute an external program
+ * SHEXEC <command> - execute a shell command
  * WORKSPACE_MENU - places the workspace submenu
  * ARRANGE_ICONS
  * RESTART [<window manager>] - restarts the window manager
@@ -264,7 +264,8 @@ static void
 restartCommand(WMenu *menu, WMenuEntry *entry)
 {
     Shutdown(WSRestartPreparationMode);
-    Restart((char*)entry->clientdata);
+    Restart((char*)entry->clientdata, False);
+    Restart(NULL, True);
 }
 
 
@@ -600,7 +601,7 @@ static void
 separateCommand(char *line, char ***file, char **command)
 {
     char *token, *tmp = line;
-    LinkedList *list = NULL;
+    WMBag *bag = WMCreateBag(4);
     int count, i;
 
     *file = NULL;
@@ -615,20 +616,19 @@ separateCommand(char *line, char ***file, char **command)
                     wwarning(_("%s: missing command"), line);
                 break;
             }
-            list = list_cons(token, list);
+	    WMPutInBag(bag, token);
         }
     } while (token!=NULL && tmp!=NULL);
 
-    count = list_length(list);
+    count = WMGetBagItemCount(bag);
     if (count>0) {
         *file = wmalloc(sizeof(char*)*(count+1));
-        i = count;
         (*file)[count] = NULL;
-        while (list!=NULL) {
-            (*file)[--i] = list->head;
-            list_remove_head(&list);
+        for (i = 0; i < count; i++) {
+            (*file)[i] = WMGetFromBag(bag, i);
         }
     }
+    WMFreeBag(bag);
 }
 
 
@@ -644,7 +644,7 @@ constructMenu(WMenu *menu, WMenuEntry *entry)
     time_t last=0;
     
     separateCommand((char*)entry->clientdata, &path, &cmd);
-    if (!path || *path==NULL || **path==0) {
+    if (path == NULL || *path==NULL || **path==0) {
 	wwarning(_("invalid OPEN_MENU specification: %s"),
 		 (char*)entry->clientdata);
 	return;
@@ -805,7 +805,18 @@ addMenuEntry(WMenu *menu, char *title, char *shortcut, char *command,
 	    wwarning(_("%s:missing parameter for menu command \"%s\""),
 		     file_name, command);
 	else {
-	    entry = wMenuAddCallback(menu, title, execCommand, wstrdup(params));
+	    entry = wMenuAddCallback(menu, title, execCommand, 
+				     wstrappend("exec ", params));
+	    entry->free_cdata = free;
+	    shortcutOk = True;
+	}
+    } else if (strcmp(command, "SHEXEC")==0) {
+	if (!params)
+	    wwarning(_("%s:missing parameter for menu command \"%s\""),
+		     file_name, command);
+	else {
+	    entry = wMenuAddCallback(menu, title, execCommand,
+				     wstrdup(params));
 	    entry->free_cdata = free;
 	    shortcutOk = True;
 	}
@@ -1256,9 +1267,12 @@ typedef struct {
 
 
 static int
-myCompare(dir_data *d1, dir_data *d2)
+myCompare(const void *d1, const void *d2)
 {
-    return strcmp(d1->name, d2->name);
+    dir_data *p1 = *(dir_data**)d1;
+    dir_data *p2 = *(dir_data**)d2;
+
+    return strcmp(p1->name, p2->name);
 }
 
 
@@ -1290,11 +1304,15 @@ readMenuDirectory(WScreen *scr, char *title, char **path, char *command)
     struct stat stat_buf;
     WMenu *menu=NULL;
     char *buffer;
-    LinkedList *dirs = NULL, *files = NULL;
+    WMBag *dirs = NULL, *files = NULL;
     int length, i, have_space=0;
     dir_data *data;
     int stripExtension = 0;
 
+    
+    dirs = WMCreateBag(16);
+    files = WMCreateBag(16);
+    
     i=0;
     while (path[i]!=NULL) {
 	if (strcmp(path[i], "-noext")==0) {
@@ -1318,7 +1336,7 @@ readMenuDirectory(WScreen *scr, char *title, char **path, char *command)
 	    if (dentry->d_name[0] == '.')
 		continue;
 
-            buffer = wmalloc(strlen(path[i])+strlen(dentry->d_name)+4);
+            buffer = malloc(strlen(path[i])+strlen(dentry->d_name)+4);
             if (!buffer) {
                 wsyserror(_("out of memory while constructing directory menu %s"),
                          path[i]);
@@ -1347,7 +1365,7 @@ readMenuDirectory(WScreen *scr, char *title, char **path, char *command)
 			data->name = wstrdup(dentry->d_name);
 			data->index = i;
 
-                        list_insert_sorted(data, &dirs, (int(*)())myCompare);
+			WMPutInBag(dirs, data);
                     }
                 } else if (S_ISREG(stat_buf.st_mode) || isFilePack) {
                     /* Hack because access always returns X_OK success for user root */
@@ -1360,7 +1378,7 @@ readMenuDirectory(WScreen *scr, char *title, char **path, char *command)
 			data->name = wstrdup(dentry->d_name);
 			data->index = i;
 
-                        list_insert_sorted(data, &files, (int(*)())myCompare);
+			WMPutInBag(files, data);
                     }
                 }
             }
@@ -1370,37 +1388,47 @@ readMenuDirectory(WScreen *scr, char *title, char **path, char *command)
         closedir(dir);
         i++;
     }
-
-    if (!dirs && !files)
+    
+    if (!WMGetBagItemCount(dirs) && !WMGetBagItemCount(files)) {
+	WMFreeBag(dirs);
+	WMFreeBag(files);
         return NULL;
+    }
+
+    WMSortBag(dirs, myCompare);
+    WMSortBag(files, myCompare);
 
     menu = wMenuCreate(scr, title, False);
     menu->on_destroy = removeShortcutsForMenu;
-    
-    while (dirs != NULL) {
+
+    for (i = 0; i < WMGetBagItemCount(dirs); i++) {
         /* New directory. Use same OPEN_MENU command that was used
          * for the current directory. */
-        dir_data *d = (dir_data*)dirs->head;
+        dir_data *d = (dir_data*)WMGetFromBag(dirs, i);
 
         length = strlen(path[d->index])+strlen(d->name)+6;
+	if (stripExtension)
+	    length += 7;
         if (command)
             length += strlen(command) + 6;
-        buffer = wmalloc(length);
+        buffer = malloc(length);
         if (!buffer) {
             wsyserror(_("out of memory while constructing directory menu %s"),
                      path[d->index]);
             break;
         }
+	
+	buffer[0] = '\0';
+        if (stripExtension)
+            strcat(buffer, "-noext ");
 
         have_space = strchr(path[d->index], ' ')!=NULL ||
                      strchr(d->name, ' ')!=NULL;
-        if (have_space) {
-            buffer[0] = '"';
-            buffer[1] = 0;
-            strcat(buffer, path[d->index]);
-        } else {
-            strcpy(buffer, path[d->index]);
-        }
+
+	if (have_space)
+            strcat(buffer, "\"");
+	strcat(buffer, path[d->index]);
+
         strcat(buffer, "/");
         strcat(buffer, d->name);
         if (have_space)
@@ -1413,23 +1441,20 @@ readMenuDirectory(WScreen *scr, char *title, char **path, char *command)
         addMenuEntry(menu, d->name, NULL, "OPEN_MENU", buffer, path[d->index]);
 
         free(buffer);
-        if (dirs->head) {
-            if (d->name)
-                free(d->name);
-            free(dirs->head);
-        }
-        list_remove_head(&dirs);
+	if (d->name)
+	    free(d->name);
+	free(d);
     }
 
-    while (files != NULL) {
+    for (i = 0; i < WMGetBagItemCount(files); i++) {
         /* executable: add as entry */
-        dir_data *f = (dir_data*) files->head;;
+        dir_data *f = (dir_data*)WMGetFromBag(files, i);
 
         length = strlen(path[f->index])+strlen(f->name)+6;
         if (command)
             length += strlen(command);
 
-        buffer = wmalloc(length);
+        buffer = malloc(length);
         if (!buffer) {
             wsyserror(_("out of memory while constructing directory menu %s"),
                      path[f->index]);
@@ -1463,16 +1488,16 @@ readMenuDirectory(WScreen *scr, char *title, char **path, char *command)
 	    if (ptr && ptr!=f->name)
 		*ptr = 0;
 	}
-        addMenuEntry(menu, f->name, NULL, "EXEC", buffer, path[f->index]);
+        addMenuEntry(menu, f->name, NULL, "SHEXEC", buffer, path[f->index]);
 
         free(buffer);
-        if (files->head) {
-            if (f->name)
-                free(f->name);
-            free(files->head);
-        }
-        list_remove_head(&files);
+	if (f->name)
+	    free(f->name);
+	free(f);
     }
+    
+    WMFreeBag(files);
+    WMFreeBag(dirs);
 
     return menu;
 }
@@ -1708,8 +1733,7 @@ OpenRootMenu(WScreen *scr, int x, int y, int keyboard)
 	    if (keyboard)
 		wMenuMapAt(menu, 0, 0, True);
 	    else
-		wMenuMapCopyAt(menu, x-menu->frame->core->width/2, 
-			       y-menu->frame->top_width/2);
+		wMenuMapCopyAt(menu, x-menu->frame->core->width/2, y);
 	}
 	return;
     }
@@ -1739,9 +1763,9 @@ OpenRootMenu(WScreen *scr, int x, int y, int keyboard)
 	/* menu hasn't changed or could not be read */
 	if (!scr->root_menu) {
 	    wMessageDialog(scr, _("Error"),
-			  _("The applications menu could not be loaded."
-			    "Look at the console output for a detailed"
-			    "description of the errors"),
+			  _("The applications menu could not be loaded. "
+			    "Look at the console output for a detailed "
+			    "description of the errors."),
 			   _("OK"), NULL, NULL);
 
 	    menu = makeDefaultMenu(scr);
@@ -1755,8 +1779,7 @@ OpenRootMenu(WScreen *scr, int x, int y, int keyboard)
 	scr->root_menu = menu;
     }
     if (menu) {
-	wMenuMapAt(menu, x-menu->frame->core->width/2, y-menu->frame->top_width/2,
-		   keyboard);
+	wMenuMapAt(menu, x-menu->frame->core->width/2, y, keyboard);
     }
     
     if (scr->flags.root_menu_changed_shortcuts)
