@@ -32,6 +32,9 @@
 #ifdef SHAPE
 #include <X11/extensions/shape.h>
 #endif
+#ifdef KEEP_XKB_LOCK_STATUS     
+#include <X11/XKBlib.h>         
+#endif /* KEEP_XKB_LOCK_STATUS */
 
 #include <wraster.h>
 
@@ -49,6 +52,7 @@
 #include "workspace.h"
 #include "session.h"
 #include "balloon.h"
+#include "geomview.h"
 #ifdef KWM_HINTS
 # include "kwm.h"
 #endif
@@ -85,6 +89,10 @@ extern Atom _XA_WINDOWMAKER_NOTICEBOARD;
 
 extern int wScreenCount;
 
+#ifdef KEEP_XKB_LOCK_STATUS     
+extern int wXkbSupported;
+#endif
+
 extern WDDomain *WDWindowMaker;
 
 
@@ -172,6 +180,25 @@ allocButtonPixmaps(WScreen *scr)
     if (pix)
       pix->shared = 1;
     scr->b_pixmaps[WBUT_ICONIFY] = pix;
+#ifdef XKB_BUTTON_HINT
+    pix = wPixmapCreateFromXPMData(scr, PRED_XKBGROUP1_XPM);
+    if (pix)
+      pix->shared = 1;
+    scr->b_pixmaps[WBUT_XKBGROUP1] = pix;
+    pix = wPixmapCreateFromXPMData(scr, PRED_XKBGROUP2_XPM);
+    if (pix)
+      pix->shared = 1;
+    scr->b_pixmaps[WBUT_XKBGROUP2] = pix;
+    pix = wPixmapCreateFromXPMData(scr, PRED_XKBGROUP3_XPM);
+    if (pix)
+      pix->shared = 1;
+    scr->b_pixmaps[WBUT_XKBGROUP3] = pix;
+    pix = wPixmapCreateFromXPMData(scr, PRED_XKBGROUP4_XPM);
+    if (pix)
+      pix->shared = 1;
+    scr->b_pixmaps[WBUT_XKBGROUP4] = pix;
+#endif
+
     
     pix = wPixmapCreateFromXPMData(scr, PRED_KILL_XPM);
     if (pix)
@@ -180,7 +207,6 @@ allocButtonPixmaps(WScreen *scr)
 }
 
 
-
 static void
 draw_dot(WScreen *scr, Drawable d, int x, int y, GC gc)
 {
@@ -361,7 +387,13 @@ allocGCs(WScreen *scr)
     /* misc drawing GC */
     gcv.graphics_exposures = False;
     gcm = GCGraphicsExposures;
-    scr->draw_gc = XCreateGC(dpy, scr->w_win, gcm, &gcv);    
+    scr->draw_gc = XCreateGC(dpy, scr->w_win, gcm, &gcv);
+
+    assert (scr->stipple_bitmap!=None);
+    
+	
+    /* mono GC */
+    scr->mono_gc = XCreateGC(dpy, scr->stipple_bitmap, gcm, &gcv);
 }
 
 
@@ -425,7 +457,8 @@ createPixmaps(WScreen *scr)
     image = wDefaultGetImage(scr, "Logo", "WMPanel");
 
     if (!image) {
-	wwarning(_("could not load logo image for panels"));
+	wwarning(_("could not load logo image for panels: %s"),
+		 RMessageForError(RErrorCode));
     } else {
 	if (!RConvertImageMask(scr->rcontext, image, &p, &m, 128)) {
 	    wwarning(_("error making logo image for panel:%s"), RMessageForError(RErrorCode));
@@ -439,9 +472,7 @@ createPixmaps(WScreen *scr)
 	RDestroyImage(image);
     }
 
-    if (!wPreferences.flags.nodock || !wPreferences.flags.noclip) {
-	scr->dock_dots = make3Dots(scr);
-    }
+    scr->dock_dots = make3Dots(scr);
 
     /* titlebar button pixmaps */
     allocButtonPixmaps(scr);
@@ -468,29 +499,6 @@ createInternalWindows(WScreen *scr)
 {
     int vmask;
     XSetWindowAttributes attribs;
-
-    /* window for displaying geometry information during resizes and moves */
-    vmask = CWBorderPixel|CWBackPixmap|CWBackPixel|CWCursor|CWSaveUnder|CWOverrideRedirect;
-    attribs.border_pixel = 0;
-    attribs.save_under = True;
-    attribs.override_redirect = True;    
-    attribs.cursor = wCursor[WCUR_DEFAULT];
-    attribs.background_pixmap = None;
-    if (scr->resizebar_texture[0])
-	attribs.background_pixel = scr->resizebar_texture[0]->normal.pixel;
-    else
-	attribs.background_pixel = scr->light_pixel;
-    vmask |= CWColormap;
-    attribs.colormap = scr->w_colormap;
-
-    wGetGeometryWindowSize(scr, &scr->geometry_display_width,
-			   &scr->geometry_display_height);
-    scr->geometry_display =
-      XCreateWindow(dpy, scr->root_win, 1, 1,
-		    scr->geometry_display_width,
-		    scr->geometry_display_height,
-		    1, scr->w_depth, CopyFromParent, scr->w_visual,
-		    vmask, &attribs);
     
     /* InputOnly window to get the focus when no other window can get it */
     vmask = CWEventMask|CWOverrideRedirect;
@@ -505,11 +513,12 @@ createInternalWindows(WScreen *scr)
  
     /* shadow window for dock buttons */
     vmask = CWBorderPixel|CWBackPixmap|CWBackPixel|CWCursor|CWSaveUnder|CWOverrideRedirect;
-    attribs.border_pixel = 0;
+    attribs.border_pixel = scr->black_pixel;
     attribs.save_under = True;
     attribs.override_redirect = True;
     attribs.background_pixmap = None;
     attribs.background_pixel = scr->white_pixel;
+    attribs.cursor = wCursor[WCUR_DEFAULT];
     vmask |= CWColormap;
     attribs.colormap = scr->w_colormap;
     scr->dock_shadow =
@@ -529,10 +538,18 @@ createInternalWindows(WScreen *scr)
       XCreateWindow(dpy, scr->root_win, 0, 0, 10, 10, 0, scr->w_depth,
 		    CopyFromParent, scr->w_visual, vmask, &attribs);
 
-    /* for our window manager info notice board */
-    scr->info_window =
-      XCreateWindow(dpy, scr->root_win, 0, 0, 10, 10, 0, CopyFromParent,
-		    CopyFromParent, CopyFromParent, 0, NULL);
+    
+    /* workspace name */
+    vmask = CWBackPixel|CWSaveUnder|CWOverrideRedirect|CWColormap
+	|CWBorderPixel;
+    attribs.save_under = True;
+    attribs.override_redirect = True;
+    attribs.colormap = scr->w_colormap;
+    attribs.background_pixel = scr->icon_back_texture->normal.pixel;
+    attribs.border_pixel = 0; /* do not care */
+    scr->workspace_name =
+      XCreateWindow(dpy, scr->root_win, 0, 0, 10, 10, 0, scr->w_depth,
+		    CopyFromParent, scr->w_visual, vmask, &attribs);
 
     /*
      * If the window is clicked without having ButtonPress selected, the
@@ -637,6 +654,8 @@ wScreenInit(int screen_number)
     scr = wmalloc(sizeof(WScreen));
     memset(scr, 0, sizeof(WScreen));
 
+    scr->stacking_list = WMCreateTreeBag();
+    
     /* initialize globals */
     scr->screen = screen_number;
     scr->root_win = RootWindow(dpy, screen_number);
@@ -669,6 +688,17 @@ wScreenInit(int screen_number)
     
     XSelectInput(dpy, scr->root_win, event_mask);
 
+#ifdef KEEP_XKB_LOCK_STATUS     
+    /* Only GroupLock doesn't work correctly in my system since right-alt
+     * can change mode while holding it too - ]d
+     */
+    if (wXkbSupported) {
+        XkbSelectEvents(dpy,XkbUseCoreKbd,
+                XkbStateNotifyMask,
+                XkbStateNotifyMask);
+    }
+#endif /* KEEP_XKB_LOCK_STATUS */
+
     XSync(dpy, False);
     XSetErrorHandler(oldHandler);
     
@@ -677,20 +707,40 @@ wScreenInit(int screen_number)
 	return NULL;
     }
 
+#ifndef DEFINABLE_CURSOR
     XDefineCursor(dpy, scr->root_win, wCursor[WCUR_DEFAULT]);
+#endif
 
     /* screen descriptor for raster graphic library */
-    rattr.flags = RC_RenderMode | RC_ColorsPerChannel;
-    rattr.render_mode = wPreferences.no_dithering?RM_MATCH:RM_DITHER;
+    rattr.flags = RC_RenderMode | RC_ColorsPerChannel | RC_StandardColormap;
+    rattr.render_mode = wPreferences.no_dithering 
+				? RBestMatchRendering
+				: RDitheredRendering;
+
+    /* if the std colormap stuff works ok, this will be ignored */
     rattr.colors_per_channel = wPreferences.cmap_size;
     if (rattr.colors_per_channel<2)
 	rattr.colors_per_channel = 2;
 
+    /* will only be accounted for in PseudoColor */
+    rattr.standard_colormap_mode = RCreateStdColormap;
+
     if (wVisualID>=0) {
 	rattr.flags |= RC_VisualID;
 	rattr.visualid = wVisualID;
     }
+    
     scr->rcontext = RCreateContext(dpy, screen_number, &rattr);
+
+    if (!scr->rcontext && RErrorCode == RERR_STDCMAPFAIL) {
+	wwarning(RMessageForError(RErrorCode));
+
+	rattr.flags &= ~RC_StandardColormap;
+	rattr.standard_colormap_mode = RUseStdColormap;
+
+	scr->rcontext = RCreateContext(dpy, screen_number, &rattr);
+    }
+
     if (!scr->rcontext) {
 	wwarning(_("could not initialize graphics library context: %s"),
 		 RMessageForError(RErrorCode));
@@ -746,6 +796,13 @@ wScreenInit(int screen_number)
     /* create GCs with default values */
     allocGCs(scr);
 
+    
+    /* for our window manager info notice board. Need to
+     * create before reading the defaults, because it will be used there. 
+     */
+    scr->info_window = XCreateSimpleWindow(dpy, scr->root_win, 0, 0, 10, 10, 
+					   0, 0, 0);
+    
     /* read defaults for this screen */
     wReadDefaults(scr, WDWindowMaker->dictionary);
 
@@ -794,15 +851,15 @@ wScreenInit(int screen_number)
     /* initialize balloon text stuff */
     wBalloonInitialize(scr);
 #endif
+    
+    scr->info_text_font = WMBoldSystemFontOfSize(scr->wmscreen, 12);
+					     
+
+    scr->gview = WCreateGeometryView(scr->wmscreen);
+    WMRealizeWidget(scr->gview);
 
     wScreenUpdateUsableArea(scr);
 
-#ifndef LITE
-    /* kluge to load menu configurations at startup */
-    OpenRootMenu(scr, -10000, -10000, False);
-    wMenuUnmap(scr->root_menu);
-#endif
-
     return scr;
 }
 
@@ -934,6 +991,12 @@ wScreenRestoreState(WScreen *scr)
     proplist_t state;
     char *path;
 
+    
+#ifndef LITE
+    OpenRootMenu(scr, -10000, -10000, False);
+    wMenuUnmap(scr->root_menu);
+#endif
+
     make_keys();
 
     if (wScreenCount == 1)
@@ -960,6 +1023,13 @@ wScreenRestoreState(WScreen *scr)
 
     wWorkspaceRestoreState(scr);
 
+#ifdef VIRTUAL_DESKTOP
+        /*
+         *      * create inputonly windows at the border of screen
+         *           */
+        wWorkspaceManageEdge(scr);
+#endif
+
     if (!wPreferences.flags.nodock) {
         state = PLGetDictionaryEntry(scr->session_state, dDock);
         scr->dock = wDockRestoreState(scr, state, WM_DOCK);
@@ -1012,7 +1082,7 @@ wScreenSaveState(WScreen *scr)
 
     /* save dock state to file */
     if (!wPreferences.flags.nodock) {
-        wDockSaveState(scr);
+        wDockSaveState(scr, old_state);
     } else {
         if ((foo = PLGetDictionaryEntry(old_state, dDock))!=NULL) {
             PLInsertDictionaryEntry(scr->session_state, dDock, foo);
@@ -1055,7 +1125,7 @@ wScreenSaveState(WScreen *scr)
     free(str);
     PLSetFilename(scr->session_state, path);
     if (!PLSave(scr->session_state, YES)) {
-	wwarning(_("could not save session state in %s"), PLGetString(path));
+	wsyserror(_("could not save session state in %s"), PLGetString(path));
     }
     PLRelease(path);
     PLRelease(old_state);
