@@ -91,7 +91,6 @@
 
 #include "dock.h"
 
-#include "list.h"
 
 #include <proplist.h>
 
@@ -126,6 +125,7 @@ static proplist_t sShaded;
 static proplist_t sMiniaturized;
 static proplist_t sHidden;
 static proplist_t sGeometry;
+static proplist_t sShortcutMask;
 
 static proplist_t sDock;
 
@@ -148,6 +148,7 @@ make_keys()
     sHidden = PLMakeString("Hidden");
     sGeometry = PLMakeString("Geometry");
     sDock = PLMakeString("Dock");
+    sShortcutMask = PLMakeString("ShortcutMask");
 
     sYes = PLMakeString("Yes");
     sNo = PLMakeString("No");
@@ -186,6 +187,24 @@ getBool(proplist_t value)
 }
 
 
+static unsigned
+getInt(proplist_t value)
+{
+    char *val;
+    unsigned n;
+    
+    if (!PLIsString(value))
+	return 0;
+    val = PLGetString(value);
+    if (!val)
+	return 0;
+    if (sscanf(val, "%u", &n) != 1)
+	return 0;
+
+    return n;
+}
+
+
 
 static proplist_t
 makeWindowState(WWindow *wwin, WApplication *wapp)
@@ -194,10 +213,12 @@ makeWindowState(WWindow *wwin, WApplication *wapp)
     Window win;
     int argc;
     char **argv;
+    int i;
+    unsigned mask;
     char *class, *instance, *command=NULL, buffer[256];
     proplist_t win_state, cmd, name, workspace;
     proplist_t shaded, miniaturized, hidden, geometry;
-    proplist_t dock;
+    proplist_t dock, shortcut;
 
     if (wwin->main_window!=None && wwin->main_window!=wwin->client_win)
         win = wwin->main_window;
@@ -232,6 +253,16 @@ makeWindowState(WWindow *wwin, WApplication *wapp)
         sprintf(buffer, "%ix%i+%i+%i", wwin->client.width, wwin->client.height,
 		wwin->frame_x, wwin->frame_y);
         geometry = PLMakeString(buffer);
+	
+	for (mask = 0, i = 0; i < MAX_WINDOW_SHORTCUTS; i++) {
+	    if (scr->shortcutWindows[i] != NULL &&
+		WMGetFirstInBag(scr->shortcutWindows[i], wwin) != WBNotFound) {
+		mask |= 1<<i;
+	    }
+	}
+
+	sprintf(buffer, "%u", mask);
+	shortcut = PLMakeString(buffer);
 
         win_state = PLMakeDictionaryFromEntries(sName, name,
                                                 sCommand, cmd,
@@ -239,6 +270,7 @@ makeWindowState(WWindow *wwin, WApplication *wapp)
                                                 sShaded, shaded,
                                                 sMiniaturized, miniaturized,
                                                 sHidden, hidden,
+						sShortcutMask, shortcut,
                                                 sGeometry, geometry,
                                                 NULL);
 
@@ -246,6 +278,7 @@ makeWindowState(WWindow *wwin, WApplication *wapp)
         PLRelease(cmd);
         PLRelease(workspace);
         PLRelease(geometry);
+	PLRelease(shortcut);
         if (wapp && wapp->app_icon && wapp->app_icon->dock) {
             int i;
             char *name;
@@ -281,7 +314,7 @@ wSessionSaveState(WScreen *scr)
     WWindow *wwin = scr->focused_window;
     proplist_t win_info, wks;
     proplist_t list=NULL;
-    LinkedList *wapp_list=NULL;
+    WMBag *wapp_list=NULL;
 
 
     make_keys();
@@ -294,10 +327,13 @@ wSessionSaveState(WScreen *scr)
 
     list = PLMakeArrayFromElements(NULL);
 
+    wapp_list = WMCreateBag(16);
+
     while (wwin) {
         WApplication *wapp=wApplicationOf(wwin->main_window);
 
-        if (wwin->transient_for==None && list_find(wapp_list, wapp)==NULL
+        if (wwin->transient_for==None 
+	    && WMGetFirstInBag(wapp_list, wapp)==WBNotFound
 	    && !WFLAGP(wwin, dont_save_session)) {
             /* A entry for this application was not yet saved. Save one. */
             if ((win_info = makeWindowState(wwin, wapp))!=NULL) {
@@ -308,7 +344,7 @@ wSessionSaveState(WScreen *scr)
                  * application list, so no multiple entries for the same
                  * application are saved.
                  */
-                wapp_list = list_cons(wapp, wapp_list);
+		WMPutInBag(wapp_list, wapp);
             }
         }
         wwin = wwin->prev;
@@ -321,7 +357,7 @@ wSessionSaveState(WScreen *scr)
     PLInsertDictionaryEntry(scr->session_state, sWorkspace, wks);
     PLRelease(wks);
 
-    list_free(wapp_list);
+    WMFreeBag(wapp_list);
 }
 
 
@@ -380,6 +416,7 @@ getWindowState(WScreen *scr, proplist_t win_state)
     WSavedState *state = wmalloc(sizeof(WSavedState));
     proplist_t value;
     char *tmp;
+    unsigned mask;
     int i;
 
     memset(state, 0, sizeof(WSavedState));
@@ -405,6 +442,10 @@ getWindowState(WScreen *scr, proplist_t win_state)
         state->miniaturized = getBool(value);
     if ((value = PLGetDictionaryEntry(win_state, sHidden))!=NULL)
         state->hidden = getBool(value);
+    if ((value = PLGetDictionaryEntry(win_state, sHidden))!=NULL) {
+        mask = getInt(value);
+	state->window_shortcuts = mask;
+    }
 
     value = PLGetDictionaryEntry(win_state, sGeometry);
     if (value && PLIsString(value)) {
@@ -508,7 +549,8 @@ wSessionRestoreState(WScreen *scr)
                 btn = dock->icon_array[j];
                 if (btn && SAME(instance, btn->wm_instance) &&
                     SAME(class, btn->wm_class) &&
-                    SAME(command, btn->command)) {
+                    SAME(command, btn->command) &&
+		    !btn->launching) {
                     found = 1;
                     break;
                 }
@@ -682,15 +724,35 @@ wSessionSendSaveYourself(WScreen *scr)
  */
 
 
+static char*
+getWindowRole(Window window)
+{
+    XTextProperty prop;
+    static Atom atom = 0;
+    
+    if (!atom)
+	atom = XInternAtom(dpy, "WM_WINDOW_ROLE", False);
+
+    if (XGetTextProperty(dpy, window, &prop, atom)) {
+	if (prop.encoding == XA_STRING && prop.format == 8 && prop.nitems > 0)
+	    return prop.value;
+    }
+    
+    return NULL;
+}
+
+
 /*
- * Windows are identified as:
- * 	WM_CLASS(instance.class).WM_WINDOW_ROLE
  *
  * Saved Info:
- * 
- * WM_CLASS.instance 
- * WM_CLASSS.class
+ *
  * WM_WINDOW_ROLE
+ *
+ * WM_CLASS.instance
+ * WM_CLASS.class
+ * WM_NAME
+ * WM_COMMAND
+ * 
  * geometry
  * state = (miniaturized, shaded, etc)
  * attribute
@@ -736,6 +798,64 @@ makeAppState(WWindow *wwin)
 }
 
 
+
+Bool
+wSessionGetStateFor(WWindow *wwin, WSessionData *state)
+{
+    char *str;
+    proplist_t slist;
+    proplist_t elem;
+    proplist_t value;
+    int index = 0;
+
+    index = 3;
+
+    /* geometry */
+    value = PLGetArrayElement(slist, index++);    
+    str = PLGetString(value);
+
+    sscanf(str, "%i %i %i %i %i %i", &state->x, &state->y,
+	   &state->width, &state->height,
+	   &state->user_changed_width, &state->user_changed_height);
+
+
+    /* state */
+    value = PLGetArrayElement(slist, index++);
+    str = PLGetString(value);
+
+    sscanf(str, "%i %i %i", &state->miniaturized, &state->shaded,
+	   &state->maximized);
+
+
+    /* attributes */
+    value = PLGetArrayElement(slist, index++);
+    str = PLGetString(value);
+
+    getAttributeState(str, &state->mflags, &state->flags);
+
+
+    /* workspace */
+    value = PLGetArrayElement(slist, index++);
+    str = PLGetString(value);
+
+    sscanf(str, "%i", &state->workspace);
+
+
+    /* app state (repeated for all windows of the app) */
+    value = PLGetArrayElement(slist, index++);
+    str = PLGetString(value);
+
+    /* ???? */
+
+    /* shortcuts */
+    value = PLGetArrayElement(slist, index++);
+    str = PLGetString(value);
+
+    sscanf(str, "%i", &state->shortcuts);
+}
+
+
+
 static proplist_t
 makeAttributeState(WWindow *wwin)
 {
@@ -745,7 +865,8 @@ makeAttributeState(WWindow *wwin)
 #define W_FLAG(wwin, FLAG)	((wwin)->defined_user_flags.FLAG \
 					? (wwin)->user_flags.FLAG : -1)
 
-    sprintf(buffer, "%i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i",
+    sprintf(buffer, 
+	    "%i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i",
 	    W_FLAG(no_titlebar),
 	    W_FLAG(no_resizable),
 	    W_FLAG(no_closable),
@@ -779,6 +900,17 @@ makeAttributeState(WWindow *wwin)
 }
 
 
+static void
+appendStringInArray(proplist_t array, char *str)
+{
+    proplist_t val;
+    
+    val = PLMakeString(str);
+    PLAppendArrayElement(array, val);
+    PLRelease(val);
+}
+
+
 static proplist_t
 makeClientState(WWindow *wwin)
 {
@@ -787,47 +919,65 @@ makeClientState(WWindow *wwin)
     char *str;
     char buffer[256];
     int i;
-    
+    unsigned shortcuts;
+
     state = PLMakeArrayWithElements(NULL, NULL);
     
-    /* spec */
-    PLAppendArrayElement(state, PLMakeString(wwin->wm_instance));
-    PLAppendArrayElement(state, PLMakeString(wwin->wm_class));
+    /* WM_WINDOW_ROLE */
+    str = getWindowRole(wwin->client_win);
+    if (!str)
+	appendStringInArray(state, "");
+    else {
+	appendStringInArray(state, str);
+	XFree(str);
+    }
 
-    PLAppendArrayElement(state, PLMakeString(str));
+    /* WM_CLASS.instance */
+    appendStringInArray(state, wwin->wm_instance);
+
+    /* WM_CLASS.class */
+    appendStringInArray(state, wwin->wm_class);
+
+    /* WM_NAME */
+    if (wwin->flags.wm_name_changed)
+	appendStringInArray(state, "");
+    else
+	appendStringInArray(state, wwin->frame->name);
 
     /* geometry */
-    sprintf(buffer, "%i %i %i %i", wwin->frame_x, wwin->frame_y,
-	    wwin->frame->core->width, wwin->frame->core->height);
-    PLAppendArrayElement(state, PLMakeString(buffer));
+    sprintf(buffer, "%i %i %i %i %i %i", wwin->frame_x, wwin->frame_y,
+	    wwin->client.width, wwin->client.height,
+	    wwin->flags.user_changed_width, wwin->flags.user_changed_height);
+    appendStringInArray(state, buffer);
 
     /* state */
     sprintf(buffer, "%i %i %i", wwin->flags.miniaturized,
 	    wwin->flags.shaded, wwin->flags.maximized);
-    PLAppendArrayElement(state, PLMakeString(buffer));
+    appendStringInArray(state, buffer);
 
     /* attributes */
-    PLAppendArrayElement(state, makeAttributeState(wwin));
+    tmp = makeAttributeState(wwin);
+    PLAppendArrayElement(state, tmp);
+    PLRelease(tmp);
 
     /* workspace */
     sprintf(buffer, "%i", wwin->frame->workspace);
-    PLAppendArrayElement(state, PLMakeString(buffer));
+    appendStringInArray(state, buffer);
 
     /* app state (repeated for all windows of the app) */
-    PLAppendArrayElement(state, makeAppState(wwin));
+    tmp = makeAppState(wwin);
+    PLAppendArrayElement(state, tmp);
+    PLRelease(tmp);
 
     /* shortcuts */
-    tmp = PLMakeArrayWithElements(NULL, NULL);
-    
+    shortcuts = 0;
     for (i = 0; i < MAX_WINDOW_SHORTCUTS; i++) {
-
 	if (scr->shortcutWindow[i] == wwin) {
-
-	    sprintf(buffer, "%i", i);
-	    PLAppendArrayElement(tmp, PLMakeString(buffer));
+	    shortcuts |= 1 << i;
 	}
     }
-    PLAppendArrayElement(state, tmp);
+    sprintf(buffer, "%ui", shortcuts);
+    appendStringInArray(tmp, buffer);
 
     return state;
 }
