@@ -3,7 +3,7 @@
  * 
  *  WindowMaker window manager
  * 
- *  Copyright (c) 1997 Alfredo K. Kojima
+ *  Copyright (c) 1997-1999 Alfredo K. Kojima
  * 
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -37,13 +37,16 @@
  *--------------------------------------------------------------------
  */
 
-/* undefine WEENDOZE_CYCLING to remove all Windoze style Alt-Tab cycling
- * If define, windoze cycling must still be enabled with the 
- * WindozeCycling = YES and FinishedCyclingKey = Meta_L options
- 
- not working yet
+
+/*
+ * #undefine if you dont want texture plugin support or your system have
+ * some sort of problem with them.
  */
-#undef WEENDOZE_CYCLING
+#define TEXTURE_PLUGIN
+
+#ifdef TEXTURE_PLUGIN
+#define DRAWSTRING_PLUGIN
+#endif
 
 
 /* undefine ANIMATIONS if you don't want animations for iconification,
@@ -57,7 +60,7 @@
  */
 #define USECPP
 
-/* #define CPP_PATH /lib/cpp */
+/* #define CPP_PATH /usr/local/bin/cpp */
 
 /*
  * support for OffiX DND drag and drop in the Dock
@@ -65,9 +68,11 @@
 #define OFFIX_DND
 
 /*
- * support for XDE drang and drop in the Dock. Experimental
+ * support for XDND drop in the Dock. Experimental
  */
-#undef XDE_DND
+#ifdef GNOME_STUFF
+#define XDND
+#endif
 
 /*
  * support for Motif window manager (mwm) window hints
@@ -121,12 +126,9 @@
 
 
 /*
- * define EXTEND_WINDOWSHORTCUT if you have a great memory and want 10
- * window shortcuts instead of 4. You will also need to define the
- * same thing in WPrefs.app/WPrefs.h to get the corresponding entries
- * in WPrefs.
+ * define EXTEND_WINDOWSHORTCUT if you want 10 window shortcuts instead of 4. 
  */
-#undef EXTEND_WINDOWSHORTCUT
+#define EXTEND_WINDOWSHORTCUT
 
 
 
@@ -150,20 +152,40 @@
 /*
  * disable/enable workspace indicator in the dock
  */
-
 #undef WS_INDICATOR 
 
 
+/*
+ * define HIDDENDOT if you want a dot to be shown in the application icon
+ * of applications that are hidden.
+ */
+#define HIDDENDOT
+
+
 /*
  * Ignores the PPosition hint from clients. This is needed for some
  * programs that have buggy implementations of such hint and place
  * themselves in strange locations.
  */
-#undef IGNORE_NOPPOSITION
+#undef IGNORE_PPOSITION
+
+
+/*
+ * Do not scale application icon and miniwindow icon images.
+ */
+#undef DONT_SCALE_ICONS
+
+
+
+
 
 
 #define SILLYNESS
 
+
+
+
+
 /*
  *..........................................................................
  * The following options WILL NOT BE MADE RUN-TIME. Please do not request.
@@ -173,19 +195,16 @@
 
 /*
  * define SHADOW_RESIZEBAR if you want a resizebar with shadows like in
- * NextStep 3.x ??, instead of the default Openstep look. NEXTSTEP 3.3
- * also does not have these shadows.
+ * AfterStep, instead of the default Openstep look.
+ * NEXTSTEP 3.3 also does not have these shadows.
  */
 #undef SHADOW_RESIZEBAR
 
-/*
- * Titlebar text shadow
- */
-#undef TITLE_TEXT_SHADOW
-
 /*
  * Define DEMATERIALIZE_ICON if you want the undocked icon animation
- * to be a progressive disaparison animation. 
+ * to be a progressive disaparison animation.
+ * This will cause all application icons to be created with Save Under
+ * enable.
  */
 #undef DEMATERIALIZE_ICON
 
@@ -201,6 +220,17 @@
  */
 #undef WINDOW_BIRTH_ZOOM
 
+/*
+ * whether arrow drawing in clip buttons should be gradiented
+ */
+#undef GRADIENT_CLIP_ARROWS
+
+/*
+ * define DEFINABLE_CURSOR if you want WindowMaker's default cursor
+ * to be user-definable instead of using a hard-coded left_ptr.
+ */
+#undef DEFINABLE_CURSOR
+
 
 /*
  *--------------------------------------------------------------------
@@ -228,6 +258,12 @@
 
 #define DEFAULTS_DIR "Defaults"
 
+#ifdef USE_TIFF
+#define DEF_BUTTON_IMAGES PKGDATADIR"/buttons.tiff"
+#else
+#define DEF_BUTTON_IMAGES PKGDATADIR"/buttons.xpm"
+#endif
+
 /* the file of the system wide submenu to be forced into the main menu */
 /* remove the comments if you want it */
 /*
@@ -238,6 +274,12 @@
 #define DEF_PIXMAP_PATHS \
 "(\"~/pixmaps\",\"~/GNUstep/Library/WindowMaker/Pixmaps\",\""PIXMAPDIR"\")"
 
+#ifdef USER_MENU
+#define GLOBAL_USER_MENU_PATH PKGDATADIR"/UserMenus"
+#define DEF_USER_MENU_PATHS \
+"~/GNUstep/Library/WindowMaker/UserMenus:"GLOBAL_USER_MENU_PATH
+#endif
+
 /* icon path */
 #define DEF_ICON_PATHS \
 "(\"~/pixmaps\",\"~/GNUstep/Library/Icons\",\"/usr/include/X11/pixmaps/\",\""PIXMAPDIR"\")"
@@ -257,19 +299,29 @@
 #define DEF_ICON_TITLE_FONT "\"-*-*-medium-r-normal--10-*\""
 #define DEF_CLIP_TITLE_FONT "\"-*-*-medium-r-normal--10-*\""
 #define DEF_INFO_TEXT_FONT "\"-*-*-medium-r-normal--14-*\""
-#else /* I18N_MB */
+#define DEF_ADDITONAL_FONT "-*-*-medium-r-normal--12-*"
+
+#define DEF_WORKSPACE_NAME_FONT "\"-*-*-medium-r-normal--24-*\""
+#else /* !I18N_MB */
 #define DEF_TITLE_FONT "\"-*-helvetica-bold-r-normal-*-12-*-*-*-*-*-*-*\""
 #define DEF_MENU_TITLE_FONT "\"-*-helvetica-bold-r-normal-*-12-*-*-*-*-*-*-*\""
 #define DEF_MENU_ENTRY_FONT "\"-*-helvetica-medium-r-normal-*-12-*-*-*-*-*-*-*\""
 #define DEF_ICON_TITLE_FONT "\"-*-helvetica-medium-r-normal-*-8-*-*-*-*-*-*-*\""
 #define DEF_CLIP_TITLE_FONT "\"-*-helvetica-bold-r-normal-*-10-*-*-*-*-*-*-*\""
 #define DEF_INFO_TEXT_FONT "\"-*-helvetica-medium-r-normal-*-12-*-*-*-*-*-*-*\""
+
+#define DEF_WORKSPACE_NAME_FONT "\"-*-lucida-bold-r-*-*-24-*-*-*-*-*-*-*\""
 #endif /* !I18N_MB */
 
 #define HELVETICA10_FONT "-*-helvetica-medium-r-normal-*-10-*-*-*-*-*-*-*"
 
 #define DEF_FRAME_THICKNESS 1	       /* linewidth of the move/resize frame */
 
+#define DEF_WINDOW_TITLE_EXTEND_SPACE	"0"
+#define DEF_MENU_TITLE_EXTEND_SPACE	"0"
+#define DEF_MENU_TEXT_EXTEND_SPACE	"0"
+#define TITLEBAR_EXTEND_SPACE 4
+
 #define DEF_XPM_CLOSENESS	40000 
 
 /* default position of application menus */
@@ -316,7 +368,10 @@
 /* delay for menu item selection hysteresis */
 #define MENU_SELECT_DELAY 200
 
-/* animation speed constants */
+/* delay for jumpback of scrolled menus */
+#define MENU_JUMP_BACK_DELAY 	400
+
+/* *** animation speed constants *** */
 
 /* icon slide */
 #define ICON_SLIDE_SLOWDOWN_UF	20
@@ -355,6 +410,7 @@
 #define MENU_SCROLL_STEPS_U	1
 #define MENU_SCROLL_DELAY_U	8
 
+
 /* shade animation */
 #define SHADE_STEPS_UF		5
 #define SHADE_DELAY_UF		0
@@ -371,6 +427,12 @@
 #define SHADE_STEPS_U		20
 #define SHADE_DELAY_U		10
 
+
+/* workspace name on switch display */
+#define WORKSPACE_NAME_FADE_DELAY 30
+
+#define WORKSPACE_NAME_DELAY	400
+
 /* window birth animation steps (DO NOT MAKE IT RUN-TIME) */
 #define WINDOW_BIRTH_STEPS	20
 
@@ -395,6 +457,9 @@
 /* Vicinity in which an icon can be attached to the clip */
 #define CLIP_ATTACH_VICINITY	1
 
+#define CLIP_BUTTON_SIZE  23
+
+
 /* The amount of space (in multiples of the icon size)
  * a docked icon must be dragged out to detach it */
 #define DOCK_DETTACH_THRESHOLD	3
@@ -402,12 +467,16 @@
 /* Delay (in ms) after which the clip will autocollapse when leaved */
 #define AUTO_COLLAPSE_DELAY     1000
 
+/* Delay (in ms) after which the clip will autoexpand when entered.
+ * Set this to zero if you want instant expanding. */
+#define AUTO_EXPAND_DELAY       600
+
 /* Delay (in ms) after which the clip will be lowered when leaved */
 #define AUTO_LOWER_DELAY        1000
 
 /* Delay (in ms) after which the clip will be raised when entered.
  * Set this to zero if you want instant raise. */
-#define AUTO_RAISE_DELAY        650
+#define AUTO_RAISE_DELAY        600
 
 
 /* Max. number of icons the clip can have */
@@ -422,6 +491,8 @@
 #define MOVE_THRESHOLD	5 /* how many pixels to move before dragging windows
 			   * and other objects */
 
+#define KEY_CONTROL_WINDOW_WEIGHT 1
+
 #define HRESIZE_THRESHOLD	3
 
 #define MAX_WORKSPACENAME_WIDTH	16
@@ -442,12 +513,8 @@
 #define FRAME_BORDER_COLOR "black"
 
 
-#ifdef TITLE_TEXT_SHADOW
-#define TITLE_TEXT_SHADOW_X_OFFSET 1
-#define TITLE_TEXT_SHADOW_Y_OFFSET 1
-#define TITLE_TEXT_SHADOW_WIDTH 1
-#define TITLE_TEXT_SHADOW_HEIGHT 1
-#endif
+/* for boxes with high mouse sampling rates (SGI) */
+#define DELAY_BETWEEN_MOUSE_SAMPLING  10
 
 
 /*
@@ -474,8 +541,6 @@
 #define RESIZEBAR_MIN_WIDTH 20	       /* min. width of handles-corner_width */
 #define RESIZEBAR_CORNER_WIDTH 28      /* width of the corner of resizebars */
 
-#define TITLEBAR_EXTRA_HEIGHT 8
-
 #define MENU_INDICATOR_SPACE	12
 
 /* minimum size for windows */
@@ -524,9 +589,19 @@
 #undef SHAPED_BALLOON
 #endif
 
-
 #ifdef XKB_MODELOCK
 #define KEEP_XKB_LOCK_STATUS
+/* This is a hidden feature.
+ * choose one between 2 of these hints. Icon can be changed
+ * in def_pixmaps.h and default icon is now for Thai language only.
+ * More icons are welcome.
+ */
+#define XKB_BUTTON_HINT
+#undef LANGUAGE_TH
+#endif
+
+#ifndef DEMATERIALIZE_ICON
+# define NORMAL_ICON_KABOOM
 #endif
 
 #if HAVE_LIBINTL_H && I18N
