@@ -66,6 +66,7 @@ static proplist_t ANoTitlebar = NULL;
 static proplist_t ANoResizebar;
 static proplist_t ANoMiniaturizeButton;
 static proplist_t ANoCloseButton;
+static proplist_t ANoBorder;
 static proplist_t ANoHideOthers;
 static proplist_t ANoMouseBindings;
 static proplist_t ANoKeyBindings;
@@ -83,6 +84,9 @@ static proplist_t AStartHidden;	       /* app */
 static proplist_t ADontSaveSession;    /* app */
 static proplist_t AEmulateAppIcon;
 static proplist_t AFullMaximize;
+#ifdef XKB_BUTTON_HINT
+static proplist_t ANoLanguageButton;
+#endif
 
 static proplist_t AStartWorkspace;
 
@@ -102,6 +106,7 @@ init_wdefaults(WScreen *scr)
     ANoResizebar = PLMakeString("NoResizebar");
     ANoMiniaturizeButton = PLMakeString("NoMiniaturizeButton");
     ANoCloseButton = PLMakeString("NoCloseButton");
+    ANoBorder = PLMakeString("NoBorder");
     ANoHideOthers = PLMakeString("NoHideOthers");
     ANoMouseBindings = PLMakeString("NoMouseBindings");
     ANoKeyBindings = PLMakeString("NoKeyBindings");
@@ -119,6 +124,9 @@ init_wdefaults(WScreen *scr)
     ADontSaveSession = PLMakeString("DontSaveSession");
     AEmulateAppIcon = PLMakeString("EmulateAppIcon");
     AFullMaximize = PLMakeString("FullMaximize");
+#ifdef XKB_BUTTON_HINT
+    ANoLanguageButton = PLMakeString("NoLanguageButton");
+#endif
 
     AStartWorkspace = PLMakeString("StartWorkspace");
     
@@ -191,12 +199,14 @@ wDefaultFillAttributes(WScreen *scr, char *instance, char *class,
     proplist_t value;
     proplist_t key1, key2, key3;
     proplist_t dw, dc, dn, da;
-    char buffer[256];
 
 
-    if (class && instance)
+    if (class && instance) {
+      char *buffer = NULL;
+	  buffer = wmalloc(strlen(class)+strlen(instance)+4);
       key1 = PLMakeString(strcat(strcat(strcpy(buffer,instance),"."),class));
-    else
+	  free(buffer);
+    } else
       key1 = NULL;
 
     if (instance)
@@ -253,6 +263,9 @@ wDefaultFillAttributes(WScreen *scr, char *instance, char *class,
     value = get_value(dw, dc, dn, da, ANoCloseButton, No, useGlobalDefault);
     APPLY_VAL(value, no_close_button, ANoCloseButton);
 
+    value = get_value(dw, dc, dn, da, ANoBorder, No, useGlobalDefault);
+    APPLY_VAL(value, no_border, ANoBorder);
+
     value = get_value(dw, dc, dn, da, ANoHideOthers, No, useGlobalDefault);
     APPLY_VAL(value, no_hide_others, ANoHideOthers);
     
@@ -302,7 +315,12 @@ wDefaultFillAttributes(WScreen *scr, char *instance, char *class,
     APPLY_VAL(value, emulate_appicon, AEmulateAppIcon);
 
     value = get_value(dw, dc, dn, da, AFullMaximize, No, useGlobalDefault);
-    APPLY_VAL(value, emulate_appicon, AFullMaximize);
+    APPLY_VAL(value, full_maximize, AFullMaximize);
+
+#ifdef XKB_BUTTON_HINT
+    value = get_value(dw, dc, dn, da, ANoLanguageButton, No, useGlobalDefault);
+    APPLY_VAL(value, no_language_button, ANoLanguageButton);
+#endif
 
     /* clean up */
     PLSetStringCmpHook(StringCompareHook);
@@ -315,17 +333,19 @@ get_generic_value(WScreen *scr, char *instance, char *class, proplist_t option,
 		  Bool noDefault)
 {
     proplist_t value, key, dict;
-    char buffer[256];
     
     value = NULL;
 
     PLSetStringCmpHook(NULL);
 
     if (class && instance) {
+	char *buffer = NULL;
+	buffer = wmalloc(strlen(class)+strlen(instance)+4);
 	key = PLMakeString(strcat(strcat(strcpy(buffer,instance),"."),class));
-	
+
 	dict = PLGetDictionaryEntry(WDWindowAttributes->dictionary, key);
 	PLRelease(key);
+	free(buffer);
 
 	if (dict) {
 	    value = PLGetDictionaryEntry(dict, option);
@@ -337,7 +357,6 @@ get_generic_value(WScreen *scr, char *instance, char *class, proplist_t option,
 	
 	dict = PLGetDictionaryEntry(WDWindowAttributes->dictionary, key);
 	PLRelease(key);
-
 	if (dict) {
 	    value = PLGetDictionaryEntry(dict, option);
 	}
@@ -447,14 +466,14 @@ wDefaultGetStartWorkspace(WScreen *scr, char *instance, char *class)
 
     if (!tmp || strlen(tmp)==0)
 	return -1;
-    
+
     if (sscanf(tmp, "%i", &w)!=1) {
 	w = -1;
 	for (i=0; i < scr->workspace_count; i++) {
 	    if (strcmp(scr->workspaces[i]->name, tmp)==0) {
 		w = i;
 		break;
-	    }	
+	    }
 	}
     } else {
 	w--;
@@ -470,7 +489,6 @@ wDefaultChangeIcon(WScreen *scr, char *instance, char* class, char *file)
     WDDomain *db = WDWindowAttributes;
     proplist_t icon_value=NULL, value, attr, key, def_win, def_icon=NULL;
     proplist_t dict = db->dictionary;
-    char *buffer;
     int same = 0;
 
     if (!dict) {
@@ -488,6 +506,7 @@ wDefaultChangeIcon(WScreen *scr, char *instance, char* class, char *file)
     PLSetStringCmpHook(NULL);
 
     if (instance && class) {
+    	char *buffer;
         buffer = wmalloc(strlen(instance) + strlen(class) + 2);
         strcat(strcat(strcpy(buffer, instance), "."), class);
         key = PLMakeString(buffer);
