@@ -27,8 +27,8 @@
 #ifdef SHAPE
 #include <X11/extensions/shape.h>
 #endif
-#ifdef KEEP_XKB_LOCK_STATUS
-# include <X11/XKBlib.h>
+#ifdef KEEP_XKB_LOCK_STATUS     
+#include <X11/XKBlib.h>         
 #endif /* KEEP_XKB_LOCK_STATUS */
 #include <stdlib.h>
 #include <stdio.h>
@@ -52,6 +52,7 @@
 #include "defaults.h"
 #include "workspace.h"
 
+
 #ifdef MWM_HINTS
 # include "motif.h"
 #endif
@@ -122,6 +123,9 @@ static void windowCloseDblClick(WCoreWindow *sender, void *data, XEvent *event);
 /* iconify button */
 static void windowIconifyClick(WCoreWindow *sender, void *data, XEvent *event);
 
+#ifdef XKB_BUTTON_HINT
+static void windowLanguageClick(WCoreWindow *sender, void *data, XEvent *event);
+#endif
 
 static void titlebarMouseDown(WCoreWindow *sender, void *data, XEvent *event);
 static void titlebarDblClick(WCoreWindow *sender, void *data, XEvent *event);
@@ -129,9 +133,40 @@ static void titlebarDblClick(WCoreWindow *sender, void *data, XEvent *event);
 static void resizebarMouseDown(WCoreWindow *sender, void *data, XEvent *event);
 
 
+/****** Notification Observers ******/
 
+static void
+appearanceObserver(void *self, WMNotification *notif)
+{
+    WWindow *wwin = (WWindow*)self;
+    int flags = (int)WMGetNotificationClientData(notif);
 
+    if (!wwin->frame || (!wwin->frame->titlebar && !wwin->frame->resizebar))
+	return;
 
+    if (flags & WFontSettings) {
+        wWindowConfigureBorders(wwin);
+        if(wwin->flags.shaded) {
+            wFrameWindowResize(wwin->frame, wwin->frame->core->width,
+                    wwin->frame->top_width - 1);
+
+            wwin->client.y = wwin->frame_y - wwin->client.height 
+                + wwin->frame->top_width;
+            wWindowSynthConfigureNotify(wwin);
+        }
+    }
+    if (flags & WTextureSettings) {
+	wwin->frame->flags.need_texture_remake = 1;
+    }
+    if (flags & (WTextureSettings | WColorSettings)) {
+	if (wwin->frame->titlebar)
+	    XClearWindow(dpy, wwin->frame->titlebar->window);
+
+	wFrameWindowPaint(wwin->frame);
+    }
+}
+
+/************************************/
 
 WWindow*
 wWindowFor(Window window)
@@ -151,6 +186,7 @@ wWindowFor(Window window)
 	if (frame->flags.is_client_window_frame)
 	    return frame->child;
     }
+
     return NULL;
 }
 
@@ -169,6 +205,7 @@ wWindowCreate()
     wwin->client_descriptor.parent = wwin;
     wwin->client_descriptor.self = wwin;
     wwin->client_descriptor.parent_type = WCLASS_WINDOW;
+
     return wwin;
 }
 
@@ -178,16 +215,28 @@ wWindowDestroy(WWindow *wwin)
 {
     int i;
 
+    if (wwin->screen_ptr->cmap_window == wwin) {
+	wwin->screen_ptr->cmap_window = NULL;
+    }
+    
+    WMRemoveNotificationObserver(wwin);
+    
     wwin->flags.destroyed = 1;
 
     for (i = 0; i < MAX_WINDOW_SHORTCUTS; i++) {
-	if (wwin->screen_ptr->shortcutWindow[i] == wwin) {
-	    wwin->screen_ptr->shortcutWindow[i] = NULL;
+	if (!wwin->screen_ptr->shortcutWindows[i])
+	    continue;
+
+	WMRemoveFromBag(wwin->screen_ptr->shortcutWindows[i], wwin);
+
+	if (!WMGetBagItemCount(wwin->screen_ptr->shortcutWindows[i])) {
+	    WMFreeBag(wwin->screen_ptr->shortcutWindows[i]);
+	    wwin->screen_ptr->shortcutWindows[i] = NULL;
 	}
     }
 
     if (wwin->normal_hints)
-      free(wwin->normal_hints);
+      XFree(wwin->normal_hints);
     
     if (wwin->wm_hints)
       XFree(wwin->wm_hints);
@@ -225,28 +274,48 @@ static void
 setupGNUstepHints(WWindow *wwin, GNUstepWMAttributes *gs_hints)
 {
     if (gs_hints->flags & GSWindowStyleAttr) {
-	wwin->client_flags.no_titlebar = 
-		  ((gs_hints->window_style & WMTitledWindowMask)?0:1);
+	if (gs_hints->window_style == WMBorderlessWindowMask) {
+	    wwin->client_flags.no_border = 1;
+	    wwin->client_flags.no_titlebar = 1;
+	    wwin->client_flags.no_closable = 1;
+	    wwin->client_flags.no_miniaturizable = 1;
+	    wwin->client_flags.no_resizable = 1;
+	    wwin->client_flags.no_close_button = 1;
+	    wwin->client_flags.no_miniaturize_button = 1;
+	    wwin->client_flags.no_resizebar = 1;
+	} else {
+	    wwin->client_flags.no_close_button =
+		((gs_hints->window_style & WMClosableWindowMask)?0:1);
 
-	wwin->client_flags.no_close_button =
-		  ((gs_hints->window_style & WMClosableWindowMask)?0:1);
+	    wwin->client_flags.no_closable =
+		((gs_hints->window_style & WMClosableWindowMask)?0:1);
 
-	wwin->client_flags.no_closable =
-		  ((gs_hints->window_style & WMClosableWindowMask)?0:1);
+	    wwin->client_flags.no_miniaturize_button =
+		((gs_hints->window_style & WMMiniaturizableWindowMask)?0:1);
 
-	wwin->client_flags.no_miniaturize_button =
-		  ((gs_hints->window_style & WMMiniaturizableWindowMask)?0:1);
+	    wwin->client_flags.no_miniaturizable = 
+		wwin->client_flags.no_miniaturize_button;
 
-	wwin->client_flags.no_miniaturizable =
-		  ((gs_hints->window_style & WMMiniaturizableWindowMask)?0:1);
+	    wwin->client_flags.no_resizebar =
+		((gs_hints->window_style & WMResizableWindowMask)?0:1);
 
-	wwin->client_flags.no_resizebar =
-		  ((gs_hints->window_style & WMResizableWindowMask)?0:1);
+	    wwin->client_flags.no_resizable = wwin->client_flags.no_resizebar;
 
-	wwin->client_flags.no_resizable =
-		  ((gs_hints->window_style & WMResizableWindowMask)?0:1);
+	    /* these attributes supposedly imply in the existence
+	     * of a titlebar */
+	    if (gs_hints->window_style & (WMResizableWindowMask|
+					  WMClosableWindowMask|
+					  WMMiniaturizableWindowMask)) {
+		wwin->client_flags.no_titlebar = 0;
+	    } else {
+		wwin->client_flags.no_titlebar = 
+		    ((gs_hints->window_style & WMTitledWindowMask)?0:1);
+	    }
+
+	}
     } else {
 	/* setup the defaults */
+	wwin->client_flags.no_border = 0;
 	wwin->client_flags.no_titlebar = 0;
 	wwin->client_flags.no_closable = 0;
 	wwin->client_flags.no_miniaturizable = 0;
@@ -360,7 +429,7 @@ wWindowSetupInitialAttributes(WWindow *wwin, int *level, int *workspace)
 #endif /* MWM_HINTS */
 
 #ifdef KWM_HINTS
-	wKWMCheckClientHints(wwin, &tmp_workspace);
+	wKWMCheckClientHints(wwin, &tmp_level, &tmp_workspace);
 #endif /* KWM_HINTS */
 	
 #ifdef GNOME_STUFF
@@ -502,7 +571,7 @@ wManageWindow(WScreen *scr, Window window)
     Bool withdraw = False;
 
     /* mutex. */
-//    XGrabServer(dpy);
+    /* XGrabServer(dpy); */
     XSync(dpy, False);
     /* make sure the window is still there */
     if (!XGetWindowAttributes(dpy, window, &wattribs)) {
@@ -582,6 +651,9 @@ wManageWindow(WScreen *scr, Window window)
     XSetWindowBorderWidth(dpy, window, 0);
 
     /* get hints from GNUstep app */
+    if (wwin->wm_class != 0 && strcmp(wwin->wm_class, "GNUstep") == 0) {
+	wwin->flags.is_gnustep = 1;
+    }
     if (!PropGetGNUstepWMAttr(window, &wwin->wm_gnustep_attr)) {
 	wwin->wm_gnustep_attr = NULL;
     }
@@ -761,6 +833,19 @@ wManageWindow(WScreen *scr, Window window)
 	    wwin->flags.hidden = wstate->hidden;
 	    wwin->flags.miniaturized = wstate->miniaturized;
 	    workspace = wstate->workspace;
+
+	    if (scr->flags.startup && wstate->window_shortcuts > 0) {
+		int i;
+
+		for (i = 0; i < MAX_WINDOW_SHORTCUTS; i++) {
+		    if (wstate->window_shortcuts & (1<<i)) {
+			if (!scr->shortcutWindows[i])
+			    scr->shortcutWindows[i] = WMCreateBag(4);
+
+			WMPutInBag(scr->shortcutWindows[i], wwin);
+		    }
+		}
+	    }
 	    free(wstate);
 	}
     }
@@ -816,12 +901,34 @@ wManageWindow(WScreen *scr, Window window)
 	if (win_state && win_state->state->use_geometry) {
 	    x = win_state->state->x;
 	    y = win_state->state->y;
-	} else if (wwin->transient_for==None && !scr->flags.startup 
+	} else if ((wwin->transient_for==None
+		    || wPreferences.window_placement!=WPM_MANUAL)
+		   && !scr->flags.startup 
 		   && workspace == scr->current_workspace
 		   && !wwin->flags.miniaturized
 		   && !wwin->flags.maximized
 		   && !(wwin->normal_hints->flags & (USPosition|PPosition))) {
-	    PlaceWindow(wwin, &x, &y, width, height);
+	    
+	    if (transientOwner && transientOwner->flags.mapped) {
+		int offs = WMAX(20, 2*transientOwner->frame->top_width);
+		
+		x = transientOwner->frame_x + 
+		    abs((transientOwner->frame->core->width - width)/2) + offs;
+		y = transientOwner->frame_y +
+		    abs((transientOwner->frame->core->height - height)/3) + offs;
+
+		if (x < 0)
+		    x = 0;
+		else if (x + width > scr->scr_width)
+		    x = scr->scr_width - width;
+
+		if (y < 0)
+		    y = 0;
+		else if (y + height > scr->scr_height)
+		    y = scr->scr_height - height;
+	    } else {
+		PlaceWindow(wwin, &x, &y, width, height);
+	    }
 	    if (wPreferences.window_placement == WPM_MANUAL)
 		dontBring = True;
 	} 
@@ -843,16 +950,26 @@ wManageWindow(WScreen *scr, Window window)
      *--------------------------------------------------
      */
     foo = WFF_LEFT_BUTTON | WFF_RIGHT_BUTTON;
+#ifdef XKB_BUTTON_HINT
+    if (wPreferences.modelock)
+        foo |= WFF_LANGUAGE_BUTTON;
+#endif
     if (!WFLAGP(wwin, no_titlebar))
 	foo |= WFF_TITLEBAR;
     if (!WFLAGP(wwin, no_resizebar))
 	foo |= WFF_RESIZEBAR;
+    if (!WFLAGP(wwin, no_border))
+	foo |= WFF_BORDER;
 
     wwin->frame = wFrameWindowCreate(scr, window_level, 
-				     x, y, width, height, foo,
+				     x, y, width, height,
+                     &wPreferences.window_title_clearance, foo,
 				     scr->window_title_texture,
-				     (WTexture**)scr->resizebar_texture,
+				     scr->resizebar_texture,
 				     scr->window_title_pixel, 
+#ifdef DRAWSTRING_PLUGIN
+                     W_STRING_FTITLE,
+#endif
 				     &scr->window_title_gc, 
 				     &scr->title_font);
 
@@ -861,13 +978,17 @@ wManageWindow(WScreen *scr, Window window)
 
     /* setup button images */
     wWindowUpdateButtonImages(wwin);
-    
+
     /* hide unused buttons */
     foo = 0;
     if (WFLAGP(wwin, no_close_button))
 	foo |= WFF_RIGHT_BUTTON;
     if (WFLAGP(wwin, no_miniaturize_button))
 	foo |= WFF_LEFT_BUTTON;
+#ifdef XKB_BUTTON_HINT
+    if (WFLAGP(wwin, no_language_button) || WFLAGP(wwin, no_focusable))
+    foo |= WFF_LANGUAGE_BUTTON;
+#endif
     if (foo!=0)
 	wFrameWindowHideButton(wwin->frame, foo);
 
@@ -889,6 +1010,10 @@ wManageWindow(WScreen *scr, Window window)
     wwin->frame->workspace = workspace;
 
     wwin->frame->on_click_left = windowIconifyClick;
+#ifdef XKB_BUTTON_HINT
+    if (wPreferences.modelock)
+        wwin->frame->on_click_language = windowLanguageClick;
+#endif
 
     wwin->frame->on_click_right = windowCloseClick;
     wwin->frame->on_dblclick_right = windowCloseDblClick;
@@ -917,6 +1042,7 @@ wManageWindow(WScreen *scr, Window window)
 	if (gy > 0)
 	    y -= wwin->frame->top_width + wwin->frame->bottom_width;
     }
+
     /* 
      * wWindowConfigure() will init the client window's size
      * (wwin->client.{width,height}) and all other geometry
@@ -993,6 +1119,7 @@ wManageWindow(WScreen *scr, Window window)
             wClientSetState(wwin, NormalState, None);
         }
 
+#if 0
 	/* if not auto focus, then map the window under the currently
 	 * focused window */
 #define _WIDTH(w) (w)->frame->core->width
@@ -1010,6 +1137,8 @@ wManageWindow(WScreen *scr, Window window)
 #undef _WIDTH
 #undef _HEIGHT
 
+#endif
+
 	if (wPreferences.superfluous && !wPreferences.no_animations
 	    && !scr->flags.startup && wwin->transient_for==None
 	    /* 
@@ -1088,8 +1217,8 @@ wManageWindow(WScreen *scr, Window window)
 
     if (!wwin->flags.miniaturized && workspace == scr->current_workspace
 	&& !wwin->flags.hidden) {
-	if ((transientOwner && transientOwner->flags.focused)
-	    || wPreferences.auto_focus)
+	if (((transientOwner && transientOwner->flags.focused)
+	     || wPreferences.auto_focus) && !WFLAGP(wwin, no_focusable))
 	    wSetFocusTo(scr, wwin);
     }
     wWindowResetMouseGrabs(wwin);
@@ -1120,6 +1249,15 @@ wManageWindow(WScreen *scr, Window window)
     }
 #endif
 
+    /*
+     *------------------------------------------------------------
+     * Setup Notification Observers
+     *------------------------------------------------------------
+     */
+    WMAddNotificationObserver(appearanceObserver, wwin,
+			      WNWindowAppearanceSettingsChanged, wwin);
+
+    
     /*
      *--------------------------------------------------
      * 
@@ -1156,6 +1294,9 @@ wManageInternalWindow(WScreen *scr, Window window, Window owner,
 
     wwin = wWindowCreate();
 
+    WMAddNotificationObserver(appearanceObserver, wwin,
+			      WNWindowAppearanceSettingsChanged, wwin);
+
     wwin->flags.internal_window = 1;
 
     WSETUFLAG(wwin, omnipresent, 1);
@@ -1179,15 +1320,22 @@ wManageInternalWindow(WScreen *scr, Window window, Window owner,
     wwin->frame_y = wwin->client.y;
 
     
-    foo = WFF_RIGHT_BUTTON;
+    foo = WFF_RIGHT_BUTTON|WFF_BORDER;
     foo |= WFF_TITLEBAR;
+#ifdef XKB_BUTTON_HINT
+    foo |= WFF_LANGUAGE_BUTTON;
+#endif
 
     wwin->frame = wFrameWindowCreate(scr, WMFloatingLevel,
 				     wwin->frame_x, wwin->frame_y,
-				     width, height, foo,
+				     width, height,
+                     &wPreferences.window_title_clearance, foo,
 				     scr->window_title_texture,
-				     (WTexture**)scr->resizebar_texture,
+				     scr->resizebar_texture,
 				     scr->window_title_pixel, 
+#ifdef DRAWSTRING_PLUGIN
+                     W_STRING_FTITLE,
+#endif
 				     &scr->window_title_gc, 
 				     &scr->title_font);
     
@@ -1207,6 +1355,11 @@ wManageInternalWindow(WScreen *scr, Window window, Window owner,
     wwin->frame->child = wwin;
     
     wwin->frame->workspace = wwin->screen_ptr->current_workspace;
+    
+#ifdef XKB_BUTTON_HINT
+    if (wPreferences.modelock)
+        wwin->frame->on_click_language = windowLanguageClick;
+#endif
         
     wwin->frame->on_click_right = windowCloseClick;
     
@@ -1264,7 +1417,8 @@ wManageInternalWindow(WScreen *scr, Window window, Window owner,
 	wwin->prev = NULL;
     }
 
-    wFrameWindowChangeState(wwin->frame, WS_UNFOCUSED);
+    if (wwin->flags.is_gnustep == 0)
+      wFrameWindowChangeState(wwin->frame, WS_UNFOCUSED);
 
 /*    if (wPreferences.auto_focus)*/
 	wSetFocusTo(scr, wwin);
@@ -1310,8 +1464,7 @@ wUnmanageWindow(WWindow *wwin, Bool restore, Bool destroyed)
 
     /* First close attribute editor window if open */
     if (wwin->flags.inspector_open) {
-        WWindow *pwin = wwin->inspector->frame; /* the inspector window */
-        (*pwin->frame->on_click_right)(NULL, pwin, NULL);
+	wCloseInspectorForWindow(wwin);
     }
 
     /* Close window menu if it's open for this window */
@@ -1331,6 +1484,8 @@ wUnmanageWindow(WWindow *wwin, Bool restore, Bool destroyed)
 
     XUnmapWindow(dpy, frame->window);
 
+    XUnmapWindow(dpy, wwin->client_win);
+
     /* deselect window */
     wSelectWindow(wwin, False);
     
@@ -1377,28 +1532,40 @@ wUnmanageWindow(WWindow *wwin, Bool restore, Bool destroyed)
 	    scr->focused_window->next = NULL;
 	}
 
-	/* if in click to focus mode and the window
-	 * was a transient, focus the owner window  
-	 */
-	tmp = NULL;
 	if (wPreferences.focus_mode==WKF_CLICK) {
-	    tmp = wWindowFor(wwin->transient_for);
-	    if (tmp && (!tmp->flags.mapped || WFLAGP(tmp, no_focusable))) {
-		tmp = NULL;
-	    }
-	}
-	/* otherwise, focus the next one in the focus list */
-	if (!tmp) {
-	    tmp = scr->focused_window;
-	    while (tmp) {
-		if (!WFLAGP(tmp, no_focusable)
-		    && (tmp->flags.mapped || tmp->flags.shaded))
-		    break;
-		tmp = tmp->prev;
-	    }
-	}
-	if (wPreferences.focus_mode==WKF_CLICK) {
-	    newFocusedWindow = tmp;
+
+           /* if in click to focus mode and the window
+            * was a transient, focus the owner window  
+            */
+           tmp = NULL;
+           if (wPreferences.focus_mode==WKF_CLICK) {
+               tmp = wWindowFor(wwin->transient_for);
+               if (tmp && (!tmp->flags.mapped || WFLAGP(tmp, no_focusable))) {
+                   tmp = NULL;
+               }
+           }
+           /* otherwise, focus the next one in the focus list */
+           if (!tmp) {
+              tmp = scr->focused_window;
+              while (tmp) {               /* look for one in the window list first */
+                 if (!WFLAGP(tmp, no_focusable) && !WFLAGP(tmp, skip_window_list)
+                     && (tmp->flags.mapped || tmp->flags.shaded))
+                     break;
+                 tmp = tmp->prev;
+              }
+              if (!tmp) {                 /* if unsuccessful, choose any focusable window */
+                 tmp = scr->focused_window;
+                 while (tmp) {
+                    if (!WFLAGP(tmp, no_focusable)
+                         && (tmp->flags.mapped || tmp->flags.shaded))
+                         break;
+                    tmp = tmp->prev;
+                 }
+              }
+           }
+
+	   newFocusedWindow = tmp;
+
 	} else if (wPreferences.focus_mode==WKF_SLOPPY
 		   || wPreferences.focus_mode==WKF_POINTER) {
             unsigned int mask;
@@ -1437,7 +1604,8 @@ wUnmanageWindow(WWindow *wwin, Bool restore, Bool destroyed)
 
     if (wasFocused) {
 	if (newFocusedWindow != owner && owner) {
-	    wFrameWindowChangeState(owner->frame, WS_UNFOCUSED);
+	    if (wwin->flags.is_gnustep == 0)
+	      wFrameWindowChangeState(owner->frame, WS_UNFOCUSED);
 	}
 	wSetFocusTo(scr, newFocusedWindow);
     }
@@ -1486,20 +1654,19 @@ wWindowFocus(WWindow *wwin, WWindow *owin)
 
 #ifdef KEEP_XKB_LOCK_STATUS
     if (wPreferences.modelock) {
-        if (!wwin->flags.focused) {
-            XkbLockGroup(dpy, XkbUseCoreKbd, wwin->languagemode);
-        }
+        XkbLockGroup(dpy, XkbUseCoreKbd, wwin->frame->languagemode);
     }
 #endif /* KEEP_XKB_LOCK_STATUS */
 
     wwin->flags.semi_focused = 0;
 
-    wFrameWindowChangeState(wwin->frame, WS_FOCUSED);
-
-    wWindowResetMouseGrabs(wwin);
+    if (wwin->flags.is_gnustep == 0)
+      wFrameWindowChangeState(wwin->frame, WS_FOCUSED);
 
     wwin->flags.focused = 1;
 
+    wWindowResetMouseGrabs(wwin);
+
     UpdateSwitchMenu(wwin->screen_ptr, wwin, ACTION_CHANGE_STATE);
 
     if (owin == wwin || !owin)
@@ -1514,6 +1681,8 @@ wWindowFocus(WWindow *wwin, WWindow *owin)
 	return;
     }
 
+    oowner = wWindowFor(owin->transient_for);
+
     /* new window is owner of old window */
     if (wwin == oowner) {
 	wWindowUnfocus(owin);
@@ -1526,7 +1695,6 @@ wWindowFocus(WWindow *wwin, WWindow *owin)
     }
 
     /* new window has same owner of old window */
-    oowner = wWindowFor(owin->transient_for);
     if (oowner == nowner) {
 	/* prevent unfocusing of owner */
 	oowner->flags.semi_focused = 0;
@@ -1546,20 +1714,10 @@ wWindowFocus(WWindow *wwin, WWindow *owin)
 void
 wWindowUnfocus(WWindow *wwin)
 {
-#ifdef KEEP_XKB_LOCK_STATUS
-    static XkbStateRec staterec;
-    if (wPreferences.modelock) {
-        if (wwin->flags.focused) {
-	    XkbGetState(dpy,XkbUseCoreKbd,&staterec);
-	    wwin->languagemode=staterec.compat_state&32?1:0;
-	    XkbLockGroup(dpy,XkbUseCoreKbd,0); /* reset to workspace */
-        }
-    }
-#endif /* KEEP_XKB_LOCK_STATUS */
-
     CloseWindowMenu(wwin->screen_ptr);
 
-    wFrameWindowChangeState(wwin->frame, wwin->flags.semi_focused 
+    if (wwin->flags.is_gnustep == 0)
+      wFrameWindowChangeState(wwin->frame, wwin->flags.semi_focused 
 			    ? WS_PFOCUSED : WS_UNFOCUSED);
 
     if (wwin->transient_for!=None 
@@ -1575,6 +1733,7 @@ wWindowUnfocus(WWindow *wwin)
 	}
     }
     wwin->flags.focused = 0;
+
     wWindowResetMouseGrabs(wwin);
 
     UpdateSwitchMenu(wwin->screen_ptr, wwin, ACTION_CHANGE_STATE);
@@ -1599,79 +1758,97 @@ wWindowUnfocus(WWindow *wwin)
 void
 wWindowConstrainSize(WWindow *wwin, int *nwidth, int *nheight)
 {
-    XSizeHints *sizeh = wwin->normal_hints;
     int width = *nwidth;
     int height = *nheight;
-    int winc = sizeh->width_inc;
-    int hinc = sizeh->height_inc;
+    int winc = 1;
+    int hinc = 1;
+    int minW = 1, minH = 1;
+    int maxW = wwin->screen_ptr->scr_width*2;
+    int maxH = wwin->screen_ptr->scr_height*2;
+    int minAX = -1, minAY = -1;
+    int maxAX = -1, maxAY = -1;
+    int baseW = 0;
+    int baseH = 0;
 
-    if (width < sizeh->min_width)
-      width = sizeh->min_width;
-    if (height < sizeh->min_height)
-      height = sizeh->min_height;
+    if (wwin->normal_hints) {
+	winc = wwin->normal_hints->width_inc;
+	hinc = wwin->normal_hints->height_inc;
+	minW = wwin->normal_hints->min_width;
+	minH = wwin->normal_hints->min_height;
+	maxW = wwin->normal_hints->max_width;
+	maxH = wwin->normal_hints->max_height;
+	if (wwin->normal_hints->flags & PAspect) {
+	    minAX = wwin->normal_hints->min_aspect.x;
+	    minAY = wwin->normal_hints->min_aspect.y;
+	    maxAX = wwin->normal_hints->max_aspect.x;
+	    maxAY = wwin->normal_hints->max_aspect.y;
+	}
+    }
+
+    if (width < minW)
+      width = minW;
+    if (height < minH)
+      height = minH;
+
+    if (width > maxW)
+      width = maxW;
+    if (height > maxH) 
+      height = maxH;
 
-    if (width > sizeh->max_width) 
-      width = sizeh->max_width;
-    if (height > sizeh->max_height) 
-      height = sizeh->max_height;
-    
     /* aspect ratio code borrowed from olwm */
-    if (sizeh->flags & PAspect) {
+    if (minAX > 0) {
 	/* adjust max aspect ratio */
-	if (!(sizeh->max_aspect.x==1 && sizeh->max_aspect.y==1)
-	    && width * sizeh->max_aspect.y > height * sizeh->max_aspect.x) {
-	    if (sizeh->max_aspect.x > sizeh->max_aspect.y) {
-		height = (width * sizeh->max_aspect.y) / sizeh->max_aspect.x;
-		if (height > sizeh->max_height) {
-		    height = sizeh->max_height;
-		    width = (height*sizeh->max_aspect.x) / sizeh->max_aspect.y;
+	if (!(maxAX == 1 && maxAY == 1) && width * maxAY > height * maxAX) {
+	    if (maxAX > maxAY) {
+		height = (width * maxAY) / maxAX;
+		if (height > maxH) {
+		    height = maxH;
+		    width = (height * maxAX) / maxAY;
 		}
 	    } else {
-		width = (height * sizeh->max_aspect.x) / sizeh->max_aspect.y;
-		if (width > sizeh->max_width) {
-		    width = sizeh->max_width;
-		    height = (width*sizeh->max_aspect.y) / sizeh->max_aspect.x;
+		width = (height * maxAX) / maxAY;
+		if (width > maxW) {
+		    width = maxW;
+		    height = (width * maxAY) / maxAX;
 		}
 	    }
 	}
 
 	/* adjust min aspect ratio */
-	if (!(sizeh->min_aspect.x==1 && sizeh->min_aspect.y==1)
-	    && width * sizeh->min_aspect.y < height * sizeh->min_aspect.x) {
-	    if (sizeh->min_aspect.x > sizeh->min_aspect.y) {
-		height = (width * sizeh->min_aspect.y) / sizeh->min_aspect.x;
-		if (height < sizeh->min_height) {
-		    height = sizeh->min_height;
-		    width = (height*sizeh->min_aspect.x) / sizeh->min_aspect.y;
+	if (!(minAX == 1 && minAY == 1) && width * minAY < height * minAX) {
+	    if (minAX > minAY) {
+		height = (width * minAY) / minAX;
+		if (height < minH) {
+		    height = minH;
+		    width = (height * minAX) / minAY;
 		}
 	    } else {
-		width = (height * sizeh->min_aspect.x) / sizeh->min_aspect.y;
-		if (width > sizeh->min_width) {
-		    width = sizeh->min_width;
-		    height = (width*sizeh->min_aspect.y) / sizeh->min_aspect.x;
+		width = (height * minAX) / minAY;
+		if (width < minW) {
+		    width = minW;
+		    height = (width * minAY) / minAX;
 		}
 	    }
 	}
     }
 
-    if (sizeh->base_width != 0) {
-	width = (((width - sizeh->base_width) / winc) * winc) 
-	    + sizeh->base_width;
+    if (baseW != 0) {
+	width = (((width - baseW) / winc) * winc) + baseW;
     } else {
-	width = (((width - sizeh->min_width) / winc) * winc)
-	  + sizeh->min_width;
+	width = (((width - minW) / winc) * winc) + minW;
     }
 
-    if (sizeh->base_width != 0) {
-	height = (((height - sizeh->base_height) / hinc) * hinc)
-	  + sizeh->base_height;
+    if (baseW != 0) {
+	height = (((height - baseH) / hinc) * hinc) + baseH;
     } else {
-	height = (((height - sizeh->min_height) / hinc) * hinc) 
-	  + sizeh->min_height;
+	height = (((height - minH) / hinc) * hinc) + minH;
     }
-    
-    *nwidth = width;
-    *nheight = height;
+
+    /* broken stupid apps may cause preposterous values for these.. */
+    if (width > 0)
+	*nwidth = width;
+    if (height > 0)
+	*nheight = height;
 }
 
 
@@ -1698,8 +1875,10 @@ wWindowChangeWorkspace(WWindow *wwin, int workspace)
 		wapp->last_workspace = workspace;
 	    }
 	    if (wwin->flags.miniaturized) {
-		XUnmapWindow(dpy, wwin->icon->core->window);
-		wwin->icon->mapped = 0;
+		if (wwin->icon) {
+		    XUnmapWindow(dpy, wwin->icon->core->window);
+		    wwin->icon->mapped = 0;
+		}
 	    } else {
 		unmap = 1;
 		wSetFocusTo(scr, NULL);
@@ -1708,8 +1887,10 @@ wWindowChangeWorkspace(WWindow *wwin, int workspace)
     } else {
 	/* brought to current workspace. Map window */
 	if (wwin->flags.miniaturized && !wPreferences.sticky_icons) {
-	    XMapWindow(dpy, wwin->icon->core->window);
-	    wwin->icon->mapped = 1;
+	    if (wwin->icon) {
+		XMapWindow(dpy, wwin->icon->core->window);
+		wwin->icon->mapped = 1;
+	    }
 	} else if (!wwin->flags.mapped && 
 	    !(wwin->flags.miniaturized || wwin->flags.hidden)) {
 	    wWindowMap(wwin);
@@ -1846,6 +2027,10 @@ int req_width, req_height;	       /* new size of the client */
     }
     wwin->frame_x = req_x;
     wwin->frame_y = req_y;
+    if (!WFLAGP(wwin, no_border)) {
+	wwin->client.x += FRAME_BORDER_WIDTH;
+	wwin->client.y += FRAME_BORDER_WIDTH;
+    }
 
 #ifdef SHAPE
     if (wShapeSupported && wwin->flags.shaped && resize) {
@@ -1882,6 +2067,10 @@ int req_x, req_y;		       /* new position of the frame */
 
     wwin->client.x = req_x;
     wwin->client.y = req_y + wwin->frame->top_width;
+    if (!WFLAGP(wwin, no_border)) {
+	wwin->client.x += FRAME_BORDER_WIDTH;
+	wwin->client.y += FRAME_BORDER_WIDTH;
+    }
 
     XMoveWindow(dpy, wwin->frame->core->window, req_x, req_y);
 
@@ -1895,36 +2084,36 @@ int req_x, req_y;		       /* new position of the frame */
 }
 
 
-void 
+void
 wWindowUpdateButtonImages(WWindow *wwin)
 {
     WScreen *scr = wwin->screen_ptr;
     Pixmap pixmap, mask;
     WFrameWindow *fwin = wwin->frame;
-    
+
     if (WFLAGP(wwin, no_titlebar))
 	return;
 
     /* miniaturize button */
-    
+
     if (!WFLAGP(wwin, no_miniaturize_button)) {
 	if (wwin->wm_gnustep_attr
 	    && wwin->wm_gnustep_attr->flags & GSMiniaturizePixmapAttr) {
 	    pixmap = wwin->wm_gnustep_attr->miniaturize_pixmap;
-	      
+	
 	    if (wwin->wm_gnustep_attr->flags&GSMiniaturizeMaskAttr) {
 		mask = wwin->wm_gnustep_attr->miniaturize_mask;
 	    } else {
 		mask = None;
 	    }
 
-	    if (fwin->lbutton_image 
+	    if (fwin->lbutton_image
 		&& (fwin->lbutton_image->image != pixmap
 		    || fwin->lbutton_image->mask != mask)) {
 		wPixmapDestroy(fwin->lbutton_image);
 		fwin->lbutton_image = NULL;
 	    }
-	    
+	
 	    if (!fwin->lbutton_image) {
 		fwin->lbutton_image = wPixmapCreate(scr, pixmap, mask);
 		fwin->lbutton_image->client_owned = 1;
@@ -1937,46 +2126,64 @@ wWindowUpdateButtonImages(WWindow *wwin)
 	    fwin->lbutton_image = scr->b_pixmaps[WBUT_ICONIFY];
 	}
     }
-    
+
+#ifdef XKB_BUTTON_HINT
+    if (!WFLAGP(wwin, no_language_button)) {
+	    if (fwin->languagebutton_image &&  
+		!fwin->languagebutton_image->shared) {
+		wPixmapDestroy(fwin->languagebutton_image);
+	    }
+	fwin->languagebutton_image = 
+	    scr->b_pixmaps[WBUT_XKBGROUP1 + fwin->languagemode];
+    }
+#endif
+
     /* close button */
 
+/* redefine WFLAGP to MGFLAGP to allow broken close operation */
+#define MGFLAGP(wwin, FLAG)	(wwin)->client_flags.FLAG
+
     if (!WFLAGP(wwin, no_close_button)) {
-	if (wwin->wm_gnustep_attr
-	    && wwin->wm_gnustep_attr->flags&GSClosePixmapAttr) {
-	    pixmap = wwin->wm_gnustep_attr->close_pixmap;
-	    
-	    if (wwin->wm_gnustep_attr->flags&GSCloseMaskAttr) {
-		mask = wwin->wm_gnustep_attr->close_mask;
-	    } else {
-		mask = None;
-	    }
+        if (wwin->wm_gnustep_attr 
+	    && wwin->wm_gnustep_attr->flags & GSClosePixmapAttr) {
+            pixmap = wwin->wm_gnustep_attr->close_pixmap;
+	
+            if (wwin->wm_gnustep_attr->flags&GSCloseMaskAttr)
+                 mask = wwin->wm_gnustep_attr->close_mask;
+            else
+                 mask = None;
+
+           if (fwin->rbutton_image && (fwin->rbutton_image->image != pixmap
+				       || fwin->rbutton_image->mask != mask)) {
+	       wPixmapDestroy(fwin->rbutton_image);
+	       fwin->rbutton_image = NULL;
+	   }
+			
+           if (!fwin->rbutton_image) {
+               fwin->rbutton_image = wPixmapCreate(scr, pixmap, mask);
+               fwin->rbutton_image->client_owned = 1;
+               fwin->rbutton_image->client_owned_mask = 1;
+	   }
 
-	    if (fwin->rbutton_image
-		&& (fwin->rbutton_image->image != pixmap
-		    || fwin->rbutton_image->mask != mask)) {
-		wPixmapDestroy(fwin->rbutton_image);
-		fwin->rbutton_image = NULL;
-	    }
-	    
-	    if (!fwin->rbutton_image) {
-		fwin->rbutton_image = wPixmapCreate(scr, pixmap, mask);
-		fwin->rbutton_image->client_owned = 1;
-		fwin->rbutton_image->client_owned_mask = 1;
-	    }
 	} else if (WFLAGP(wwin, kill_close)) {
-	    if (fwin->rbutton_image && !fwin->rbutton_image->shared) {
+
+	    if (fwin->rbutton_image && !fwin->rbutton_image->shared)
 		wPixmapDestroy(fwin->rbutton_image);
-	    }
+
 	    fwin->rbutton_image = scr->b_pixmaps[WBUT_KILL];
-	} else if (WFLAGP(wwin, broken_close)) {
-	    if (fwin->rbutton_image && !fwin->rbutton_image->shared) {
+	    
+	} else if (MGFLAGP(wwin, broken_close)) {
+
+	    if (fwin->rbutton_image && !fwin->rbutton_image->shared) 
 		wPixmapDestroy(fwin->rbutton_image);
-	    }
+
 	    fwin->rbutton_image = scr->b_pixmaps[WBUT_BROKENCLOSE];
+
 	} else {
-	    if (fwin->rbutton_image && !fwin->rbutton_image->shared) {
+
+	    if (fwin->rbutton_image && !fwin->rbutton_image->shared) 
 		wPixmapDestroy(fwin->rbutton_image);
-	    }
+
 	    fwin->rbutton_image = scr->b_pixmaps[WBUT_CLOSE];
 	}
     }
@@ -2002,10 +2209,15 @@ wWindowConfigureBorders(WWindow *wwin)
 	int newy, oldh;
 
 	flags = WFF_LEFT_BUTTON|WFF_RIGHT_BUTTON;
+#ifdef XKB_BUTTON_HINT
+    flags |= WFF_LANGUAGE_BUTTON;
+#endif
 	if (!WFLAGP(wwin, no_titlebar))
 	    flags |= WFF_TITLEBAR;
 	if (!WFLAGP(wwin, no_resizebar))
 	    flags |= WFF_RESIZEBAR;
+	if (!WFLAGP(wwin, no_border))
+	    flags |= WFF_BORDER;
 	if (wwin->flags.shaded)
 	    flags |= WFF_IS_SHADED;
 
@@ -2024,6 +2236,12 @@ wWindowConfigureBorders(WWindow *wwin)
 	    && wwin->frame->flags.hide_left_button)
 	    flags |= WFF_LEFT_BUTTON;
 
+#ifdef XKB_BUTTON_HINT
+    if (!WFLAGP(wwin, no_language_button)
+            && wwin->frame->flags.hide_language_button)
+        flags |= WFF_LANGUAGE_BUTTON;
+#endif
+
 	if (!WFLAGP(wwin, no_close_button)
 	    && wwin->frame->flags.hide_right_button)
 	    flags |= WFF_RIGHT_BUTTON;
@@ -2038,6 +2256,12 @@ wWindowConfigureBorders(WWindow *wwin)
 	    && !wwin->frame->flags.hide_left_button)
 	    flags |= WFF_LEFT_BUTTON;
 
+#ifdef XKB_BUTTON_HINT
+	if (WFLAGP(wwin, no_language_button)
+	    && !wwin->frame->flags.hide_language_button)
+	    flags |= WFF_LANGUAGE_BUTTON;
+#endif
+
 	if (WFLAGP(wwin, no_close_button)
 	    && !wwin->frame->flags.hide_right_button)
 	    flags |= WFF_RIGHT_BUTTON;
@@ -2057,17 +2281,23 @@ wWindowConfigureBorders(WWindow *wwin)
 void
 wWindowSaveState(WWindow *wwin)
 {
-    CARD32 data[9];
-    
-    memset(data, 0, sizeof(CARD32)*9);
+    CARD32 data[10];
+    int i;
+
+    memset(data, 0, sizeof(CARD32)*10);
     data[0] = wwin->frame->workspace;
     data[1] = wwin->flags.miniaturized;
     data[2] = wwin->flags.shaded;
     data[3] = wwin->flags.hidden;
 
+    for (i = 0; i < MAX_WINDOW_SHORTCUTS; i++) {
+	if (wwin->screen_ptr->shortcutWindows[i] &&
+	    WMCountInBag(wwin->screen_ptr->shortcutWindows[i], wwin))
+	    data[9] |= 1<<i;
+    }
     XChangeProperty(dpy, wwin->client_win, _XA_WINDOWMAKER_STATE,
 		    _XA_WINDOWMAKER_STATE, 32, PropModeReplace,
-		    (unsigned char *)data, 9);
+		    (unsigned char *)data, 10);
 }
 
 
@@ -2080,7 +2310,7 @@ getSavedState(Window window, WSavedState **state)
     unsigned long bytes_after_ret;
     CARD32 *data;
     
-    if (XGetWindowProperty(dpy, window, _XA_WINDOWMAKER_STATE, 0, 9,
+    if (XGetWindowProperty(dpy, window, _XA_WINDOWMAKER_STATE, 0, 10,
 			   True, _XA_WINDOWMAKER_STATE,
 			   &type_ret, &fmt_ret, &nitems_ret, &bytes_after_ret,
 			   (unsigned char **)&data)!=Success || !data)
@@ -2098,6 +2328,7 @@ getSavedState(Window window, WSavedState **state)
 	(*state)->y = data[6];
 	(*state)->w = data[7];
 	(*state)->h = data[8];
+	(*state)->window_shortcuts = data[9];
     }
     XFree(data);
     
@@ -2282,7 +2513,8 @@ wWindowResetMouseGrabs(WWindow *wwin)
 			  GrabModeAsync, None, None);
     }
     
-    if (!wwin->flags.focused) {
+    if (!wwin->flags.focused && !WFLAGP(wwin, no_focusable)
+	&& !wwin->flags.is_gnustep) {
 	/* the passive grabs to focus the window */
 	if (wPreferences.focus_mode == WKF_CLICK)
 	    XGrabButton(dpy, AnyButton, AnyModifier, wwin->client_win,
@@ -2296,17 +2528,13 @@ wWindowResetMouseGrabs(WWindow *wwin)
 void
 wWindowUpdateGNUstepAttr(WWindow *wwin, GNUstepWMAttributes *attr)
 {
-    
-    if (attr->flags & GSExtraFlagsAttr) {
-	if (WFLAGP(wwin, broken_close) != 
-	    (attr->extra_flags & GSDocumentEditedFlag)) {
-
-	    wwin->client_flags.broken_close = !WFLAGP(wwin, broken_close);
-
-	    wWindowUpdateButtonImages(wwin);
-	}
-    }
-    
+   if (attr->flags & GSExtraFlagsAttr) {
+       if (MGFLAGP(wwin, broken_close) != 
+	   (attr->extra_flags & GSDocumentEditedFlag)) {
+	   wwin->client_flags.broken_close = !MGFLAGP(wwin, broken_close);
+	   wWindowUpdateButtonImages(wwin);
+       }
+   }
 }
 
 
@@ -2494,7 +2722,8 @@ resizebarMouseDown(WCoreWindow *sender, void *data, XEvent *event)
     CloseWindowMenu(wwin->screen_ptr);
 
     if (wPreferences.focus_mode==WKF_CLICK 
-	&& !(event->xbutton.state&ControlMask)) {
+	&& !(event->xbutton.state&ControlMask)
+	&& !WFLAGP(wwin, no_focusable)) {
 	wSetFocusTo(wwin->screen_ptr, wwin);
     }
 
@@ -2581,7 +2810,8 @@ frameMouseDown(WObjDescriptor *desc, XEvent *event)
     CloseWindowMenu(wwin->screen_ptr);
 
     if (wPreferences.focus_mode==WKF_CLICK
-	&& !(event->xbutton.state&ControlMask)) {
+	&& !(event->xbutton.state&ControlMask)
+	&& !WFLAGP(wwin, no_focusable)) {
 	wSetFocusTo(wwin->screen_ptr, wwin);
     }
     if (event->xbutton.button == Button1) {
@@ -2625,7 +2855,8 @@ titlebarMouseDown(WCoreWindow *sender, void *data, XEvent *event)
     CloseWindowMenu(wwin->screen_ptr);
 
     if (wPreferences.focus_mode==WKF_CLICK 
-	&& !(event->xbutton.state&ControlMask)) {
+	&& !(event->xbutton.state&ControlMask)
+	&& !WFLAGP(wwin, no_focusable)) {
 	wSetFocusTo(wwin->screen_ptr, wwin);
     }
 
@@ -2744,6 +2975,33 @@ windowCloseDblClick(WCoreWindow *sender, void *data, XEvent *event)
 }
 
 
+#ifdef XKB_BUTTON_HINT
+static void
+windowLanguageClick(WCoreWindow *sender, void *data, XEvent *event)
+{
+    WWindow *wwin = data;
+    WFrameWindow *fwin = wwin->frame;
+    WScreen *scr = fwin->screen_ptr;
+    XkbStateRec staterec;
+    int tl;
+
+    if (event->xbutton.button != Button1 && event->xbutton.button != Button3)
+        return;
+    tl = wwin->frame->languagemode;
+    wwin->frame->languagemode = wwin->frame->last_languagemode;
+    wwin->frame->last_languagemode = tl;
+    wSetFocusTo(scr, wwin);
+    wwin->frame->languagebutton_image = 
+        wwin->frame->screen_ptr->b_pixmaps[WBUT_XKBGROUP1 + 
+        wwin->frame->languagemode];
+    wFrameWindowUpdateLanguageButton(wwin->frame);
+    if (event->xbutton.button == Button3)
+	return;
+    wRaiseFrame(fwin->core);
+}
+#endif
+
+
 static void
 windowIconifyClick(WCoreWindow *sender, void *data, XEvent *event)
 {
