@@ -23,11 +23,16 @@
 
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
+#ifdef SHAPE
+#include <X11/extensions/shape.h>
+#endif
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <unistd.h>
 #include <ctype.h>
 #include <string.h>
+#include <time.h>
 
 #include "WindowMaker.h"
 #include "wcore.h"
@@ -58,6 +63,9 @@ extern XContext wWinContext;
 static proplist_t dWorkspaces=NULL;
 static proplist_t dClip, dName;
 
+#ifdef VIRTUAL_DESKTOP
+static BOOL initVDesk=False;
+#endif
 
 static void
 make_keys()
@@ -115,7 +123,8 @@ wWorkspaceNew(WScreen *scr)
 	    list[i] = scr->workspaces[i];
 	}
 	list[i] = wspace;
-	free(scr->workspaces);
+	if (scr->workspaces)
+	    free(scr->workspaces);
 	scr->workspaces = list;
 
 	wWorkspaceMenuUpdate(scr, scr->workspace_menu);
@@ -220,16 +229,225 @@ wWorkspaceDelete(WScreen *scr, int workspace)
 }
 
 
+typedef struct WorkspaceNameData {
+    int count;
+    RImage *back;
+    RImage *text;
+    time_t timeout;
+} WorkspaceNameData;
+
+
+
+static void
+hideWorkpaceName(void *data)
+{
+    WScreen *scr = (WScreen*)data;
+
+    if (!scr->workspace_name_data || scr->workspace_name_data->count == 0
+	|| time(NULL) > scr->workspace_name_data->timeout) {
+	XUnmapWindow(dpy, scr->workspace_name);
+
+	if (scr->workspace_name_data) {
+	    RDestroyImage(scr->workspace_name_data->back);
+	    RDestroyImage(scr->workspace_name_data->text);
+	    free(scr->workspace_name_data);
+
+	    scr->workspace_name_data = NULL;
+	}
+	scr->workspace_name_timer = NULL;
+    } else {
+	RImage *img = RCloneImage(scr->workspace_name_data->back);
+	Pixmap pix;
+
+	scr->workspace_name_timer = 
+	    WMAddTimerHandler(WORKSPACE_NAME_FADE_DELAY, hideWorkpaceName,
+			      scr);
+
+	RCombineImagesWithOpaqueness(img, scr->workspace_name_data->text,
+				     scr->workspace_name_data->count*255/10);
+
+	RConvertImage(scr->rcontext, img, &pix);
+
+	RDestroyImage(img);
+
+	XSetWindowBackgroundPixmap(dpy, scr->workspace_name, pix);
+	XClearWindow(dpy, scr->workspace_name);
+	XFreePixmap(dpy, pix);
+	XFlush(dpy);
+
+	scr->workspace_name_data->count--;
+    }
+}
+
+
+
+static void
+showWorkspaceName(WScreen *scr, int workspace)
+{
+    WorkspaceNameData *data;
+    RXImage *ximg;
+    Pixmap text, mask;
+    int w, h;
+    int px, py;
+    char *name = scr->workspaces[workspace]->name;
+    int len = strlen(name);
+    int x, y;
+
+    if (wPreferences.workspace_name_display_position == WD_NONE
+	|| scr->workspace_count < 2)
+	return;
+
+    if (scr->workspace_name_timer) {
+	WMDeleteTimerHandler(scr->workspace_name_timer);
+	XUnmapWindow(dpy, scr->workspace_name);
+	XFlush(dpy);
+    }
+    scr->workspace_name_timer = WMAddTimerHandler(WORKSPACE_NAME_DELAY,
+						  hideWorkpaceName, scr);
+
+    if (scr->workspace_name_data) {
+	RDestroyImage(scr->workspace_name_data->back);
+	RDestroyImage(scr->workspace_name_data->text);
+	free(scr->workspace_name_data);
+    }
+
+    data = wmalloc(sizeof(WorkspaceNameData));
+
+    w = WMWidthOfString(scr->workspace_name_font, name, len);
+    h = WMFontHeight(scr->workspace_name_font);
+
+    switch (wPreferences.workspace_name_display_position) {
+     case WD_TOP:
+	px = (scr->scr_width - (w+4))/2;
+	py = 0;
+	break;
+     case WD_BOTTOM:
+	px = (scr->scr_width - (w+4))/2;
+	py = scr->scr_height - (h+4);
+	break;
+     case WD_TOPLEFT:
+	px = 0;
+	py = 0;
+	break;
+     case WD_TOPRIGHT:
+	px = scr->scr_width - (w+4);
+	py = 0;
+	break;
+     case WD_BOTTOMLEFT:
+	px = 0;
+	py = scr->scr_height - (h+4);
+	break;
+     case WD_BOTTOMRIGHT:
+	px = scr->scr_width - (w+4);
+	py = scr->scr_height - (h+4);
+	break;
+     case WD_CENTER:
+     default:
+	px = (scr->scr_width - (w+4))/2;
+	py = (scr->scr_height - (h+4))/2;
+	break;
+    }
+    XResizeWindow(dpy, scr->workspace_name, w+4, h+4);
+    XMoveWindow(dpy, scr->workspace_name, px, py);
+
+    text = XCreatePixmap(dpy, scr->w_win, w+4, h+4, scr->w_depth);
+    mask = XCreatePixmap(dpy, scr->w_win, w+4, h+4, 1);
+
+    XSetForeground(dpy, scr->draw_gc, scr->black_pixel);
+    XFillRectangle(dpy, text, scr->draw_gc, 0, 0, w+4, h+4);
+
+    XSetForeground(dpy, scr->mono_gc, 0);
+    XFillRectangle(dpy, mask, scr->mono_gc, 0, 0, w+4, h+4);
+
+    XSetForeground(dpy, scr->mono_gc, 1);
+    for (x = 0; x <= 4; x++) {
+	for (y = 0; y <= 4; y++) {
+	    WMDrawString(scr->wmscreen, mask, scr->mono_gc,
+			 scr->workspace_name_font, x, y, name, len);
+	}
+    }
+
+    XSetForeground(dpy, scr->draw_gc, scr->white_pixel);
+    WMDrawString(scr->wmscreen, text, scr->draw_gc, scr->workspace_name_font, 
+		 2, 2, scr->workspaces[workspace]->name,
+		 strlen(scr->workspaces[workspace]->name));
+#ifdef SHAPE
+    XShapeCombineMask(dpy, scr->workspace_name, ShapeBounding, 0, 0, mask,
+		      ShapeSet);
+#endif
+    XSetWindowBackgroundPixmap(dpy, scr->workspace_name, text);
+    XClearWindow(dpy, scr->workspace_name);
+
+    data->text = RCreateImageFromDrawable(scr->rcontext, text, None);
+
+    XFreePixmap(dpy, text);
+    XFreePixmap(dpy, mask);
+
+    if (!data->text) {
+	XMapRaised(dpy, scr->workspace_name);
+	XFlush(dpy);
+
+	goto erro;
+    }
+
+    ximg = RGetXImage(scr->rcontext, scr->root_win, px, py,
+		      data->text->width, data->text->height);
+
+    if (!ximg || !ximg->image) {
+	goto erro;
+    }
+
+    XMapRaised(dpy, scr->workspace_name);
+    XFlush(dpy);
+
+    data->back = RCreateImageFromXImage(scr->rcontext, ximg->image, NULL);
+    RDestroyXImage(scr->rcontext, ximg);
+
+    if (!data->back) {
+	goto erro;
+    }
+
+    data->count = 10;
+
+    /* set a timeout for the effect */
+    data->timeout = time(NULL) + 2 +
+	(WORKSPACE_NAME_DELAY + WORKSPACE_NAME_FADE_DELAY*data->count)/1000;
+
+    scr->workspace_name_data = data;
+
+
+    return;
+
+erro:
+    if (scr->workspace_name_timer)
+	WMDeleteTimerHandler(scr->workspace_name_timer);
+
+    if (data->text)
+	RDestroyImage(data->text);
+    if (data->back)
+	RDestroyImage(data->back);
+    free(data);
+
+    scr->workspace_name_data = NULL;
+
+    scr->workspace_name_timer = WMAddTimerHandler(WORKSPACE_NAME_DELAY +
+						  10*WORKSPACE_NAME_FADE_DELAY,
+						  hideWorkpaceName, scr);
+}
+
+
 void
 wWorkspaceChange(WScreen *scr, int workspace)
 {
     if (scr->flags.startup || scr->flags.startup2) {
 	return;
     }
-    
+
     if (workspace != scr->current_workspace) {
         wWorkspaceForceChange(scr, workspace);
-    }
+    } /*else {
+	showWorkspaceName(scr, workspace);
+    }*/
 }
 
 
@@ -283,8 +501,14 @@ wWorkspaceForceChange(WScreen *scr, int workspace)
     wWorkspaceMenuUpdate(scr, scr->clip_ws_menu);
 
     if ((tmp = scr->focused_window)!= NULL) {
-	if (IS_OMNIPRESENT(tmp) || tmp->flags.changing_workspace)
+	if ((IS_OMNIPRESENT(tmp) && !WFLAGP(tmp, skip_window_list))
+	    || tmp->flags.changing_workspace) {
 	    foc = tmp;
+	}
+ 
+	/* foc2 = tmp; will fix annoyance with gnome panel
+	 * but will create annoyance for every other application 
+	 */
 
 	while (tmp) {
 	    if (tmp->frame->workspace!=workspace && !tmp->flags.selected) {
@@ -301,11 +525,14 @@ wWorkspaceForceChange(WScreen *scr, int workspace)
                     if (!wPreferences.sticky_icons) {
                         XUnmapWindow(dpy, tmp->icon->core->window);
 			tmp->icon->mapped = 0;
-		    } else {
+		    }
+#if 0
+		    else {
 			tmp->icon->mapped = 1;
 			/* Why is this here? -Alfredo */
                         XMapWindow(dpy, tmp->icon->core->window);
 		    }
+#endif
                 }
 		/* update current workspace of omnipresent windows */
 		if (IS_OMNIPRESENT(tmp)) {
@@ -331,7 +558,7 @@ wWorkspaceForceChange(WScreen *scr, int workspace)
 			if (!(tmp->flags.mapped || tmp->flags.miniaturized)) {
 			    /* remap windows that are on this workspace */
 			    wWindowMap(tmp);
-			    if (!foc)
+			    if (!foc && !WFLAGP(tmp, skip_window_list))
 				foc = tmp;
 			}
 			/* Also map miniwindow if not omnipresent */
@@ -384,6 +611,7 @@ wWorkspaceForceChange(WScreen *scr, int workspace)
 
     if (scr->dock)
         wAppIconPaint(scr->dock->icon_array[0]);
+
     if (scr->clip_icon) {
         if (scr->workspaces[workspace]->clip->auto_collapse ||
             scr->workspaces[workspace]->clip->auto_raise_lower) {
@@ -395,6 +623,8 @@ wWorkspaceForceChange(WScreen *scr, int workspace)
         }
     }
 
+    showWorkspaceName(scr, workspace);
+
 #ifdef GNOME_STUFF
     wGNOMEUpdateCurrentWorkspaceHint(scr);
 #endif
@@ -404,6 +634,102 @@ wWorkspaceForceChange(WScreen *scr, int workspace)
 /*   XSync(dpy, False); */
 }
 
+#ifdef VIRTUAL_DESKTOP
+
+void wWorkspaceManageEdge(WScreen *scr) {
+    int i;
+    int vmask;
+    XSetWindowAttributes attribs;
+
+    puts("ok");
+    if (wPreferences.vedge_thickness) {
+        initVDesk=True;
+        for (i = 0;i < scr->workspace_count; i++) {
+            puts("reset workspace");
+            scr->workspaces[i]->x = scr->workspaces[i]->y = 0;
+            wWorkspaceResizeViewPort(scr, i, wPreferences.vedge_width, wPreferences.vedge_height);
+        }
+
+        vmask = CWEventMask|CWOverrideRedirect;
+        attribs.event_mask = (EnterWindowMask | LeaveWindowMask | VisibilityChangeMask);
+        attribs.override_redirect = True;
+        scr->virtual_edge_u =
+            XCreateWindow(dpy, scr->root_win, 0, 0,
+                    scr->scr_width, wPreferences.vedge_thickness, 0,
+                    CopyFromParent, InputOnly, CopyFromParent, vmask, &attribs);
+        scr->virtual_edge_d =
+            XCreateWindow(dpy, scr->root_win, 0, scr->scr_height-wPreferences.vedge_thickness,
+                    scr->scr_width, wPreferences.vedge_thickness, 0,
+                    CopyFromParent, InputOnly, CopyFromParent, vmask, &attribs);
+        scr->virtual_edge_l =
+            XCreateWindow(dpy, scr->root_win, 0, 0,
+                    wPreferences.vedge_thickness, scr->scr_height, 0,
+                    CopyFromParent, InputOnly, CopyFromParent, vmask, &attribs);
+        scr->virtual_edge_r =
+            XCreateWindow(dpy, scr->root_win, scr->scr_width-wPreferences.vedge_thickness, 0,
+                    wPreferences.vedge_thickness, scr->scr_height, 0,
+                    CopyFromParent, InputOnly, CopyFromParent, vmask, &attribs);
+        XMapWindow(dpy, scr->virtual_edge_u);
+        XMapWindow(dpy, scr->virtual_edge_d);
+        XMapWindow(dpy, scr->virtual_edge_l);
+        XMapWindow(dpy, scr->virtual_edge_r);
+        wWorkspaceRaiseEdge(scr);
+    }
+}
+
+void wWorkspaceRaiseEdge(WScreen *scr) {
+    if (wPreferences.vedge_thickness && initVDesk) {
+        XRaiseWindow(dpy, scr->virtual_edge_u);
+        XRaiseWindow(dpy, scr->virtual_edge_d);
+        XRaiseWindow(dpy, scr->virtual_edge_l);
+        XRaiseWindow(dpy, scr->virtual_edge_r);
+    }
+}
+
+void wWorkspaceResizeViewPort(WScreen *scr, int workspace, int width, int height)
+{
+    if (width < scr->scr_width) return;
+    if (height < scr->scr_height) return;
+
+    scr->workspaces[workspace]->width = WMAX(width,scr->scr_width);
+    scr->workspaces[workspace]->height = WMAX(height,scr->scr_height);
+
+}
+
+void wWorkspaceSetViewPort(WScreen *scr, int workspace, int x, int y)
+{
+    int diff_x, diff_y;
+    WWindow *wwin;
+
+    if (x < 0) return;
+    if (y < 0) return;
+    if (x + scr->scr_width > scr->workspaces[workspace]->width) return;
+    if (y + scr->scr_height > scr->workspaces[workspace]->height) return;
+
+    diff_x = scr->workspaces[workspace]->x - x;
+    diff_y = scr->workspaces[workspace]->y - y;
+
+    scr->workspaces[workspace]->x = WMIN(x, scr->workspaces[workspace]->width - scr->scr_width);
+    scr->workspaces[workspace]->y = WMIN(y, scr->workspaces[workspace]->height - scr->scr_height);
+
+    printf("set view %d %d, %d\n",workspace, scr->workspaces[workspace]->x, scr->workspaces[workspace]->y);
+
+    wwin = scr->focused_window;
+    while (wwin) {
+        if (wwin->frame->workspace == workspace) {
+            wWindowMove(wwin, wwin->frame_x + diff_x, wwin->frame_y + diff_y);
+            wWindowSynthConfigureNotify(wwin);
+        }
+        wwin = wwin->prev;
+    }
+}
+
+void wWorkspaceGetViewPosition(WScreen *scr, int workspace, int *x, int *y) {
+    *x = scr->workspaces[workspace]->x;
+    *y = scr->workspaces[workspace]->y;
+}
+
+#endif
 
 static void
 switchWSCommand(WMenu *menu, WMenuEntry *entry)
@@ -646,7 +972,7 @@ wWorkspaceRestoreState(WScreen *scr)
 {
     proplist_t parr, pstr, wks_state;
     proplist_t clip_state;
-    int i, wscount;
+    int i, j, wscount;
 
     make_keys();
 
@@ -656,7 +982,7 @@ wWorkspaceRestoreState(WScreen *scr)
 	return;
     
     wscount = scr->workspace_count;
-    for (i=0; i < PLGetNumberOfElements(parr); i++) {
+    for (i=0; i < WMIN(PLGetNumberOfElements(parr), MAX_WORKSPACES); i++) {
         wks_state = PLGetArrayElement(parr, i);
         if (PLIsDictionary(wks_state))
             pstr = PLGetDictionaryEntry(wks_state, dName);
@@ -676,9 +1002,24 @@ wWorkspaceRestoreState(WScreen *scr)
             if (scr->workspaces[i]->clip)
                 wDockDestroy(scr->workspaces[i]->clip);
             scr->workspaces[i]->clip = wDockRestoreState(scr, clip_state,
-                                                          WM_CLIP);
+                                                         WM_CLIP);
             if (i>0)
                 wDockHideIcons(scr->workspaces[i]->clip);
+
+            /* We set the global icons here, because scr->workspaces[i]->clip
+             * was not valid in wDockRestoreState().
+             * There we only set icon->omnipresent to know which icons we
+             * need to set here.
+             */
+            for (j=0; j<scr->workspaces[i]->clip->max_icons; j++) {
+                WAppIcon *aicon = scr->workspaces[i]->clip->icon_array[j];
+
+                if (aicon && aicon->omnipresent) {
+                    aicon->omnipresent = 0;
+                    wClipMakeIconOmnipresent(aicon, True);
+                    XMapWindow(dpy, aicon->icon->core->window);
+                }
+            }
         }
 #ifdef KWM_HINTS
 	wKWMUpdateWorkspaceNameHint(scr, i);
