@@ -38,15 +38,20 @@
 #include <sys/types.h>
 #include <ctype.h>
 
+#include "../src/config.h"
+
+#ifdef HAVE_DLFCN_H
+#include <dlfcn.h>
+#endif
+
 #include "../src/wconfig.h"
 
-#include "../WINGs/WINGs.h"
-#include "../WINGs/WUtil.h"
-#include "../wrlib/wraster.h"
+#include <WINGs.h>
+#include <wraster.h>
 
 #include <proplist.h>
 
-#define PROG_VERSION	"wmsetbg (Window Maker) 2.1"
+#define PROG_VERSION	"wmsetbg (Window Maker) 2.6"
 
 
 #define WORKSPACE_COUNT (MAX_WORKSPACES+1)
@@ -59,6 +64,10 @@ int scr;
 int scrWidth;
 int scrHeight;
 
+
+Bool smooth = False;
+
+
 Pixmap CurrentPixmap = None;
 char *PixmapPath = NULL;
 
@@ -118,15 +127,14 @@ parseTexture(RContext *rc, char *text)
     char *tmp;
     char *type;
 
-#define GETSTR(val, str, i) \
+#define GETSTRORGOTO(val, str, i, label) \
 		val = PLGetArrayElement(texarray, i);\
 		if (!PLIsString(val)) {\
 			wwarning("could not parse texture %s", text);\
-			goto error;\
+			goto label;\
 		}\
 		str = PLGetString(val)
 
-    
     texarray = PLGetProplistWithDescription(text);
     if (!texarray || !PLIsArray(texarray) 
 	|| (count = PLGetNumberOfElements(texarray)) < 2) {
@@ -140,7 +148,7 @@ parseTexture(RContext *rc, char *text)
     texture = wmalloc(sizeof(BackgroundTexture));
     memset(texture, 0, sizeof(BackgroundTexture));
 
-    GETSTR(val, type, 0);
+    GETSTRORGOTO(val, type, 0, error);
 
     if (strcasecmp(type, "solid")==0) {
 	XColor color;
@@ -148,7 +156,7 @@ parseTexture(RContext *rc, char *text)
 
 	texture->solid = 1;
 	
-	GETSTR(val, tmp, 1);
+	GETSTRORGOTO(val, tmp, 1, error);
 
 	if (!XParseColor(dpy, DefaultColormap(dpy, scr), tmp, &color)) {
 	    wwarning("could not parse color %s in texture %s", tmp, text);
@@ -156,14 +164,14 @@ parseTexture(RContext *rc, char *text)
 	}
 	XAllocColor(dpy, DefaultColormap(dpy, scr), &color);
 
-	pixmap = XCreatePixmap(dpy, root, 32, 32, DefaultDepth(dpy, scr));
+	pixmap = XCreatePixmap(dpy, root, 8, 8, DefaultDepth(dpy, scr));
 	XSetForeground(dpy, DefaultGC(dpy, scr), color.pixel);
-	XFillRectangle(dpy, pixmap, DefaultGC(dpy, scr), 0, 0, 32, 32);
+	XFillRectangle(dpy, pixmap, DefaultGC(dpy, scr), 0, 0, 8, 8);
 
 	texture->pixmap = pixmap;
 	texture->color = color;
-	texture->width = 32;
-	texture->height = 32;
+	texture->width = 8;
+	texture->height = 8;
     } else if (strcasecmp(type, "vgradient")==0
 	       || strcasecmp(type, "dgradient")==0
 	       || strcasecmp(type, "hgradient")==0) {
@@ -174,7 +182,7 @@ parseTexture(RContext *rc, char *text)
 	int gtype;
 	int iwidth, iheight;
 
-	GETSTR(val, tmp, 1);
+	GETSTRORGOTO(val, tmp, 1, error);
 
 	if (!XParseColor(dpy, DefaultColormap(dpy, scr), tmp, &color)) {
 	    wwarning("could not parse color %s in texture %s", tmp, text);
@@ -185,7 +193,7 @@ parseTexture(RContext *rc, char *text)
 	color1.green = color.green >> 8;
 	color1.blue = color.blue >> 8;
 
-	GETSTR(val, tmp, 2);
+	GETSTRORGOTO(val, tmp, 2, error);
 
 	if (!XParseColor(dpy, DefaultColormap(dpy, scr), tmp, &color)) {
 	    wwarning("could not parse color %s in texture %s", tmp, text);
@@ -201,12 +209,12 @@ parseTexture(RContext *rc, char *text)
 	 case 'H':
 	    gtype = RHorizontalGradient;
 	    iwidth = scrWidth;
-	    iheight = 32;
+	    iheight = 8;
 	    break;
 	 case 'V':
 	 case 'v':
 	    gtype = RVerticalGradient;
-	    iwidth = 32;
+	    iwidth = 8;
 	    iheight = scrHeight;
 	    break;
 	 default:
@@ -294,12 +302,12 @@ parseTexture(RContext *rc, char *text)
 	 case 'H':
 	    gtype = RHorizontalGradient;
 	    iwidth = scrWidth;
-	    iheight = 32;
+	    iheight = 8;
 	    break;
 	 case 'V':
 	 case 'v':
 	    gtype = RVerticalGradient;
-	    iwidth = 32;
+	    iwidth = 8;
 	    iheight = scrHeight;
 	    break;
 	 default:
@@ -342,8 +350,10 @@ parseTexture(RContext *rc, char *text)
 	RImage *image = NULL;
 	int w, h;
 	int iwidth, iheight;
+	RColor rcolor;
 
-	GETSTR(val, tmp, 1);
+
+	GETSTRORGOTO(val, tmp, 1, error);
 /*
 	if (toupper(type[0]) == 'T' || toupper(type[0]) == 'C')
 	    pixmap = LoadJPEG(rc, tmp, &iwidth, &iheight);
@@ -358,7 +368,7 @@ parseTexture(RContext *rc, char *text)
 	    iheight = image->height;
 	}
 
-	GETSTR(val, tmp, 2);
+	GETSTRORGOTO(val, tmp, 2, error);
 
 	if (!XParseColor(dpy, DefaultColormap(dpy, scr), tmp, &color)) {
 	    wwarning("could not parse color %s in texture %s\n", tmp, text);
@@ -366,13 +376,20 @@ parseTexture(RContext *rc, char *text)
 	    goto error;
 	}
 	if (!XAllocColor(dpy, DefaultColormap(dpy, scr), &color)) {
-	    RColor rcolor;
-
 	    rcolor.red = color.red >> 8;
 	    rcolor.green = color.green >> 8;
 	    rcolor.blue = color.blue >> 8;
 	    RGetClosestXColor(rc, &rcolor, &color);
+	} else {
+	    rcolor.red = 0;
+	    rcolor.green = 0;
+	    rcolor.blue = 0;
 	}
+	/* for images with a transparent color */
+	if (image->data[3]) {
+	    RCombineImageWithColor(image, &rcolor);
+	}
+
 	switch (toupper(type[0])) {
 	 case 'T':
 	    texture->width = iwidth;
@@ -400,10 +417,13 @@ parseTexture(RContext *rc, char *text)
 		    w = (scrHeight*iwidth)/iheight;
 		}
 	    }
-	    {
+	    if (w != image->width || h != image->height) {
 		RImage *simage;
 
-		simage = RScaleImage(image, w, h);
+		if (smooth)
+		    simage = RSmoothScaleImage(image, w, h);
+		else
+		    simage = RScaleImage(image, w, h);
 		if (!simage) {
 		    wwarning("could not scale image:%s", 
 			     RMessageForError(RErrorCode));
@@ -412,9 +432,10 @@ parseTexture(RContext *rc, char *text)
 		}
 		RDestroyImage(image);
 		image = simage;
-		iwidth = image->width;
-		iheight = image->height;
 	    }
+	    iwidth = image->width;
+	    iheight = image->height;
+
 	    /* fall through */
 	 case 'C':
 	    {
@@ -486,13 +507,13 @@ parseTexture(RContext *rc, char *text)
 	int gtype;
 	int twidth, theight;
 
-	GETSTR(val, file, 1);
+	GETSTRORGOTO(val, file, 1, error);
 
-	GETSTR(val, tmp, 2);
+	GETSTRORGOTO(val, tmp, 2, error);
 
 	opaq = atoi(tmp);
 
-	GETSTR(val, tmp, 3);
+	GETSTRORGOTO(val, tmp, 3, error);
 
 	if (!XParseColor(dpy, DefaultColormap(dpy, scr), tmp, &color)) {
 	    wwarning("could not parse color %s in texture %s", tmp, text);
@@ -503,7 +524,7 @@ parseTexture(RContext *rc, char *text)
 	color1.green = color.green >> 8;
 	color1.blue = color.blue >> 8;
 
-	GETSTR(val, tmp, 4);
+	GETSTRORGOTO(val, tmp, 4, error);
 
 	if (!XParseColor(dpy, DefaultColormap(dpy, scr), tmp, &color)) {
 	    wwarning("could not parse color %s in texture %s", tmp, text);
@@ -516,7 +537,6 @@ parseTexture(RContext *rc, char *text)
 
 	image = loadImage(rc, file);
 	if (!image) {
-	    RDestroyImage(gradient);
 	    goto error;
 	}
 
@@ -544,7 +564,8 @@ parseTexture(RContext *rc, char *text)
 	if (!gradient) {
 	    wwarning("could not render texture:%s",
 		     RMessageForError(RErrorCode));
-	    RDestroyImage(gradient);
+            RDestroyImage(gradient);
+            RDestroyImage(image);
 	    goto error;
 	}
 
@@ -564,7 +585,7 @@ parseTexture(RContext *rc, char *text)
 	if (!RConvertImage(rc, tiled, &pixmap)) {
 	    wwarning("could not convert texture:%s", 
 		     RMessageForError(RErrorCode));
-	    RDestroyImage(image);
+	    RDestroyImage(tiled);
 	    goto error;
 	}
 	texture->width = tiled->width;
@@ -573,6 +594,86 @@ parseTexture(RContext *rc, char *text)
 	RDestroyImage(tiled);
 
 	texture->pixmap = pixmap;	
+    } else if (strcasecmp(type, "function")==0) {
+#ifdef HAVE_DLFCN_H
+	void (*initFunc) (Display*, Colormap);
+	RImage* (*mainFunc) (int, char**, int, int, int);
+	Pixmap pixmap;
+	RImage *image = 0;
+	int success = 0;
+	char *lib, *func, **argv = 0;
+	void *handle = 0;
+	int i, argc;
+
+        if (count < 3)
+	    goto function_cleanup;
+
+	/* get the library name */
+	GETSTRORGOTO(val, lib, 1, function_cleanup);
+
+	/* get the function name */
+	GETSTRORGOTO(val, func, 2, function_cleanup);
+
+	argc = count - 2;
+	argv = (char**)wmalloc(argc * sizeof(char*));
+
+	/* get the parameters */
+	argv[0] = func;
+	for (i=0; i<argc-1; i++) {
+	    GETSTRORGOTO(val, tmp, 3+i, function_cleanup);
+	    argv[i+1] = wstrdup(tmp);
+	}
+
+	handle = dlopen(lib, RTLD_LAZY);
+	if (!handle) {
+            wwarning("could not find library %s", lib);
+	    goto function_cleanup;
+	}
+
+	initFunc = dlsym(handle, "initWindowMaker");
+        if (!initFunc) {
+	    wwarning("could not initialize library %s", lib);
+	    goto function_cleanup;
+        }
+        initFunc(dpy, DefaultColormap(dpy, scr));
+
+        mainFunc = dlsym(handle, func);
+        if (!mainFunc) {
+            wwarning("could not find function %s::%s", lib, func);
+	    goto function_cleanup;
+        }
+        image = mainFunc(argc, argv, scrWidth, scrHeight, 0);
+
+        if (!RConvertImage(rc, image, &pixmap)) {
+	    wwarning("could not convert texture:%s", 
+	        RMessageForError(RErrorCode));
+	    goto function_cleanup;
+	}
+	texture->width = scrWidth;
+        texture->height = scrHeight;
+	texture->pixmap = pixmap;	
+        success = 1;
+
+function_cleanup:
+        if (argv) {
+	    int i;
+	    for (i=0; i<argc; i++) {
+	        free(argv[i]);
+	    }
+	}
+	if (handle) {
+	    dlclose(handle);
+	}
+	if (image) {
+            RDestroyImage(image);
+	}
+	if (!success) {
+	    goto error;
+	}
+#else
+	wwarning("function textures not supported");
+	goto error;
+#endif
     } else {
 	wwarning("invalid texture type %s", text);
 	goto error;
@@ -604,6 +705,9 @@ freeTexture(BackgroundTexture *texture)
 	    && pixel[0]!=WhitePixelOfScreen(DefaultScreenOfDisplay(dpy)))
 	    XFreeColors(dpy, DefaultColormap(dpy, scr), pixel, 1, 0);
     }
+    if (texture->pixmap) {
+	XFreePixmap(dpy, texture->pixmap);
+    }
     free(texture->spec);
     free(texture);
 }
@@ -747,8 +851,9 @@ setPixmapProperty(Pixmap pixmap)
 void
 changeTexture(BackgroundTexture *texture)
 {
-    if (!texture)
+    if (!texture) {
 	return;
+    }
 
     if (texture->solid) {
 	XSetWindowBackground(dpy, root, texture->color.pixel);
@@ -863,10 +968,11 @@ helperLoop(RContext *rc)
 #ifdef DEBUG
 	    printf("change texture %i\n", workspace);
 #endif
-	    if (!textures[workspace])
+	    if (!textures[workspace]) {
 		changeTexture(textures[0]);
-	    else
+	    } else {
 		changeTexture(textures[workspace]);
+	    }
 	    break;
 
 	 case 'P':
@@ -904,6 +1010,10 @@ updateDomain(char *domain, char *key, char *texture)
 {
     char *program = "wdwrite";
 
+    system(wstrappend("wdwrite ", 
+		      wstrappend(domain, smooth ? " SmoothWorkspaceBack YES"
+				 : " SmoothWorkspaceBack NO")));
+
     execlp(program, program, domain, key, texture, NULL);
     wwarning("warning could not run \"%s\"", program);
 }
@@ -915,7 +1025,7 @@ globalDefaultsPathForDomain(char *domain)
 {
     char path[1024];
 
-    sprintf(path, "%s/%s", SYSCONFDIR, domain);
+    sprintf(path, "%s/WindowMaker/%s", SYSCONFDIR, domain);
 
     return wstrdup(path);
 }
@@ -981,6 +1091,8 @@ getValueForKey(char *domain, char *keyName)
     return val;
 }
 
+
+
 char*
 getPixmapPath(char *domain)
 {
@@ -1033,6 +1145,37 @@ getPixmapPath(char *domain)
 }
 
 
+char*
+getFullPixmapPath(char *file)
+{
+    char *tmp;
+
+    if (!PixmapPath || !(tmp = wfindfile(PixmapPath, file))) {
+	int bsize = 512;
+	char *path = wmalloc(bsize);
+	
+	while (!getcwd(path, bsize)) {
+	    free(path);
+	    bsize += bsize/2;
+	    path = malloc(bsize);
+	}
+
+	tmp = wstrappend(path, "/");
+	free(path);
+	path = wstrappend(tmp, file);
+	free(tmp);
+
+	return path;
+    }
+    
+    /* the file is in the PixmapPath */
+    free(tmp);
+
+    return wstrdup(file);
+}
+
+
+
 void
 wAbort()
 {
@@ -1052,6 +1195,7 @@ print_help(char *ProgName)
 P(" -display	 		display to use");
 P(" -d, --dither	 		dither image");
 P(" -m, --match			match  colors");
+P(" -S, --smooth			smooth scaled image");
 P(" -b, --back-color <color>	background color");
 P(" -t, --tile			tile   image");
 P(" -e, --center			center image");
@@ -1120,7 +1264,7 @@ main(int argc, char **argv)
     char *back_color = "gray20";
     char *image_name = NULL;
     char *domain = "WindowMaker";
-    int update=0, cpc=4, render_mode=RM_DITHER, obey_user=0;
+    int update=0, cpc=4, render_mode=RDitheredRendering, obey_user=0;
     char *texture = NULL;
     int workspace = -1;
     
@@ -1161,12 +1305,15 @@ main(int argc, char **argv)
 	    style = "mpixmap";	    
 	} else if (strcmp(argv[i], "-d")==0
 		   || strcmp(argv[i], "--dither")==0) {
-	    render_mode = RM_DITHER;
+	    render_mode = RDitheredRendering;
 	    obey_user++;
 	} else if (strcmp(argv[i], "-m")==0
 		   || strcmp(argv[i], "--match")==0) {
-	    render_mode = RM_MATCH;
+	    render_mode = RBestMatchRendering;
 	    obey_user++;
+	} else if (strcmp(argv[i], "-S")==0
+		   || strcmp(argv[i], "--smooth")==0) {
+	    smooth = True;
 	} else if (strcmp(argv[i], "-u")==0
 		   || strcmp(argv[i], "--update-wmaker")==0) {
 	    update++;
@@ -1243,13 +1390,27 @@ main(int argc, char **argv)
 
     
     PixmapPath = getPixmapPath(domain);
+    if (!smooth) {
+	proplist_t val;
+#if 0 /* some problem with Alpha... TODO: check if its right */
+	val = PLGetDictionaryEntry(domain, 
+				   PLMakeString("SmoothWorkspaceBack"));
+#else
+	val = getValueForKey(domain, "SmoothWorkspaceBack");
+#endif
+
+	if (val && PLIsString(val) && strcasecmp(PLGetString(val), "YES")==0)
+	    smooth = True;
+    }
 
     dpy = XOpenDisplay(display);
     if (!dpy) {
 	wfatal("could not open display");
 	exit(1);
     }
+#if 0
     XSynchronize(dpy, 1);
+#endif
 
     root = DefaultRootWindow(dpy);
 
@@ -1259,13 +1420,20 @@ main(int argc, char **argv)
     scrHeight = HeightOfScreen(DefaultScreenOfDisplay(dpy));
 
     if (!obey_user && DefaultDepth(dpy, scr) <= 8)
-        render_mode = RM_DITHER;
+        render_mode = RDitheredRendering;
 
-    rattr.flags = RC_RenderMode | RC_ColorsPerChannel | RC_DefaultVisual;
+    rattr.flags = RC_RenderMode | RC_ColorsPerChannel
+	| RC_StandardColormap;
     rattr.render_mode = render_mode;
     rattr.colors_per_channel = cpc;
+    rattr.standard_colormap_mode = RCreateStdColormap;
 
     rc = RCreateContext(dpy, scr, &rattr);
+    if (!rc) {
+	wfatal("could not initialize wrlib:",
+	       RMessageForError(RErrorCode));
+	exit(1);
+    }
 
     if (helperMode) {
 	/* lower priority, so that it wont use all the CPU */
@@ -1277,7 +1445,10 @@ main(int argc, char **argv)
 	char buffer[4098];
 
 	if (!texture) {
-	    sprintf(buffer, "(%s, \"%s\", %s)", style, image_name, back_color);
+	    char *image_path = getFullPixmapPath(image_name);
+	    
+	    sprintf(buffer, "(%s, \"%s\", %s)", style, image_path, back_color);
+	    free(image_path);
 	    texture = (char*)buffer;
 	}
 
@@ -1293,7 +1464,7 @@ main(int argc, char **argv)
 	    changeTexture(tex);
 	else {
 	    /* always update domain */
-	    changeTextureForWorkspace(domain, texture, workspace-1);
+	    changeTextureForWorkspace(domain, texture, workspace);
 	}
     }
 
