@@ -2,7 +2,7 @@
  * 
  *  Raster graphics library
  * 
- *  Copyright (c) 1997 Alfredo K. Kojima
+ *  Copyright (c) 1997, 1998, 1999 Alfredo K. Kojima
  * 
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
@@ -32,23 +32,139 @@
 
 #include <math.h>
 
+#include "StdCmap.h"
+
 #include "wraster.h"
 
+
+extern void _wraster_change_filter(int type);
+
+
 static Bool bestContext(Display *dpy, int screen_number, RContext *context);
 
 static RContextAttributes DEFAULT_CONTEXT_ATTRIBS = {
     RC_UseSharedMemory|RC_RenderMode|RC_ColorsPerChannel, /* flags */
-	RM_DITHER, 		       /* render_mode */
+	RDitheredRendering, 		       /* render_mode */
 	4,			       /* colors_per_channel */
 	0, 
 	0,
 	0,
 	0,
-	True				   /* use_shared_memory */
+	True,				   /* use_shared_memory */
+	RMitchellFilter,
+	RUseStdColormap
 };
 
 
-static XColor*
+
+/*
+ * 
+ * Colormap allocation for PseudoColor visuals:
+ * 
+ * 
+ * switch standardColormap:
+ * 	none:
+ * 		allocate colors according to colors_per_channel
+ * 
+ * 	best/default:
+ * 		if there's a std colormap defined then use it
+ * 
+ * 		else
+ * 			create a std colormap and set it
+ */
+
+
+
+
+/*
+ *----------------------------------------------------------------------
+ * allocateStandardPseudoColor
+ * 	Creates the internal colormap for PseudoColor, setting the
+ * color values according to the supplied standard colormap.
+ * 
+ * Returns: -
+ *
+ * Side effects: -
+ *
+ * Notes: -
+ *----------------------------------------------------------------------
+ */
+static Bool
+allocateStandardPseudoColor(RContext *ctx, XStandardColormap *stdcmap)
+{
+    int i;
+
+    ctx->ncolors = stdcmap->red_max * stdcmap->red_mult
+	+ stdcmap->green_max * stdcmap->green_mult
+	+ stdcmap->blue_max * stdcmap->blue_mult + 1;
+
+    if (ctx->ncolors <= 1) {
+	RErrorCode = RERR_INTERNAL;
+	puts("wraster: bad standard colormap");
+
+	return False;
+    }
+    
+    ctx->colors = malloc(sizeof(XColor)*ctx->ncolors);
+    if (!ctx->colors) {
+	RErrorCode = RERR_NOMEMORY;
+
+	return False;
+    }
+    
+    ctx->pixels = malloc(sizeof(unsigned long)*ctx->ncolors);
+    if (!ctx->pixels) {
+	
+	free(ctx->colors);
+	ctx->colors = NULL;
+	
+	RErrorCode = RERR_NOMEMORY;
+
+	return False;
+    }
+
+    
+#define calc(max,mult) (((i / stdcmap->mult) % \
+                         (stdcmap->max + 1)) * 65535) / stdcmap->max
+
+    for (i = 0; i < ctx->ncolors; i++) {
+	ctx->colors[i].pixel = i + stdcmap->base_pixel;
+	ctx->colors[i].red = calc(red_max, red_mult);
+	ctx->colors[i].green = calc(green_max, green_mult);
+	ctx->colors[i].blue = calc(blue_max, blue_mult);
+	
+	ctx->pixels[i] = ctx->colors[i].pixel;
+    }
+
+#undef calc
+
+    return True;
+}
+
+
+static Bool
+setupStandardColormap(RContext *ctx, Atom property)
+{
+    if (!XmuLookupStandardColormap(ctx->dpy, ctx->screen_number,
+				   ctx->visual->visualid,
+				   ctx->depth, property,
+				   True, True)) {
+	RErrorCode = RERR_STDCMAPFAIL;
+	
+	return False;
+    }
+    return True;
+}
+
+
+
+
+
+
+
+
+
+static Bool
 allocatePseudoColor(RContext *ctx)
 {
     XColor *colors;
@@ -60,10 +176,10 @@ allocatePseudoColor(RContext *ctx)
     
     ncolors = cpc * cpc * cpc;
     
-    if ( ncolors > (1<<ctx->depth) ) {
-      /* reduce colormap size */
-      cpc = ctx->attribs->colors_per_channel = 1<<((int)ctx->depth/3);
-      ncolors = cpc * cpc * cpc;
+    if (ncolors > (1<<ctx->depth)) {
+	/* reduce colormap size */
+	cpc = ctx->attribs->colors_per_channel = 1<<((int)ctx->depth/3);
+	ncolors = cpc * cpc * cpc;
     }
 
     assert(cpc >= 2 && ncolors <= (1<<ctx->depth));
@@ -71,8 +187,16 @@ allocatePseudoColor(RContext *ctx)
     colors = malloc(sizeof(XColor)*ncolors);
     if (!colors) {
 	RErrorCode = RERR_NOMEMORY;
-	return NULL;
+	return False;
     }
+
+    ctx->pixels = malloc(sizeof(unsigned long)*ncolors);
+    if (!ctx->pixels) {
+	free(colors);
+	RErrorCode = RERR_NOMEMORY;
+	return False;
+    }
+    
     i=0;
 
     if ((ctx->attribs->flags & RC_GammaCorrection) && ctx->attribs->rgamma > 0
@@ -168,7 +292,16 @@ allocatePseudoColor(RContext *ctx)
 	    }
 	}
     }
-    return colors;
+    
+    ctx->colors = colors;
+    ctx->ncolors = ncolors;
+    
+    /* fill the pixels shortcut array */
+    for (i = 0; i < ncolors; i++) {
+	ctx->pixels[i] = ctx->colors[i].pixel;
+    }
+    
+    return True;
 }
 
 
@@ -199,7 +332,7 @@ allocateGrayScale(RContext *ctx)
 
     if (ncolors>=256 && ctx->vclass==StaticGray) {
         /* don't need dithering for 256 levels of gray in StaticGray visual */
-        ctx->attribs->render_mode = RM_MATCH;
+        ctx->attribs->render_mode = RBestMatchRendering;
     }
 
     colors = malloc(sizeof(XColor)*ncolors);
@@ -278,6 +411,88 @@ allocateGrayScale(RContext *ctx)
 }
 
 
+static Bool
+setupPseudoColorColormap(RContext *context)
+{
+    Atom property = 0;
+    
+    if (context->attribs->standard_colormap_mode == RCreateStdColormap) {
+	property = XInternAtom(context->dpy, "RGB_DEFAULT_MAP", False);
+	
+	if (!setupStandardColormap(context, property)) {
+	    return False;
+	}
+    }
+
+    if (context->attribs->standard_colormap_mode != RIgnoreStdColormap) {
+	XStandardColormap *maps;
+	int count, i;
+
+	if (!property) {
+	    property = XInternAtom(context->dpy, "RGB_BEST_MAP", False);
+	    if (!XGetRGBColormaps(context->dpy, 
+				  DefaultRootWindow(context->dpy),
+				  &maps, &count, property)) {
+		maps = NULL;
+	    }
+	    
+	    if (!maps) {
+		property = XInternAtom(context->dpy, "RGB_DEFAULT_MAP", False);
+		if (!XGetRGBColormaps(context->dpy,
+				      DefaultRootWindow(context->dpy),
+				      &maps, &count, property)) {
+		    maps = NULL;
+		}
+	    }
+	} else {
+	    if (!XGetRGBColormaps(context->dpy, 
+				  DefaultRootWindow(context->dpy),
+				  &maps, &count, property)) {
+		maps = NULL;
+	    }
+	}
+	
+	if (maps) {	    
+	    int theMap = -1;
+	    
+	    for (i = 0; i < count; i++) {
+		if (maps[i].visualid == context->visual->visualid) {
+		    theMap = i;
+		    break;
+		}
+	    }
+
+	    if (theMap < 0) {
+		puts("wrlib: no std cmap found");
+	    }
+
+	    if (theMap >= 0
+		&& allocateStandardPseudoColor(context, &maps[theMap])) {
+
+		context->std_rgb_map = XAllocStandardColormap();
+		
+		*context->std_rgb_map = maps[theMap];
+		
+		context->cmap = context->std_rgb_map->colormap;
+
+		XFree(maps);
+
+		return True;
+	    }
+	    
+	    XFree(maps);
+	}
+    }
+    
+    context->attribs->standard_colormap_mode = RIgnoreStdColormap;
+
+    /* RIgnoreStdColormap and fallback */
+    return allocatePseudoColor(context);
+}
+
+
+
+
 static char*
 mygetenv(char *var, int scr)
 {
@@ -322,6 +537,14 @@ gatherconfig(RContext *context, int screen_n)
 	    context->attribs->colors_per_channel = i;
 	}
     }
+    
+    ptr = mygetenv("WRASTER_OPTIMIZE_FOR_SPEED", screen_n);
+    if (ptr) {
+	context->flags.optimize_for_speed = 1;
+    } else {
+	context->flags.optimize_for_speed = 0;
+    }
+    
 }
 
 
@@ -407,9 +630,20 @@ RCreateContext(Display *dpy, int screen_number, RContextAttributes *attribs)
     else
 	*context->attribs = *attribs;
 
+    if (!(context->attribs->flags & RC_StandardColormap)) {
+	context->attribs->standard_colormap_mode = RUseStdColormap;
+    }
+
+    if (!(context->attribs->flags & RC_ScalingFilter)) {
+	context->attribs->flags |= RC_ScalingFilter;
+	context->attribs->scaling_filter = RMitchellFilter;
+    }
+
     /* get configuration from environment variables */
     gatherconfig(context, screen_number);
-    
+#ifndef BENCH
+    _wraster_change_filter(context->attribs->scaling_filter);
+#endif
     if ((context->attribs->flags & RC_VisualID)) {
 	XVisualInfo *vinfo, templ;
 	int nret;
@@ -468,10 +702,11 @@ RCreateContext(Display *dpy, int screen_number, RContextAttributes *attribs)
 				 |GCGraphicsExposures, &gcv);
 
     if (context->vclass == PseudoColor || context->vclass == StaticColor) {
-	context->colors = allocatePseudoColor(context);
-	if (!context->colors) {
+	
+	if (!setupPseudoColorColormap(context)) {
 	    return NULL;
 	}
+    
     } else if (context->vclass == GrayScale || context->vclass == StaticGray) {
 	context->colors = allocateGrayScale(context);
 	if (!context->colors) {
@@ -484,7 +719,7 @@ RCreateContext(Display *dpy, int screen_number, RContextAttributes *attribs)
 	context->blue_offset = count_offset(context->visual->blue_mask);
 	/* disable dithering on 24 bits visuals */
 	if (context->depth >= 24)
-	    context->attribs->render_mode = RM_MATCH;
+	    context->attribs->render_mode = RBestMatchRendering;
     }
     
     /* check avaiability of MIT-SHM */
